/* Generated by Cython 0.27.2 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_27_2"
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX < 0x030700A0 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject **args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__thinc__neural___funcs
#define __PYX_HAVE_API__thinc__neural___funcs
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <vector>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include "math.h"
#include "stdlib.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "_funcs.pyx",
  "linalg.pxd",
  "structs.pxd",
  "cymem.pxd",
  "maps.pxd",
};
/* None.proto */
#if defined(__GNUC__)
#define __Pyx_PACKED __attribute__((__packed__))
#else
#define __Pyx_PACKED
#endif


/* "typedefs.pxd":7
 * 
 * 
 * ctypedef float weight_t             # <<<<<<<<<<<<<<
 * ctypedef uint64_t atom_t
 * ctypedef uint64_t feat_t
 */
typedef float __pyx_t_5thinc_8typedefs_weight_t;

/* "typedefs.pxd":8
 * 
 * ctypedef float weight_t
 * ctypedef uint64_t atom_t             # <<<<<<<<<<<<<<
 * ctypedef uint64_t feat_t
 * ctypedef uint64_t hash_t
 */
typedef uint64_t __pyx_t_5thinc_8typedefs_atom_t;

/* "typedefs.pxd":9
 * ctypedef float weight_t
 * ctypedef uint64_t atom_t
 * ctypedef uint64_t feat_t             # <<<<<<<<<<<<<<
 * ctypedef uint64_t hash_t
 * ctypedef int32_t class_t
 */
typedef uint64_t __pyx_t_5thinc_8typedefs_feat_t;

/* "typedefs.pxd":10
 * ctypedef uint64_t atom_t
 * ctypedef uint64_t feat_t
 * ctypedef uint64_t hash_t             # <<<<<<<<<<<<<<
 * ctypedef int32_t class_t
 * ctypedef uint32_t count_t
 */
typedef uint64_t __pyx_t_5thinc_8typedefs_hash_t;

/* "typedefs.pxd":11
 * ctypedef uint64_t feat_t
 * ctypedef uint64_t hash_t
 * ctypedef int32_t class_t             # <<<<<<<<<<<<<<
 * ctypedef uint32_t count_t
 * ctypedef uint32_t time_t
 */
typedef int32_t __pyx_t_5thinc_8typedefs_class_t;

/* "typedefs.pxd":12
 * ctypedef uint64_t hash_t
 * ctypedef int32_t class_t
 * ctypedef uint32_t count_t             # <<<<<<<<<<<<<<
 * ctypedef uint32_t time_t
 * ctypedef int32_t len_t
 */
typedef uint32_t __pyx_t_5thinc_8typedefs_count_t;

/* "typedefs.pxd":13
 * ctypedef int32_t class_t
 * ctypedef uint32_t count_t
 * ctypedef uint32_t time_t             # <<<<<<<<<<<<<<
 * ctypedef int32_t len_t
 * ctypedef int32_t idx_t
 */
typedef uint32_t __pyx_t_5thinc_8typedefs_time_t;

/* "typedefs.pxd":14
 * ctypedef uint32_t count_t
 * ctypedef uint32_t time_t
 * ctypedef int32_t len_t             # <<<<<<<<<<<<<<
 * ctypedef int32_t idx_t
 * 
 */
typedef int32_t __pyx_t_5thinc_8typedefs_len_t;

/* "typedefs.pxd":15
 * ctypedef uint32_t time_t
 * ctypedef int32_t len_t
 * ctypedef int32_t idx_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef int32_t __pyx_t_5thinc_8typedefs_idx_t;

/* "preshed/maps.pxd":5
 * 
 * 
 * ctypedef uint64_t key_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef uint64_t __pyx_t_7preshed_4maps_key_t;

/*--- Type declarations ---*/
struct __pyx_obj_5cymem_5cymem_Pool;
struct __pyx_obj_5cymem_5cymem_Address;
struct __pyx_obj_7preshed_4maps_PreshMap;
struct __pyx_obj_7preshed_4maps_PreshMapArray;
struct __pyx_obj_5thinc_6linalg_Matrix;
struct __pyx_obj_5thinc_6linalg_Vec;
struct __pyx_obj_5thinc_6linalg_VecVec;
struct __pyx_obj_5thinc_6linalg_Mat;
struct __pyx_obj_5thinc_6linalg_MatVec;
struct __pyx_obj_5thinc_6linalg_MatMat;
struct __pyx_t_7preshed_4maps_Cell;
struct __pyx_t_7preshed_4maps_MapStruct;

/* "preshed/maps.pxd":8
 * 
 * 
 * cdef struct Cell:             # <<<<<<<<<<<<<<
 *     key_t key
 *     void* value
 */
struct __pyx_t_7preshed_4maps_Cell {
  __pyx_t_7preshed_4maps_key_t key;
  void *value;
};

/* "preshed/maps.pxd":13
 * 
 * 
 * cdef struct MapStruct:             # <<<<<<<<<<<<<<
 *     Cell* cells
 *     void* value_for_empty_key
 */
struct __pyx_t_7preshed_4maps_MapStruct {
  struct __pyx_t_7preshed_4maps_Cell *cells;
  void *value_for_empty_key;
  void *value_for_del_key;
  __pyx_t_7preshed_4maps_key_t length;
  __pyx_t_7preshed_4maps_key_t filled;
  int is_empty_key_set;
  int is_del_key_set;
};
struct __pyx_t_5thinc_7structs_ConstantsC;
struct __pyx_t_5thinc_7structs_EmbedC;
struct __pyx_t_5thinc_7structs_NeuralNetC;
struct __pyx_t_5thinc_7structs_ExampleC;
struct __pyx_t_5thinc_7structs_MinibatchC;
struct __pyx_t_5thinc_7structs_SparseArrayC;
struct __pyx_t_5thinc_7structs_FeatureC;
struct __pyx_t_5thinc_7structs_SparseAverageC;
struct __pyx_t_5thinc_7structs_TemplateC;

/* "structs.pxd":14
 * include "compile_time_constants.pxi"
 * 
 * ctypedef vector[weight_t] vector_weight_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef std::vector<__pyx_t_5thinc_8typedefs_weight_t>  __pyx_t_5thinc_7structs_vector_weight_t;

/* "structs.pxd":17
 * 
 * 
 * ctypedef void (*do_update_t)(             # <<<<<<<<<<<<<<
 *     weight_t* weights,
 *     weight_t* gradient,
 */
typedef void (*__pyx_t_5thinc_7structs_do_update_t)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_len_t, struct __pyx_t_5thinc_7structs_ConstantsC const *);

/* "structs.pxd":25
 * 
 * 
 * ctypedef void (*do_feed_fwd_t)(             # <<<<<<<<<<<<<<
 *     weight_t** fwd,
 *         const weight_t* W,
 */
typedef void (*__pyx_t_5thinc_7structs_do_feed_fwd_t)(__pyx_t_5thinc_8typedefs_weight_t **, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_len_t const *, int, int, struct __pyx_t_5thinc_7structs_ConstantsC const *);

/* "structs.pxd":35
 * 
 * 
 * ctypedef void (*do_feed_bwd_t)(             # <<<<<<<<<<<<<<
 *     weight_t* G,
 *     weight_t** bwd,
 */
typedef void (*__pyx_t_5thinc_7structs_do_feed_bwd_t)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t **, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *const *, __pyx_t_5thinc_8typedefs_len_t const *, int, int, struct __pyx_t_5thinc_7structs_ConstantsC const *);

/* "structs.pxd":48
 * 
 * # Alias this, so that it matches our naming scheme
 * ctypedef MapStruct MapC             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef struct __pyx_t_7preshed_4maps_MapStruct __pyx_t_5thinc_7structs_MapC;

/* "structs.pxd":51
 * 
 * 
 * cdef struct ConstantsC:             # <<<<<<<<<<<<<<
 *     weight_t a
 *     weight_t b
 */
struct __pyx_t_5thinc_7structs_ConstantsC {
  __pyx_t_5thinc_8typedefs_weight_t a;
  __pyx_t_5thinc_8typedefs_weight_t b;
  __pyx_t_5thinc_8typedefs_weight_t c;
  __pyx_t_5thinc_8typedefs_weight_t d;
  __pyx_t_5thinc_8typedefs_weight_t e;
  __pyx_t_5thinc_8typedefs_weight_t g;
  __pyx_t_5thinc_8typedefs_weight_t h;
  __pyx_t_5thinc_8typedefs_weight_t i;
  __pyx_t_5thinc_8typedefs_weight_t j;
  __pyx_t_5thinc_8typedefs_weight_t k;
  __pyx_t_5thinc_8typedefs_weight_t l;
  __pyx_t_5thinc_8typedefs_weight_t m;
  __pyx_t_5thinc_8typedefs_weight_t n;
  __pyx_t_5thinc_8typedefs_weight_t o;
  __pyx_t_5thinc_8typedefs_weight_t p;
  __pyx_t_5thinc_8typedefs_weight_t q;
  __pyx_t_5thinc_8typedefs_weight_t r;
  __pyx_t_5thinc_8typedefs_weight_t s;
  __pyx_t_5thinc_8typedefs_weight_t t;
  __pyx_t_5thinc_8typedefs_weight_t u;
  __pyx_t_5thinc_8typedefs_weight_t w;
  __pyx_t_5thinc_8typedefs_weight_t x;
  __pyx_t_5thinc_8typedefs_weight_t y;
  __pyx_t_5thinc_8typedefs_weight_t z;
};

/* "structs.pxd":78
 * 
 * 
 * cdef struct EmbedC:             # <<<<<<<<<<<<<<
 *     MapC** weights
 *     MapC** gradients
 */
struct __pyx_t_5thinc_7structs_EmbedC {
  __pyx_t_5thinc_7structs_MapC **weights;
  __pyx_t_5thinc_7structs_MapC **gradients;
  __pyx_t_5thinc_8typedefs_weight_t **defaults;
  __pyx_t_5thinc_8typedefs_weight_t **d_defaults;
  __pyx_t_5thinc_8typedefs_idx_t *offsets;
  __pyx_t_5thinc_8typedefs_len_t *lengths;
  __pyx_t_5thinc_8typedefs_len_t nr;
  int nr_support;
};

/* "structs.pxd":89
 * 
 * 
 * cdef struct NeuralNetC:             # <<<<<<<<<<<<<<
 *     do_feed_fwd_t feed_fwd
 *     do_feed_bwd_t feed_bwd
 */
struct __pyx_t_5thinc_7structs_NeuralNetC {
  __pyx_t_5thinc_7structs_do_feed_fwd_t feed_fwd;
  __pyx_t_5thinc_7structs_do_feed_bwd_t feed_bwd;
  __pyx_t_5thinc_7structs_do_update_t update;
  __pyx_t_5thinc_8typedefs_len_t *widths;
  __pyx_t_5thinc_8typedefs_weight_t *weights;
  __pyx_t_5thinc_8typedefs_weight_t *gradient;
  struct __pyx_t_5thinc_7structs_EmbedC *embed;
  __pyx_t_5thinc_8typedefs_len_t nr_layer;
  __pyx_t_5thinc_8typedefs_len_t nr_weight;
  __pyx_t_5thinc_8typedefs_len_t nr_node;
  struct __pyx_t_5thinc_7structs_ConstantsC hp;
};

/* "structs.pxd":112
 * 
 * 
 * cdef struct ExampleC:             # <<<<<<<<<<<<<<
 *     int* is_valid
 *     weight_t* costs
 */
struct __pyx_t_5thinc_7structs_ExampleC {
  int *is_valid;
  __pyx_t_5thinc_8typedefs_weight_t *costs;
  uint64_t *atoms;
  struct __pyx_t_5thinc_7structs_FeatureC *features;
  __pyx_t_5thinc_8typedefs_weight_t *scores;
  int nr_class;
  int nr_atom;
  int nr_feat;
};
struct __pyx_t_5thinc_7structs_MinibatchC {

  /* "structs.pxd":124
 * 
 * 
 * cdef cppclass MinibatchC:             # <<<<<<<<<<<<<<
 *     weight_t** _fwd
 *     weight_t** _bwd
 */
  __pyx_t_5thinc_8typedefs_weight_t **_fwd;
  __pyx_t_5thinc_8typedefs_weight_t **_bwd;
  struct __pyx_t_5thinc_7structs_FeatureC **_feats;
  __pyx_t_5thinc_8typedefs_len_t *_nr_feat;
  __pyx_t_5thinc_8typedefs_weight_t *_costs;
  int *_is_valid;
  uint64_t *signatures;
  __pyx_t_5thinc_8typedefs_len_t *widths;
  int i;
  int nr_layer;
  int batch_size;
  void __pyx_f___init__MinibatchC(__pyx_t_5thinc_8typedefs_len_t *, int, int);
  void __pyx_f___dealloc__MinibatchC(void);
  virtual void reset(void);
  virtual int nr_in(void);
  virtual int nr_out(void);
  virtual int push_back(struct __pyx_t_5thinc_7structs_FeatureC const *, int, __pyx_t_5thinc_8typedefs_weight_t const *, int const *, uint64_t);
  virtual struct __pyx_t_5thinc_7structs_FeatureC *features(int);
  virtual int nr_feat(int);
  virtual __pyx_t_5thinc_8typedefs_weight_t *fwd(int, int);
  virtual __pyx_t_5thinc_8typedefs_weight_t *bwd(int, int);
  virtual __pyx_t_5thinc_8typedefs_weight_t *scores(int);
  virtual __pyx_t_5thinc_8typedefs_weight_t *losses(int);
  virtual __pyx_t_5thinc_8typedefs_weight_t *costs(int);
  virtual int *is_valid(int);
  virtual int guess(int);
  virtual int best(int);
  __pyx_t_5thinc_7structs_MinibatchC(__pyx_t_5thinc_8typedefs_len_t *__pyx_v_widths, int __pyx_v_nr_layer, int __pyx_v_batch_size) {
    __pyx_f___init__MinibatchC(__pyx_v_widths, __pyx_v_nr_layer, __pyx_v_batch_size);
  }
  virtual ~__pyx_t_5thinc_7structs_MinibatchC() {
    __pyx_f___dealloc__MinibatchC();
  }
};

/* "structs.pxd":258
 * 
 * 
 * cdef packed struct SparseArrayC:             # <<<<<<<<<<<<<<
 *     int32_t key
 *     weight_t val
 */
#if defined(__SUNPRO_C)
  #pragma pack(1)
#elif !defined(__GNUC__)
  #pragma pack(push, 1)
#endif
struct __Pyx_PACKED __pyx_t_5thinc_7structs_SparseArrayC {
  int32_t key;
  __pyx_t_5thinc_8typedefs_weight_t val;
};
#if defined(__SUNPRO_C)
  #pragma pack()
#elif !defined(__GNUC__)
  #pragma pack(pop)
#endif

/* "structs.pxd":263
 * 
 * 
 * cdef struct FeatureC:             # <<<<<<<<<<<<<<
 *     int i
 *     uint64_t key
 */
struct __pyx_t_5thinc_7structs_FeatureC {
  int i;
  uint64_t key;
  __pyx_t_5thinc_8typedefs_weight_t value;
};

/* "structs.pxd":269
 * 
 * 
 * cdef struct SparseAverageC:             # <<<<<<<<<<<<<<
 *     SparseArrayC* curr
 *     SparseArrayC* mom1
 */
struct __pyx_t_5thinc_7structs_SparseAverageC {
  struct __pyx_t_5thinc_7structs_SparseArrayC *curr;
  struct __pyx_t_5thinc_7structs_SparseArrayC *mom1;
  struct __pyx_t_5thinc_7structs_SparseArrayC *mom2;
  struct __pyx_t_5thinc_7structs_SparseArrayC *avgs;
  struct __pyx_t_5thinc_7structs_SparseArrayC *times;
  struct __pyx_t_5thinc_7structs_SparseArrayC *penalties;
  __pyx_t_5thinc_8typedefs_weight_t penalty;
};

/* "structs.pxd":279
 * 
 * 
 * cdef struct TemplateC:             # <<<<<<<<<<<<<<
 *     int[MAX_TEMPLATE_LEN] indices
 *     int length
 */
struct __pyx_t_5thinc_7structs_TemplateC {
  int indices[10];
  int length;
  __pyx_t_5thinc_8typedefs_atom_t atoms[10];
};

/* "cymem/cymem.pxd":1
 * cdef class Pool:             # <<<<<<<<<<<<<<
 *     cdef readonly size_t size
 *     cdef readonly dict addresses
 */
struct __pyx_obj_5cymem_5cymem_Pool {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cymem_5cymem_Pool *__pyx_vtab;
  size_t size;
  PyObject *addresses;
  PyObject *refs;
};


/* "cymem/cymem.pxd":11
 * 
 * 
 * cdef class Address:             # <<<<<<<<<<<<<<
 *     cdef void* ptr
 */
struct __pyx_obj_5cymem_5cymem_Address {
  PyObject_HEAD
  void *ptr;
};


/* "preshed/maps.pxd":36
 * 
 * 
 * cdef class PreshMap:             # <<<<<<<<<<<<<<
 *     cdef MapStruct* c_map
 *     cdef Pool mem
 */
struct __pyx_obj_7preshed_4maps_PreshMap {
  PyObject_HEAD
  struct __pyx_vtabstruct_7preshed_4maps_PreshMap *__pyx_vtab;
  struct __pyx_t_7preshed_4maps_MapStruct *c_map;
  struct __pyx_obj_5cymem_5cymem_Pool *mem;
};


/* "preshed/maps.pxd":44
 * 
 * 
 * cdef class PreshMapArray:             # <<<<<<<<<<<<<<
 *     cdef Pool mem
 *     cdef MapStruct* maps
 */
struct __pyx_obj_7preshed_4maps_PreshMapArray {
  PyObject_HEAD
  struct __pyx_vtabstruct_7preshed_4maps_PreshMapArray *__pyx_vtab;
  struct __pyx_obj_5cymem_5cymem_Pool *mem;
  struct __pyx_t_7preshed_4maps_MapStruct *maps;
  size_t length;
};


/* "linalg.pxd":22
 * 
 * 
 * cdef class Matrix:             # <<<<<<<<<<<<<<
 *     cdef readonly Pool mem
 *     cdef weight_t* data
 */
struct __pyx_obj_5thinc_6linalg_Matrix {
  PyObject_HEAD
  struct __pyx_obj_5cymem_5cymem_Pool *mem;
  __pyx_t_5thinc_8typedefs_weight_t *data;
  int32_t nr_row;
  int32_t nr_col;
};


/* "linalg.pxd":29
 * 
 * 
 * cdef class Vec:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline int arg_max(const weight_t* scores, const int n_classes) nogil:
 */
struct __pyx_obj_5thinc_6linalg_Vec {
  PyObject_HEAD
  struct __pyx_vtabstruct_5thinc_6linalg_Vec *__pyx_vtab;
};


/* "linalg.pxd":138
 * 
 * 
 * cdef class VecVec:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline void add(weight_t* output,
 */
struct __pyx_obj_5thinc_6linalg_VecVec {
  PyObject_HEAD
  struct __pyx_vtabstruct_5thinc_6linalg_VecVec *__pyx_vtab;
};


/* "linalg.pxd":229
 * 
 * 
 * cdef class Mat:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline void mean_row(weight_t* Ex,
 */
struct __pyx_obj_5thinc_6linalg_Mat {
  PyObject_HEAD
  struct __pyx_vtabstruct_5thinc_6linalg_Mat *__pyx_vtab;
};


/* "linalg.pxd":257
 * 
 * 
 * cdef class MatVec:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline void add_i(weight_t* mat,
 */
struct __pyx_obj_5thinc_6linalg_MatVec {
  PyObject_HEAD
  struct __pyx_vtabstruct_5thinc_6linalg_MatVec *__pyx_vtab;
};


/* "linalg.pxd":423
 * 
 * 
 * cdef class MatMat:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline void add(weight_t* output,
 */
struct __pyx_obj_5thinc_6linalg_MatMat {
  PyObject_HEAD
  struct __pyx_vtabstruct_5thinc_6linalg_MatMat *__pyx_vtab;
};



/* "cymem/cymem.pxd":1
 * cdef class Pool:             # <<<<<<<<<<<<<<
 *     cdef readonly size_t size
 *     cdef readonly dict addresses
 */

struct __pyx_vtabstruct_5cymem_5cymem_Pool {
  void *(*alloc)(struct __pyx_obj_5cymem_5cymem_Pool *, size_t, size_t);
  void (*free)(struct __pyx_obj_5cymem_5cymem_Pool *, void *);
  void *(*realloc)(struct __pyx_obj_5cymem_5cymem_Pool *, void *, size_t);
};
static struct __pyx_vtabstruct_5cymem_5cymem_Pool *__pyx_vtabptr_5cymem_5cymem_Pool;


/* "preshed/maps.pxd":36
 * 
 * 
 * cdef class PreshMap:             # <<<<<<<<<<<<<<
 *     cdef MapStruct* c_map
 *     cdef Pool mem
 */

struct __pyx_vtabstruct_7preshed_4maps_PreshMap {
  void *(*get)(struct __pyx_obj_7preshed_4maps_PreshMap *, __pyx_t_7preshed_4maps_key_t);
  void (*set)(struct __pyx_obj_7preshed_4maps_PreshMap *, __pyx_t_7preshed_4maps_key_t, void *);
};
static struct __pyx_vtabstruct_7preshed_4maps_PreshMap *__pyx_vtabptr_7preshed_4maps_PreshMap;


/* "preshed/maps.pxd":44
 * 
 * 
 * cdef class PreshMapArray:             # <<<<<<<<<<<<<<
 *     cdef Pool mem
 *     cdef MapStruct* maps
 */

struct __pyx_vtabstruct_7preshed_4maps_PreshMapArray {
  void *(*get)(struct __pyx_obj_7preshed_4maps_PreshMapArray *, size_t, __pyx_t_7preshed_4maps_key_t);
  void (*set)(struct __pyx_obj_7preshed_4maps_PreshMapArray *, size_t, __pyx_t_7preshed_4maps_key_t, void *);
};
static struct __pyx_vtabstruct_7preshed_4maps_PreshMapArray *__pyx_vtabptr_7preshed_4maps_PreshMapArray;


/* "linalg.pxd":29
 * 
 * 
 * cdef class Vec:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline int arg_max(const weight_t* scores, const int n_classes) nogil:
 */

struct __pyx_vtabstruct_5thinc_6linalg_Vec {
  int (*arg_max)(__pyx_t_5thinc_8typedefs_weight_t const *, int const );
  __pyx_t_5thinc_8typedefs_weight_t (*max)(__pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
  __pyx_t_5thinc_8typedefs_weight_t (*sum)(__pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
  __pyx_t_5thinc_8typedefs_weight_t (*norm)(__pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
  void (*add)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*add_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*mul)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*mul_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*pow)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*pow_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const , int32_t);
  void (*div)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*div_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const , int32_t);
  void (*exp)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
  void (*exp_i)(__pyx_t_5thinc_8typedefs_weight_t *, int32_t);
  void (*reciprocal_i)(__pyx_t_5thinc_8typedefs_weight_t *, int32_t);
};
static struct __pyx_vtabstruct_5thinc_6linalg_Vec *__pyx_vtabptr_5thinc_6linalg_Vec;
static CYTHON_INLINE int __pyx_f_5thinc_6linalg_3Vec_arg_max(__pyx_t_5thinc_8typedefs_weight_t const *, int const );
static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_max(__pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_sum(__pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_norm(__pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_add(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_add_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_mul(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_pow(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_pow_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const , int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_div(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_div_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const , int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_exp(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_exp_i(__pyx_t_5thinc_8typedefs_weight_t *, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_reciprocal_i(__pyx_t_5thinc_8typedefs_weight_t *, int32_t);


/* "linalg.pxd":138
 * 
 * 
 * cdef class VecVec:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline void add(weight_t* output,
 */

struct __pyx_vtabstruct_5thinc_6linalg_VecVec {
  void (*add)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*add_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*batch_add_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t, int32_t);
  void (*add_pow)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*add_pow_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
  void (*mul)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
  void (*mul_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
  __pyx_t_5thinc_8typedefs_weight_t (*dot)(__pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
  int (*arg_max_if_true)(__pyx_t_5thinc_8typedefs_weight_t const *, int const *, int const );
  int (*arg_max_if_zero)(__pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int const );
};
static struct __pyx_vtabstruct_5thinc_6linalg_VecVec *__pyx_vtabptr_5thinc_6linalg_VecVec;
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_batch_add_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_pow(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_pow_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_mul(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_6VecVec_dot(__pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t);
static CYTHON_INLINE int __pyx_f_5thinc_6linalg_6VecVec_arg_max_if_true(__pyx_t_5thinc_8typedefs_weight_t const *, int const *, int const );
static CYTHON_INLINE int __pyx_f_5thinc_6linalg_6VecVec_arg_max_if_zero(__pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int const );


/* "linalg.pxd":229
 * 
 * 
 * cdef class Mat:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline void mean_row(weight_t* Ex,
 */

struct __pyx_vtabstruct_5thinc_6linalg_Mat {
  void (*mean_row)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*var_row)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t, __pyx_t_5thinc_8typedefs_weight_t);
};
static struct __pyx_vtabstruct_5thinc_6linalg_Mat *__pyx_vtabptr_5thinc_6linalg_Mat;
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Mat_mean_row(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Mat_var_row(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t, __pyx_t_5thinc_8typedefs_weight_t);


/* "linalg.pxd":257
 * 
 * 
 * cdef class MatVec:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline void add_i(weight_t* mat,
 */

struct __pyx_vtabstruct_5thinc_6linalg_MatVec {
  void (*add_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t, int32_t);
  void (*mul)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*mul_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*dot)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*batch_dot)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t, int32_t);
  void (*T_dot)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*batch_T_dot)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t, int32_t);
};
static struct __pyx_vtabstruct_5thinc_6linalg_MatVec *__pyx_vtabptr_5thinc_6linalg_MatVec;
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_add_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_mul(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_dot(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_batch_dot(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_T_dot(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_batch_T_dot(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t, int32_t);


/* "linalg.pxd":423
 * 
 * 
 * cdef class MatMat:             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef inline void add(weight_t* output,
 */

struct __pyx_vtabstruct_5thinc_6linalg_MatMat {
  void (*add)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*add_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*mul)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*mul_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*add_outer_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
  void (*batch_add_outer_i)(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t, int32_t);
};
static struct __pyx_vtabstruct_5thinc_6linalg_MatMat *__pyx_vtabptr_5thinc_6linalg_MatMat;
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_mul(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_mul_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add_outer_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t);
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_batch_add_outer_i(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, __pyx_t_5thinc_8typedefs_weight_t const *, int32_t, int32_t, int32_t);

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* Profile.proto */
#ifndef CYTHON_PROFILE
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
  #define CYTHON_PROFILE 0
#else
  #define CYTHON_PROFILE 1
#endif
#endif
#ifndef CYTHON_TRACE_NOGIL
  #define CYTHON_TRACE_NOGIL 0
#else
  #if CYTHON_TRACE_NOGIL && !defined(CYTHON_TRACE)
    #define CYTHON_TRACE 1
  #endif
#endif
#ifndef CYTHON_TRACE
  #define CYTHON_TRACE 0
#endif
#if CYTHON_TRACE
  #undef CYTHON_PROFILE_REUSE_FRAME
#endif
#ifndef CYTHON_PROFILE_REUSE_FRAME
  #define CYTHON_PROFILE_REUSE_FRAME 0
#endif
#if CYTHON_PROFILE || CYTHON_TRACE
  #include "compile.h"
  #include "frameobject.h"
  #include "traceback.h"
  #if CYTHON_PROFILE_REUSE_FRAME
    #define CYTHON_FRAME_MODIFIER static
    #define CYTHON_FRAME_DEL(frame)
  #else
    #define CYTHON_FRAME_MODIFIER
    #define CYTHON_FRAME_DEL(frame) Py_CLEAR(frame)
  #endif
  #define __Pyx_TraceDeclarations\
  static PyCodeObject *__pyx_frame_code = NULL;\
  CYTHON_FRAME_MODIFIER PyFrameObject *__pyx_frame = NULL;\
  int __Pyx_use_tracing = 0;
  #define __Pyx_TraceFrameInit(codeobj)\
  if (codeobj) __pyx_frame_code = (PyCodeObject*) codeobj;
  #ifdef WITH_THREAD
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  if (nogil) {\
      if (CYTHON_TRACE_NOGIL) {\
          PyThreadState *tstate;\
          PyGILState_STATE state = PyGILState_Ensure();\
          tstate = __Pyx_PyThreadState_Current;\
          if (unlikely(tstate->use_tracing) && !tstate->tracing &&\
                  (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc))) {\
              __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          }\
          PyGILState_Release(state);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  } else {\
      PyThreadState* tstate = PyThreadState_GET();\
      if (unlikely(tstate->use_tracing) && !tstate->tracing &&\
              (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc))) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #else
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  {   PyThreadState* tstate = PyThreadState_GET();\
      if (unlikely(tstate->use_tracing) && !tstate->tracing &&\
              (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc))) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #endif
  #define __Pyx_TraceException()\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (tstate->use_tracing &&\
              (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc))) {\
          tstate->tracing++;\
          tstate->use_tracing = 0;\
          PyObject *exc_info = __Pyx_GetExceptionTuple(tstate);\
          if (exc_info) {\
              if (CYTHON_TRACE && tstate->c_tracefunc)\
                  tstate->c_tracefunc(\
                      tstate->c_traceobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              tstate->c_profilefunc(\
                  tstate->c_profileobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              Py_DECREF(exc_info);\
          }\
          tstate->use_tracing = 1;\
          tstate->tracing--;\
      }\
  }
  static void __Pyx_call_return_trace_func(PyThreadState *tstate, PyFrameObject *frame, PyObject *result) {
      PyObject *type, *value, *traceback;
      PyErr_Fetch(&type, &value, &traceback);
      tstate->tracing++;
      tstate->use_tracing = 0;
      if (CYTHON_TRACE && tstate->c_tracefunc)
          tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_RETURN, result);
      if (tstate->c_profilefunc)
          tstate->c_profilefunc(tstate->c_profileobj, frame, PyTrace_RETURN, result);
      CYTHON_FRAME_DEL(frame);
      tstate->use_tracing = 1;
      tstate->tracing--;
      PyErr_Restore(type, value, traceback);
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (tstate->use_tracing) {\
                  __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (tstate->use_tracing) {\
              __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
          }\
      }\
  }
  #else
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (tstate->use_tracing) {\
          __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
      }\
  }
  #endif
  static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno);
  static int __Pyx_TraceSetupAndCall(PyCodeObject** code, PyFrameObject** frame, PyThreadState* tstate, const char *funcname, const char *srcfile, int firstlineno);
#else
  #define __Pyx_TraceDeclarations
  #define __Pyx_TraceFrameInit(codeobj)
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)   if ((1)); else goto_error;
  #define __Pyx_TraceException()
  #define __Pyx_TraceReturn(result, nogil)
#endif
#if CYTHON_TRACE
  static int __Pyx_call_line_trace_func(PyThreadState *tstate, PyFrameObject *frame, int lineno) {
      int ret;
      PyObject *type, *value, *traceback;
      PyErr_Fetch(&type, &value, &traceback);
      __Pyx_PyFrame_SetLineNumber(frame, lineno);
      tstate->tracing++;
      tstate->use_tracing = 0;
      ret = tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_LINE, NULL);
      tstate->use_tracing = 1;
      tstate->tracing--;
      if (likely(!ret)) {
          PyErr_Restore(type, value, traceback);
      } else {
          Py_XDECREF(type);
          Py_XDECREF(value);
          Py_XDECREF(traceback);
      }
      return ret;
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              int ret = 0;\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (unlikely(tstate->use_tracing && tstate->c_tracefunc && __pyx_frame->f_trace)) {\
                  ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              }\
              PyGILState_Release(state);\
              if (unlikely(ret)) goto_error;\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (unlikely(tstate->use_tracing && tstate->c_tracefunc && __pyx_frame->f_trace)) {\
              int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              if (unlikely(ret)) goto_error;\
          }\
      }\
  }
  #else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (unlikely(tstate->use_tracing && tstate->c_tracefunc && __pyx_frame->f_trace)) {\
          int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
          if (unlikely(ret)) goto_error;\
      }\
  }
  #endif
#else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)   if ((1)); else goto_error;
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static CYTHON_INLINE int __pyx_f_5thinc_6linalg_3Vec_arg_max(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_scores, int const __pyx_v_n_classes); /* proto*/
static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_max(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_sum(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_norm(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_add(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_inc, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_inc, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_mul(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scal, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scal, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_pow(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scal, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_pow_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t const __pyx_v_scal, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_div(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scal, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_div_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t const __pyx_v_scal, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_exp(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_exp_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_reciprocal_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scale, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scale, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_batch_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scale, int32_t __pyx_v_nr, int32_t __pyx_v_nr_batch); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_pow(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_power, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_pow_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_power, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_mul(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_6VecVec_dot(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr); /* proto*/
static CYTHON_INLINE int __pyx_f_5thinc_6linalg_6VecVec_arg_max_if_true(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_scores, int const *__pyx_v_is_valid, int const __pyx_v_n_classes); /* proto*/
static CYTHON_INLINE int __pyx_f_5thinc_6linalg_6VecVec_arg_max_if_zero(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_scores, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_costs, int const __pyx_v_n_classes); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Mat_mean_row(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_Ex, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Mat_var_row(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_Vx, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_Ex, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_eps); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scale, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_mul(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_dot(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_batch_dot(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col, int32_t __pyx_v_nr_batch); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_T_dot(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_batch_T_dot(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col, int32_t __pyx_v_nr_batch); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_mul(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_mul_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add_outer_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col); /* proto*/
static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_batch_add_outer_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col, int32_t __pyx_v_nr_batch); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'thinc.typedefs' */

/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cymem.cymem' */
static PyTypeObject *__pyx_ptype_5cymem_5cymem_Pool = 0;
static PyTypeObject *__pyx_ptype_5cymem_5cymem_Address = 0;

/* Module declarations from 'preshed.maps' */
static PyTypeObject *__pyx_ptype_7preshed_4maps_PreshMap = 0;
static PyTypeObject *__pyx_ptype_7preshed_4maps_PreshMapArray = 0;

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'murmurhash.mrmr' */
static uint64_t (*__pyx_f_10murmurhash_4mrmr_real_hash64)(void *, int, uint64_t); /*proto*/

/* Module declarations from 'thinc.linalg' */
static PyTypeObject *__pyx_ptype_5thinc_6linalg_Matrix = 0;
static PyTypeObject *__pyx_ptype_5thinc_6linalg_Vec = 0;
static PyTypeObject *__pyx_ptype_5thinc_6linalg_VecVec = 0;
static PyTypeObject *__pyx_ptype_5thinc_6linalg_Mat = 0;
static PyTypeObject *__pyx_ptype_5thinc_6linalg_MatVec = 0;
static PyTypeObject *__pyx_ptype_5thinc_6linalg_MatMat = 0;

/* Module declarations from 'thinc.structs' */

/* Module declarations from 'thinc.neural._funcs' */
static void __pyx_f_5thinc_6neural_6_funcs_clip_gradient(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t, int); /*proto*/
static void __pyx_f_5thinc_6neural_6_funcs_update_averages(__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int, __pyx_t_5thinc_8typedefs_weight_t); /*proto*/
#define __Pyx_MODULE_NAME "thinc.neural._funcs"
int __pyx_module_is_main_thinc__neural___funcs = 0;

/* Implementation of 'thinc.neural._funcs' */
static PyObject *__pyx_builtin_range;
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_test;

/* "thinc/neural/_funcs.pyx":21
 * 
 * 
 * cdef void ELU(weight_t* out, len_t nr_out) nogil:             # <<<<<<<<<<<<<<
 *     cdef idx_t i
 *     for i in range(nr_out):
 */

static void __pyx_f_5thinc_6neural_6_funcs_ELU(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_out, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_out) {
  __pyx_t_5thinc_8typedefs_idx_t __pyx_v_i;
  __Pyx_TraceDeclarations
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_1;
  __pyx_t_5thinc_8typedefs_idx_t __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("ELU", __pyx_f[0], 21, 1, __PYX_ERR(0, 21, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":23
 * cdef void ELU(weight_t* out, len_t nr_out) nogil:
 *     cdef idx_t i
 *     for i in range(nr_out):             # <<<<<<<<<<<<<<
 *         if out[i] < 0:
 *             out[i] = ALPHA * (exp(out[i]) - 1)
 */
  __pyx_t_1 = __pyx_v_nr_out;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":24
 *     cdef idx_t i
 *     for i in range(nr_out):
 *         if out[i] < 0:             # <<<<<<<<<<<<<<
 *             out[i] = ALPHA * (exp(out[i]) - 1)
 * 
 */
    __pyx_t_3 = (((__pyx_v_out[__pyx_v_i]) < 0.0) != 0);
    if (__pyx_t_3) {

      /* "thinc/neural/_funcs.pyx":25
 *     for i in range(nr_out):
 *         if out[i] < 0:
 *             out[i] = ALPHA * (exp(out[i]) - 1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_out[__pyx_v_i]) = (1.0 * (exp((__pyx_v_out[__pyx_v_i])) - 1.0));

      /* "thinc/neural/_funcs.pyx":24
 *     cdef idx_t i
 *     for i in range(nr_out):
 *         if out[i] < 0:             # <<<<<<<<<<<<<<
 *             out[i] = ALPHA * (exp(out[i]) - 1)
 * 
 */
    }
  }

  /* "thinc/neural/_funcs.pyx":21
 * 
 * 
 * cdef void ELU(weight_t* out, len_t nr_out) nogil:             # <<<<<<<<<<<<<<
 *     cdef idx_t i
 *     for i in range(nr_out):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.ELU", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":28
 * 
 * 
 * cdef void ReLu(weight_t* out, len_t nr_out) nogil:             # <<<<<<<<<<<<<<
 *     cdef idx_t i
 *     for i in range(nr_out):
 */

static void __pyx_f_5thinc_6neural_6_funcs_ReLu(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_out, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_out) {
  __pyx_t_5thinc_8typedefs_idx_t __pyx_v_i;
  __Pyx_TraceDeclarations
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_1;
  __pyx_t_5thinc_8typedefs_idx_t __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("ReLu", __pyx_f[0], 28, 1, __PYX_ERR(0, 28, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":30
 * cdef void ReLu(weight_t* out, len_t nr_out) nogil:
 *     cdef idx_t i
 *     for i in range(nr_out):             # <<<<<<<<<<<<<<
 *         if out[i] < 0:
 *             out[i] = 0
 */
  __pyx_t_1 = __pyx_v_nr_out;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":31
 *     cdef idx_t i
 *     for i in range(nr_out):
 *         if out[i] < 0:             # <<<<<<<<<<<<<<
 *             out[i] = 0
 * 
 */
    __pyx_t_3 = (((__pyx_v_out[__pyx_v_i]) < 0.0) != 0);
    if (__pyx_t_3) {

      /* "thinc/neural/_funcs.pyx":32
 *     for i in range(nr_out):
 *         if out[i] < 0:
 *             out[i] = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_out[__pyx_v_i]) = 0.0;

      /* "thinc/neural/_funcs.pyx":31
 *     cdef idx_t i
 *     for i in range(nr_out):
 *         if out[i] < 0:             # <<<<<<<<<<<<<<
 *             out[i] = 0
 * 
 */
    }
  }

  /* "thinc/neural/_funcs.pyx":28
 * 
 * 
 * cdef void ReLu(weight_t* out, len_t nr_out) nogil:             # <<<<<<<<<<<<<<
 *     cdef idx_t i
 *     for i in range(nr_out):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.ReLu", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":35
 * 
 * 
 * cdef void softmax(weight_t* out, len_t nr_out) nogil:             # <<<<<<<<<<<<<<
 *     #w = exp(w - max(w))
 *     Vec.add_i(out,
 */

static void __pyx_f_5thinc_6neural_6_funcs_softmax(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_out, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_out) {
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_norm;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  __Pyx_TraceCall("softmax", __pyx_f[0], 35, 1, __PYX_ERR(0, 35, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":37
 * cdef void softmax(weight_t* out, len_t nr_out) nogil:
 *     #w = exp(w - max(w))
 *     Vec.add_i(out,             # <<<<<<<<<<<<<<
 *         -Vec.max(out, nr_out), nr_out)
 *     Vec.exp_i(out,
 */
  __pyx_f_5thinc_6linalg_3Vec_add_i(__pyx_v_out, (-__pyx_f_5thinc_6linalg_3Vec_max(__pyx_v_out, __pyx_v_nr_out)), __pyx_v_nr_out);

  /* "thinc/neural/_funcs.pyx":39
 *     Vec.add_i(out,
 *         -Vec.max(out, nr_out), nr_out)
 *     Vec.exp_i(out,             # <<<<<<<<<<<<<<
 *         nr_out)
 *     #w = w / sum(w)
 */
  __pyx_f_5thinc_6linalg_3Vec_exp_i(__pyx_v_out, __pyx_v_nr_out);

  /* "thinc/neural/_funcs.pyx":42
 *         nr_out)
 *     #w = w / sum(w)
 *     cdef weight_t norm = Vec.sum(out, nr_out)             # <<<<<<<<<<<<<<
 *     if norm != 0:
 *         Vec.div_i(out,
 */
  __pyx_v_norm = __pyx_f_5thinc_6linalg_3Vec_sum(__pyx_v_out, __pyx_v_nr_out);

  /* "thinc/neural/_funcs.pyx":43
 *     #w = w / sum(w)
 *     cdef weight_t norm = Vec.sum(out, nr_out)
 *     if norm != 0:             # <<<<<<<<<<<<<<
 *         Vec.div_i(out,
 *             norm, nr_out)
 */
  __pyx_t_1 = ((__pyx_v_norm != 0.0) != 0);
  if (__pyx_t_1) {

    /* "thinc/neural/_funcs.pyx":44
 *     cdef weight_t norm = Vec.sum(out, nr_out)
 *     if norm != 0:
 *         Vec.div_i(out,             # <<<<<<<<<<<<<<
 *             norm, nr_out)
 * 
 */
    __pyx_f_5thinc_6linalg_3Vec_div_i(__pyx_v_out, __pyx_v_norm, __pyx_v_nr_out);

    /* "thinc/neural/_funcs.pyx":43
 *     #w = w / sum(w)
 *     cdef weight_t norm = Vec.sum(out, nr_out)
 *     if norm != 0:             # <<<<<<<<<<<<<<
 *         Vec.div_i(out,
 *             norm, nr_out)
 */
  }

  /* "thinc/neural/_funcs.pyx":35
 * 
 * 
 * cdef void softmax(weight_t* out, len_t nr_out) nogil:             # <<<<<<<<<<<<<<
 *     #w = exp(w - max(w))
 *     Vec.add_i(out,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.softmax", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":49
 * # Backward functions
 * 
 * cdef void d_ReLu(weight_t* delta, const weight_t* signal_out, int n) nogil:             # <<<<<<<<<<<<<<
 *     # Backprop the ReLu transformation
 *     for i in range(n):
 */

static void __pyx_f_5thinc_6neural_6_funcs_d_ReLu(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_delta, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_signal_out, int __pyx_v_n) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("d_ReLu", __pyx_f[0], 49, 1, __PYX_ERR(0, 49, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":51
 * cdef void d_ReLu(weight_t* delta, const weight_t* signal_out, int n) nogil:
 *     # Backprop the ReLu transformation
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if signal_out[i] <= 0:
 *             delta[i] = 0
 */
  __pyx_t_1 = __pyx_v_n;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":52
 *     # Backprop the ReLu transformation
 *     for i in range(n):
 *         if signal_out[i] <= 0:             # <<<<<<<<<<<<<<
 *             delta[i] = 0
 * 
 */
    __pyx_t_3 = (((__pyx_v_signal_out[__pyx_v_i]) <= 0.0) != 0);
    if (__pyx_t_3) {

      /* "thinc/neural/_funcs.pyx":53
 *     for i in range(n):
 *         if signal_out[i] <= 0:
 *             delta[i] = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_delta[__pyx_v_i]) = 0.0;

      /* "thinc/neural/_funcs.pyx":52
 *     # Backprop the ReLu transformation
 *     for i in range(n):
 *         if signal_out[i] <= 0:             # <<<<<<<<<<<<<<
 *             delta[i] = 0
 * 
 */
    }
  }

  /* "thinc/neural/_funcs.pyx":49
 * # Backward functions
 * 
 * cdef void d_ReLu(weight_t* delta, const weight_t* signal_out, int n) nogil:             # <<<<<<<<<<<<<<
 *     # Backprop the ReLu transformation
 *     for i in range(n):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.d_ReLu", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":56
 * 
 * 
 * cdef inline void d_ELU(weight_t* delta, const weight_t* signal_out, int n) nogil:             # <<<<<<<<<<<<<<
 *     # Backprop the ELU transformation
 *     # Note that this is over the function _output_, not the function
 */

static CYTHON_INLINE void __pyx_f_5thinc_6neural_6_funcs_d_ELU(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_delta, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_signal_out, int __pyx_v_n) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_TraceCall("d_ELU", __pyx_f[0], 56, 1, __PYX_ERR(0, 56, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":60
 *     # Note that this is over the function _output_, not the function
 *     # _input_!
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if signal_out[i] <= 0:
 *             delta[i] *= signal_out[i] + ALPHA
 */
  __pyx_t_1 = __pyx_v_n;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":61
 *     # _input_!
 *     for i in range(n):
 *         if signal_out[i] <= 0:             # <<<<<<<<<<<<<<
 *             delta[i] *= signal_out[i] + ALPHA
 * 
 */
    __pyx_t_3 = (((__pyx_v_signal_out[__pyx_v_i]) <= 0.0) != 0);
    if (__pyx_t_3) {

      /* "thinc/neural/_funcs.pyx":62
 *     for i in range(n):
 *         if signal_out[i] <= 0:
 *             delta[i] *= signal_out[i] + ALPHA             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_4 = __pyx_v_i;
      (__pyx_v_delta[__pyx_t_4]) = ((__pyx_v_delta[__pyx_t_4]) * ((__pyx_v_signal_out[__pyx_v_i]) + 1.0));

      /* "thinc/neural/_funcs.pyx":61
 *     # _input_!
 *     for i in range(n):
 *         if signal_out[i] <= 0:             # <<<<<<<<<<<<<<
 *             delta[i] *= signal_out[i] + ALPHA
 * 
 */
    }
  }

  /* "thinc/neural/_funcs.pyx":56
 * 
 * 
 * cdef inline void d_ELU(weight_t* delta, const weight_t* signal_out, int n) nogil:             # <<<<<<<<<<<<<<
 *     # Backprop the ELU transformation
 *     # Note that this is over the function _output_, not the function
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.d_ELU", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":65
 * 
 * 
 * cdef void d_log_loss(weight_t* loss,             # <<<<<<<<<<<<<<
 *     const weight_t* costs, const weight_t* scores, len_t nr_out
 * ) nogil:
 */

static void __pyx_f_5thinc_6neural_6_funcs_d_log_loss(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_loss, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_costs, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_scores, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_out) {
  __pyx_t_5thinc_8typedefs_idx_t __pyx_v_i;
  __pyx_t_5thinc_8typedefs_idx_t __pyx_v_best;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_score;
  __Pyx_TraceDeclarations
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_1;
  __pyx_t_5thinc_8typedefs_idx_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_TraceCall("d_log_loss", __pyx_f[0], 65, 1, __PYX_ERR(0, 65, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":70
 *     # If there's more than one gold class, appoint the top scoring one the best
 *     cdef idx_t i
 *     cdef idx_t best = 0             # <<<<<<<<<<<<<<
 *     cdef weight_t score = 0.0
 *     for i in range(nr_out):
 */
  __pyx_v_best = 0;

  /* "thinc/neural/_funcs.pyx":71
 *     cdef idx_t i
 *     cdef idx_t best = 0
 *     cdef weight_t score = 0.0             # <<<<<<<<<<<<<<
 *     for i in range(nr_out):
 *         loss[i] = scores[i]
 */
  __pyx_v_score = 0.0;

  /* "thinc/neural/_funcs.pyx":72
 *     cdef idx_t best = 0
 *     cdef weight_t score = 0.0
 *     for i in range(nr_out):             # <<<<<<<<<<<<<<
 *         loss[i] = scores[i]
 *         if scores[i] >= score and costs[i] == 0:
 */
  __pyx_t_1 = __pyx_v_nr_out;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":73
 *     cdef weight_t score = 0.0
 *     for i in range(nr_out):
 *         loss[i] = scores[i]             # <<<<<<<<<<<<<<
 *         if scores[i] >= score and costs[i] == 0:
 *             score = scores[i]
 */
    (__pyx_v_loss[__pyx_v_i]) = (__pyx_v_scores[__pyx_v_i]);

    /* "thinc/neural/_funcs.pyx":74
 *     for i in range(nr_out):
 *         loss[i] = scores[i]
 *         if scores[i] >= score and costs[i] == 0:             # <<<<<<<<<<<<<<
 *             score = scores[i]
 *             best = i
 */
    __pyx_t_4 = (((__pyx_v_scores[__pyx_v_i]) >= __pyx_v_score) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_costs[__pyx_v_i]) == 0.0) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_3) {

      /* "thinc/neural/_funcs.pyx":75
 *         loss[i] = scores[i]
 *         if scores[i] >= score and costs[i] == 0:
 *             score = scores[i]             # <<<<<<<<<<<<<<
 *             best = i
 *     loss[best] -= 1
 */
      __pyx_v_score = (__pyx_v_scores[__pyx_v_i]);

      /* "thinc/neural/_funcs.pyx":76
 *         if scores[i] >= score and costs[i] == 0:
 *             score = scores[i]
 *             best = i             # <<<<<<<<<<<<<<
 *     loss[best] -= 1
 * 
 */
      __pyx_v_best = __pyx_v_i;

      /* "thinc/neural/_funcs.pyx":74
 *     for i in range(nr_out):
 *         loss[i] = scores[i]
 *         if scores[i] >= score and costs[i] == 0:             # <<<<<<<<<<<<<<
 *             score = scores[i]
 *             best = i
 */
    }
  }

  /* "thinc/neural/_funcs.pyx":77
 *             score = scores[i]
 *             best = i
 *     loss[best] -= 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_best;
  (__pyx_v_loss[__pyx_t_2]) = ((__pyx_v_loss[__pyx_t_2]) - 1.0);

  /* "thinc/neural/_funcs.pyx":65
 * 
 * 
 * cdef void d_log_loss(weight_t* loss,             # <<<<<<<<<<<<<<
 *     const weight_t* costs, const weight_t* scores, len_t nr_out
 * ) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.d_log_loss", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":80
 * 
 * 
 * cdef void d_hinge_loss(weight_t* loss,             # <<<<<<<<<<<<<<
 *         const weight_t* costs, const weight_t* scores, len_t nr_out) nogil:
 *     cdef int best = -1
 */

static void __pyx_f_5thinc_6neural_6_funcs_d_hinge_loss(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_loss, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_costs, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_scores, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_out) {
  int __pyx_v_best;
  int __pyx_v_guess;
  __pyx_t_5thinc_8typedefs_len_t __pyx_v_i;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_margin;
  __Pyx_TraceDeclarations
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_1;
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_TraceCall("d_hinge_loss", __pyx_f[0], 80, 1, __PYX_ERR(0, 80, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":82
 * cdef void d_hinge_loss(weight_t* loss,
 *         const weight_t* costs, const weight_t* scores, len_t nr_out) nogil:
 *     cdef int best = -1             # <<<<<<<<<<<<<<
 *     cdef int guess = -1
 *     for i in range(nr_out):
 */
  __pyx_v_best = -1;

  /* "thinc/neural/_funcs.pyx":83
 *         const weight_t* costs, const weight_t* scores, len_t nr_out) nogil:
 *     cdef int best = -1
 *     cdef int guess = -1             # <<<<<<<<<<<<<<
 *     for i in range(nr_out):
 *         loss[i] = 0.0
 */
  __pyx_v_guess = -1;

  /* "thinc/neural/_funcs.pyx":84
 *     cdef int best = -1
 *     cdef int guess = -1
 *     for i in range(nr_out):             # <<<<<<<<<<<<<<
 *         loss[i] = 0.0
 *         if costs[i] == 0:
 */
  __pyx_t_1 = __pyx_v_nr_out;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":85
 *     cdef int guess = -1
 *     for i in range(nr_out):
 *         loss[i] = 0.0             # <<<<<<<<<<<<<<
 *         if costs[i] == 0:
 *             if best == -1 or scores[i] >= scores[best]:
 */
    (__pyx_v_loss[__pyx_v_i]) = 0.0;

    /* "thinc/neural/_funcs.pyx":86
 *     for i in range(nr_out):
 *         loss[i] = 0.0
 *         if costs[i] == 0:             # <<<<<<<<<<<<<<
 *             if best == -1 or scores[i] >= scores[best]:
 *                 best = i
 */
    __pyx_t_3 = (((__pyx_v_costs[__pyx_v_i]) == 0.0) != 0);
    if (__pyx_t_3) {

      /* "thinc/neural/_funcs.pyx":87
 *         loss[i] = 0.0
 *         if costs[i] == 0:
 *             if best == -1 or scores[i] >= scores[best]:             # <<<<<<<<<<<<<<
 *                 best = i
 *         elif costs[i] > 0:
 */
      __pyx_t_4 = ((__pyx_v_best == -1L) != 0);
      if (!__pyx_t_4) {
      } else {
        __pyx_t_3 = __pyx_t_4;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_4 = (((__pyx_v_scores[__pyx_v_i]) >= (__pyx_v_scores[__pyx_v_best])) != 0);
      __pyx_t_3 = __pyx_t_4;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_3) {

        /* "thinc/neural/_funcs.pyx":88
 *         if costs[i] == 0:
 *             if best == -1 or scores[i] >= scores[best]:
 *                 best = i             # <<<<<<<<<<<<<<
 *         elif costs[i] > 0:
 *             if guess == -1 or scores[i] >= scores[guess]:
 */
        __pyx_v_best = __pyx_v_i;

        /* "thinc/neural/_funcs.pyx":87
 *         loss[i] = 0.0
 *         if costs[i] == 0:
 *             if best == -1 or scores[i] >= scores[best]:             # <<<<<<<<<<<<<<
 *                 best = i
 *         elif costs[i] > 0:
 */
      }

      /* "thinc/neural/_funcs.pyx":86
 *     for i in range(nr_out):
 *         loss[i] = 0.0
 *         if costs[i] == 0:             # <<<<<<<<<<<<<<
 *             if best == -1 or scores[i] >= scores[best]:
 *                 best = i
 */
      goto __pyx_L5;
    }

    /* "thinc/neural/_funcs.pyx":89
 *             if best == -1 or scores[i] >= scores[best]:
 *                 best = i
 *         elif costs[i] > 0:             # <<<<<<<<<<<<<<
 *             if guess == -1 or scores[i] >= scores[guess]:
 *                 guess = i
 */
    __pyx_t_3 = (((__pyx_v_costs[__pyx_v_i]) > 0.0) != 0);
    if (__pyx_t_3) {

      /* "thinc/neural/_funcs.pyx":90
 *                 best = i
 *         elif costs[i] > 0:
 *             if guess == -1 or scores[i] >= scores[guess]:             # <<<<<<<<<<<<<<
 *                 guess = i
 *     margin = (scores[guess] - scores[best]) + 1
 */
      __pyx_t_4 = ((__pyx_v_guess == -1L) != 0);
      if (!__pyx_t_4) {
      } else {
        __pyx_t_3 = __pyx_t_4;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_4 = (((__pyx_v_scores[__pyx_v_i]) >= (__pyx_v_scores[__pyx_v_guess])) != 0);
      __pyx_t_3 = __pyx_t_4;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_3) {

        /* "thinc/neural/_funcs.pyx":91
 *         elif costs[i] > 0:
 *             if guess == -1 or scores[i] >= scores[guess]:
 *                 guess = i             # <<<<<<<<<<<<<<
 *     margin = (scores[guess] - scores[best]) + 1
 *     if margin > 0:
 */
        __pyx_v_guess = __pyx_v_i;

        /* "thinc/neural/_funcs.pyx":90
 *                 best = i
 *         elif costs[i] > 0:
 *             if guess == -1 or scores[i] >= scores[guess]:             # <<<<<<<<<<<<<<
 *                 guess = i
 *     margin = (scores[guess] - scores[best]) + 1
 */
      }

      /* "thinc/neural/_funcs.pyx":89
 *             if best == -1 or scores[i] >= scores[best]:
 *                 best = i
 *         elif costs[i] > 0:             # <<<<<<<<<<<<<<
 *             if guess == -1 or scores[i] >= scores[guess]:
 *                 guess = i
 */
    }
    __pyx_L5:;
  }

  /* "thinc/neural/_funcs.pyx":92
 *             if guess == -1 or scores[i] >= scores[guess]:
 *                 guess = i
 *     margin = (scores[guess] - scores[best]) + 1             # <<<<<<<<<<<<<<
 *     if margin > 0:
 *         loss[best] = -margin
 */
  __pyx_v_margin = (((__pyx_v_scores[__pyx_v_guess]) - (__pyx_v_scores[__pyx_v_best])) + 1.0);

  /* "thinc/neural/_funcs.pyx":93
 *                 guess = i
 *     margin = (scores[guess] - scores[best]) + 1
 *     if margin > 0:             # <<<<<<<<<<<<<<
 *         loss[best] = -margin
 *         loss[guess] = margin
 */
  __pyx_t_3 = ((__pyx_v_margin > 0.0) != 0);
  if (__pyx_t_3) {

    /* "thinc/neural/_funcs.pyx":94
 *     margin = (scores[guess] - scores[best]) + 1
 *     if margin > 0:
 *         loss[best] = -margin             # <<<<<<<<<<<<<<
 *         loss[guess] = margin
 * 
 */
    (__pyx_v_loss[__pyx_v_best]) = (-__pyx_v_margin);

    /* "thinc/neural/_funcs.pyx":95
 *     if margin > 0:
 *         loss[best] = -margin
 *         loss[guess] = margin             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_loss[__pyx_v_guess]) = __pyx_v_margin;

    /* "thinc/neural/_funcs.pyx":93
 *                 guess = i
 *     margin = (scores[guess] - scores[best]) + 1
 *     if margin > 0:             # <<<<<<<<<<<<<<
 *         loss[best] = -margin
 *         loss[guess] = margin
 */
  }

  /* "thinc/neural/_funcs.pyx":80
 * 
 * 
 * cdef void d_hinge_loss(weight_t* loss,             # <<<<<<<<<<<<<<
 *         const weight_t* costs, const weight_t* scores, len_t nr_out) nogil:
 *     cdef int best = -1
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.d_hinge_loss", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":99
 * 
 * @cython.boundscheck(False)
 * cdef void d_batchnorm(weight_t* diff, weight_t* est_mean, weight_t* est_var,             # <<<<<<<<<<<<<<
 *         const weight_t* _x, int nr_out, int nr_batch) nogil:
 *     # Simplification by Clement Thorey, here:
 */

static void __pyx_f_5thinc_6neural_6_funcs_d_batchnorm(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_diff, __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_est_mean, __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_est_var, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v__x, int __pyx_v_nr_out, int __pyx_v_nr_batch) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_sum_dy;
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_sum_dy_x_mu;
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_Ex;
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_Vx;
  long __pyx_v_i;
  int __pyx_v_j;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  long __pyx_t_7;
  __Pyx_TraceCall("d_batchnorm", __pyx_f[0], 99, 1, __PYX_ERR(0, 99, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":118
 *     #       * inv_var \
 *     #       * np.sum(dy * x_mu, axis=0))
 *     if nr_batch == 1:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_1 = ((__pyx_v_nr_batch == 1) != 0);
  if (__pyx_t_1) {

    /* "thinc/neural/_funcs.pyx":119
 *     #       * np.sum(dy * x_mu, axis=0))
 *     if nr_batch == 1:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     sum_dy = <weight_t*>calloc(nr_out, sizeof(weight_t))
 */
    goto __pyx_L0;

    /* "thinc/neural/_funcs.pyx":118
 *     #       * inv_var \
 *     #       * np.sum(dy * x_mu, axis=0))
 *     if nr_batch == 1:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "thinc/neural/_funcs.pyx":121
 *         return
 * 
 *     sum_dy = <weight_t*>calloc(nr_out, sizeof(weight_t))             # <<<<<<<<<<<<<<
 *     sum_dy_x_mu = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Ex = <weight_t*>calloc(nr_out, sizeof(weight_t))
 */
  __pyx_v_sum_dy = ((__pyx_t_5thinc_8typedefs_weight_t *)calloc(__pyx_v_nr_out, (sizeof(__pyx_t_5thinc_8typedefs_weight_t))));

  /* "thinc/neural/_funcs.pyx":122
 * 
 *     sum_dy = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     sum_dy_x_mu = <weight_t*>calloc(nr_out, sizeof(weight_t))             # <<<<<<<<<<<<<<
 *     Ex = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Vx = <weight_t*>calloc(nr_out, sizeof(weight_t))
 */
  __pyx_v_sum_dy_x_mu = ((__pyx_t_5thinc_8typedefs_weight_t *)calloc(__pyx_v_nr_out, (sizeof(__pyx_t_5thinc_8typedefs_weight_t))));

  /* "thinc/neural/_funcs.pyx":123
 *     sum_dy = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     sum_dy_x_mu = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Ex = <weight_t*>calloc(nr_out, sizeof(weight_t))             # <<<<<<<<<<<<<<
 *     Vx = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Mat.mean_row(Ex, _x, nr_batch, nr_out)
 */
  __pyx_v_Ex = ((__pyx_t_5thinc_8typedefs_weight_t *)calloc(__pyx_v_nr_out, (sizeof(__pyx_t_5thinc_8typedefs_weight_t))));

  /* "thinc/neural/_funcs.pyx":124
 *     sum_dy_x_mu = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Ex = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Vx = <weight_t*>calloc(nr_out, sizeof(weight_t))             # <<<<<<<<<<<<<<
 *     Mat.mean_row(Ex, _x, nr_batch, nr_out)
 *     Mat.var_row(Vx, _x, Ex, nr_batch, nr_out, 0.0)
 */
  __pyx_v_Vx = ((__pyx_t_5thinc_8typedefs_weight_t *)calloc(__pyx_v_nr_out, (sizeof(__pyx_t_5thinc_8typedefs_weight_t))));

  /* "thinc/neural/_funcs.pyx":125
 *     Ex = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Vx = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Mat.mean_row(Ex, _x, nr_batch, nr_out)             # <<<<<<<<<<<<<<
 *     Mat.var_row(Vx, _x, Ex, nr_batch, nr_out, 0.0)
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:
 */
  __pyx_f_5thinc_6linalg_3Mat_mean_row(__pyx_v_Ex, __pyx_v__x, __pyx_v_nr_batch, __pyx_v_nr_out);

  /* "thinc/neural/_funcs.pyx":126
 *     Vx = <weight_t*>calloc(nr_out, sizeof(weight_t))
 *     Mat.mean_row(Ex, _x, nr_batch, nr_out)
 *     Mat.var_row(Vx, _x, Ex, nr_batch, nr_out, 0.0)             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:
 *         for j in range(nr_out):
 */
  __pyx_f_5thinc_6linalg_3Mat_var_row(__pyx_v_Vx, __pyx_v__x, __pyx_v_Ex, __pyx_v_nr_batch, __pyx_v_nr_out, 0.0);

  /* "thinc/neural/_funcs.pyx":127
 *     Mat.mean_row(Ex, _x, nr_batch, nr_out)
 *     Mat.var_row(Vx, _x, Ex, nr_batch, nr_out, 0.0)
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:             # <<<<<<<<<<<<<<
 *         for j in range(nr_out):
 *             sum_dy[j] += diff[i+j]
 */
  __pyx_t_2 = (__pyx_v_nr_batch * __pyx_v_nr_out);
  __pyx_t_3 = __pyx_v_nr_out;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i+=__pyx_t_3) {

    /* "thinc/neural/_funcs.pyx":128
 *     Mat.var_row(Vx, _x, Ex, nr_batch, nr_out, 0.0)
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:
 *         for j in range(nr_out):             # <<<<<<<<<<<<<<
 *             sum_dy[j] += diff[i+j]
 *             sum_dy_x_mu[j] += diff[i+j] * (_x[i+j] - Ex[j])
 */
    __pyx_t_4 = __pyx_v_nr_out;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_j = __pyx_t_5;

      /* "thinc/neural/_funcs.pyx":129
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:
 *         for j in range(nr_out):
 *             sum_dy[j] += diff[i+j]             # <<<<<<<<<<<<<<
 *             sum_dy_x_mu[j] += diff[i+j] * (_x[i+j] - Ex[j])
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:
 */
      __pyx_t_6 = __pyx_v_j;
      (__pyx_v_sum_dy[__pyx_t_6]) = ((__pyx_v_sum_dy[__pyx_t_6]) + (__pyx_v_diff[(__pyx_v_i + __pyx_v_j)]));

      /* "thinc/neural/_funcs.pyx":130
 *         for j in range(nr_out):
 *             sum_dy[j] += diff[i+j]
 *             sum_dy_x_mu[j] += diff[i+j] * (_x[i+j] - Ex[j])             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:
 *         Vec.mul_i(&diff[i], nr_batch, nr_out)
 */
      __pyx_t_6 = __pyx_v_j;
      (__pyx_v_sum_dy_x_mu[__pyx_t_6]) = ((__pyx_v_sum_dy_x_mu[__pyx_t_6]) + ((__pyx_v_diff[(__pyx_v_i + __pyx_v_j)]) * ((__pyx_v__x[(__pyx_v_i + __pyx_v_j)]) - (__pyx_v_Ex[__pyx_v_j]))));
    }
  }

  /* "thinc/neural/_funcs.pyx":131
 *             sum_dy[j] += diff[i+j]
 *             sum_dy_x_mu[j] += diff[i+j] * (_x[i+j] - Ex[j])
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:             # <<<<<<<<<<<<<<
 *         Vec.mul_i(&diff[i], nr_batch, nr_out)
 *         VecVec.add_i(&diff[i], sum_dy, -1., nr_out)
 */
  __pyx_t_3 = (__pyx_v_nr_batch * __pyx_v_nr_out);
  __pyx_t_2 = __pyx_v_nr_out;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_3; __pyx_v_i+=__pyx_t_2) {

    /* "thinc/neural/_funcs.pyx":132
 *             sum_dy_x_mu[j] += diff[i+j] * (_x[i+j] - Ex[j])
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:
 *         Vec.mul_i(&diff[i], nr_batch, nr_out)             # <<<<<<<<<<<<<<
 *         VecVec.add_i(&diff[i], sum_dy, -1., nr_out)
 *         for j in range(nr_out):
 */
    __pyx_f_5thinc_6linalg_3Vec_mul_i((&(__pyx_v_diff[__pyx_v_i])), __pyx_v_nr_batch, __pyx_v_nr_out);

    /* "thinc/neural/_funcs.pyx":133
 *     for i from 0 <= i < (nr_batch * nr_out) by nr_out:
 *         Vec.mul_i(&diff[i], nr_batch, nr_out)
 *         VecVec.add_i(&diff[i], sum_dy, -1., nr_out)             # <<<<<<<<<<<<<<
 *         for j in range(nr_out):
 *             diff[i+j] -= (_x[i+j] - Ex[j]) * Vx[j] ** -1. * sum_dy_x_mu[j]
 */
    __pyx_f_5thinc_6linalg_6VecVec_add_i((&(__pyx_v_diff[__pyx_v_i])), __pyx_v_sum_dy, -1., __pyx_v_nr_out);

    /* "thinc/neural/_funcs.pyx":134
 *         Vec.mul_i(&diff[i], nr_batch, nr_out)
 *         VecVec.add_i(&diff[i], sum_dy, -1., nr_out)
 *         for j in range(nr_out):             # <<<<<<<<<<<<<<
 *             diff[i+j] -= (_x[i+j] - Ex[j]) * Vx[j] ** -1. * sum_dy_x_mu[j]
 *     Vec.mul_i(diff, 1. / nr_batch, nr_batch * nr_out)
 */
    __pyx_t_4 = __pyx_v_nr_out;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_j = __pyx_t_5;

      /* "thinc/neural/_funcs.pyx":135
 *         VecVec.add_i(&diff[i], sum_dy, -1., nr_out)
 *         for j in range(nr_out):
 *             diff[i+j] -= (_x[i+j] - Ex[j]) * Vx[j] ** -1. * sum_dy_x_mu[j]             # <<<<<<<<<<<<<<
 *     Vec.mul_i(diff, 1. / nr_batch, nr_batch * nr_out)
 *     for i in range(nr_out):
 */
      __pyx_t_7 = (__pyx_v_i + __pyx_v_j);
      (__pyx_v_diff[__pyx_t_7]) = ((__pyx_v_diff[__pyx_t_7]) - ((((__pyx_v__x[(__pyx_v_i + __pyx_v_j)]) - (__pyx_v_Ex[__pyx_v_j])) * pow(((double)(__pyx_v_Vx[__pyx_v_j])), -1.)) * (__pyx_v_sum_dy_x_mu[__pyx_v_j])));
    }
  }

  /* "thinc/neural/_funcs.pyx":136
 *         for j in range(nr_out):
 *             diff[i+j] -= (_x[i+j] - Ex[j]) * Vx[j] ** -1. * sum_dy_x_mu[j]
 *     Vec.mul_i(diff, 1. / nr_batch, nr_batch * nr_out)             # <<<<<<<<<<<<<<
 *     for i in range(nr_out):
 *         est_mean[i] = (0.9 * est_mean[i]) + (0.1 * Ex[i])
 */
  __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_diff, (1. / __pyx_v_nr_batch), (__pyx_v_nr_batch * __pyx_v_nr_out));

  /* "thinc/neural/_funcs.pyx":137
 *             diff[i+j] -= (_x[i+j] - Ex[j]) * Vx[j] ** -1. * sum_dy_x_mu[j]
 *     Vec.mul_i(diff, 1. / nr_batch, nr_batch * nr_out)
 *     for i in range(nr_out):             # <<<<<<<<<<<<<<
 *         est_mean[i] = (0.9 * est_mean[i]) + (0.1 * Ex[i])
 *         est_var[i] = (0.9 * est_var[i]) + (0.1 * Vx[i])
 */
  __pyx_t_2 = __pyx_v_nr_out;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_2; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "thinc/neural/_funcs.pyx":138
 *     Vec.mul_i(diff, 1. / nr_batch, nr_batch * nr_out)
 *     for i in range(nr_out):
 *         est_mean[i] = (0.9 * est_mean[i]) + (0.1 * Ex[i])             # <<<<<<<<<<<<<<
 *         est_var[i] = (0.9 * est_var[i]) + (0.1 * Vx[i])
 *     free(Ex)
 */
    (__pyx_v_est_mean[__pyx_v_i]) = ((0.9 * (__pyx_v_est_mean[__pyx_v_i])) + (0.1 * (__pyx_v_Ex[__pyx_v_i])));

    /* "thinc/neural/_funcs.pyx":139
 *     for i in range(nr_out):
 *         est_mean[i] = (0.9 * est_mean[i]) + (0.1 * Ex[i])
 *         est_var[i] = (0.9 * est_var[i]) + (0.1 * Vx[i])             # <<<<<<<<<<<<<<
 *     free(Ex)
 *     free(Vx)
 */
    (__pyx_v_est_var[__pyx_v_i]) = ((0.9 * (__pyx_v_est_var[__pyx_v_i])) + (0.1 * (__pyx_v_Vx[__pyx_v_i])));
  }

  /* "thinc/neural/_funcs.pyx":140
 *         est_mean[i] = (0.9 * est_mean[i]) + (0.1 * Ex[i])
 *         est_var[i] = (0.9 * est_var[i]) + (0.1 * Vx[i])
 *     free(Ex)             # <<<<<<<<<<<<<<
 *     free(Vx)
 *     free(sum_dy)
 */
  free(__pyx_v_Ex);

  /* "thinc/neural/_funcs.pyx":141
 *         est_var[i] = (0.9 * est_var[i]) + (0.1 * Vx[i])
 *     free(Ex)
 *     free(Vx)             # <<<<<<<<<<<<<<
 *     free(sum_dy)
 *     free(sum_dy_x_mu)
 */
  free(__pyx_v_Vx);

  /* "thinc/neural/_funcs.pyx":142
 *     free(Ex)
 *     free(Vx)
 *     free(sum_dy)             # <<<<<<<<<<<<<<
 *     free(sum_dy_x_mu)
 * 
 */
  free(__pyx_v_sum_dy);

  /* "thinc/neural/_funcs.pyx":143
 *     free(Vx)
 *     free(sum_dy)
 *     free(sum_dy_x_mu)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_sum_dy_x_mu);

  /* "thinc/neural/_funcs.pyx":99
 * 
 * @cython.boundscheck(False)
 * cdef void d_batchnorm(weight_t* diff, weight_t* est_mean, weight_t* est_var,             # <<<<<<<<<<<<<<
 *         const weight_t* _x, int nr_out, int nr_batch) nogil:
 *     # Simplification by Clement Thorey, here:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.d_batchnorm", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":149
 * 
 * @cython.cdivision(True)
 * cdef void clip_gradient(weight_t* gradient, weight_t threshold, int nr_weight) nogil:             # <<<<<<<<<<<<<<
 *     # Clip gradient
 *     grad_norm = Vec.norm(gradient, nr_weight)
 */

static void __pyx_f_5thinc_6neural_6_funcs_clip_gradient(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_gradient, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_threshold, int __pyx_v_nr_weight) {
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_grad_norm;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  __Pyx_TraceCall("clip_gradient", __pyx_f[0], 149, 1, __PYX_ERR(0, 149, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":151
 * cdef void clip_gradient(weight_t* gradient, weight_t threshold, int nr_weight) nogil:
 *     # Clip gradient
 *     grad_norm = Vec.norm(gradient, nr_weight)             # <<<<<<<<<<<<<<
 *     if grad_norm >= threshold:
 *         Vec.mul_i(gradient, threshold / grad_norm, nr_weight)
 */
  __pyx_v_grad_norm = __pyx_f_5thinc_6linalg_3Vec_norm(__pyx_v_gradient, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":152
 *     # Clip gradient
 *     grad_norm = Vec.norm(gradient, nr_weight)
 *     if grad_norm >= threshold:             # <<<<<<<<<<<<<<
 *         Vec.mul_i(gradient, threshold / grad_norm, nr_weight)
 * 
 */
  __pyx_t_1 = ((__pyx_v_grad_norm >= __pyx_v_threshold) != 0);
  if (__pyx_t_1) {

    /* "thinc/neural/_funcs.pyx":153
 *     grad_norm = Vec.norm(gradient, nr_weight)
 *     if grad_norm >= threshold:
 *         Vec.mul_i(gradient, threshold / grad_norm, nr_weight)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_gradient, (__pyx_v_threshold / __pyx_v_grad_norm), __pyx_v_nr_weight);

    /* "thinc/neural/_funcs.pyx":152
 *     # Clip gradient
 *     grad_norm = Vec.norm(gradient, nr_weight)
 *     if grad_norm >= threshold:             # <<<<<<<<<<<<<<
 *         Vec.mul_i(gradient, threshold / grad_norm, nr_weight)
 * 
 */
  }

  /* "thinc/neural/_funcs.pyx":149
 * 
 * @cython.cdivision(True)
 * cdef void clip_gradient(weight_t* gradient, weight_t threshold, int nr_weight) nogil:             # <<<<<<<<<<<<<<
 *     # Clip gradient
 *     grad_norm = Vec.norm(gradient, nr_weight)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.clip_gradient", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":157
 * 
 * @cython.cdivision(True)
 * cdef void update_averages(weight_t* ema,             # <<<<<<<<<<<<<<
 *         const weight_t* weights, int nr_weight, weight_t t) nogil:
 *     cdef weight_t decay = (1.0 + t) / (10.0 + t)
 */

static void __pyx_f_5thinc_6neural_6_funcs_update_averages(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_ema, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_weights, int __pyx_v_nr_weight, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_t) {
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_decay;
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_TraceCall("update_averages", __pyx_f[0], 157, 1, __PYX_ERR(0, 157, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":159
 * cdef void update_averages(weight_t* ema,
 *         const weight_t* weights, int nr_weight, weight_t t) nogil:
 *     cdef weight_t decay = (1.0 + t) / (10.0 + t)             # <<<<<<<<<<<<<<
 *     if decay > 0.9999:
 *         decay = 0.9999
 */
  __pyx_v_decay = ((1.0 + __pyx_v_t) / (10.0 + __pyx_v_t));

  /* "thinc/neural/_funcs.pyx":160
 *         const weight_t* weights, int nr_weight, weight_t t) nogil:
 *     cdef weight_t decay = (1.0 + t) / (10.0 + t)
 *     if decay > 0.9999:             # <<<<<<<<<<<<<<
 *         decay = 0.9999
 *     for i in range(nr_weight):
 */
  __pyx_t_1 = ((__pyx_v_decay > 0.9999) != 0);
  if (__pyx_t_1) {

    /* "thinc/neural/_funcs.pyx":161
 *     cdef weight_t decay = (1.0 + t) / (10.0 + t)
 *     if decay > 0.9999:
 *         decay = 0.9999             # <<<<<<<<<<<<<<
 *     for i in range(nr_weight):
 *         ema[i] -= (1-decay) * (ema[i] - weights[i])
 */
    __pyx_v_decay = 0.9999;

    /* "thinc/neural/_funcs.pyx":160
 *         const weight_t* weights, int nr_weight, weight_t t) nogil:
 *     cdef weight_t decay = (1.0 + t) / (10.0 + t)
 *     if decay > 0.9999:             # <<<<<<<<<<<<<<
 *         decay = 0.9999
 *     for i in range(nr_weight):
 */
  }

  /* "thinc/neural/_funcs.pyx":162
 *     if decay > 0.9999:
 *         decay = 0.9999
 *     for i in range(nr_weight):             # <<<<<<<<<<<<<<
 *         ema[i] -= (1-decay) * (ema[i] - weights[i])
 * 
 */
  __pyx_t_2 = __pyx_v_nr_weight;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "thinc/neural/_funcs.pyx":163
 *         decay = 0.9999
 *     for i in range(nr_weight):
 *         ema[i] -= (1-decay) * (ema[i] - weights[i])             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_4 = __pyx_v_i;
    (__pyx_v_ema[__pyx_t_4]) = ((__pyx_v_ema[__pyx_t_4]) - ((1.0 - __pyx_v_decay) * ((__pyx_v_ema[__pyx_v_i]) - (__pyx_v_weights[__pyx_v_i]))));
  }

  /* "thinc/neural/_funcs.pyx":157
 * 
 * @cython.cdivision(True)
 * cdef void update_averages(weight_t* ema,             # <<<<<<<<<<<<<<
 *         const weight_t* weights, int nr_weight, weight_t t) nogil:
 *     cdef weight_t decay = (1.0 + t) / (10.0 + t)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.update_averages", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":167
 * 
 * @cython.cdivision(True)
 * cdef void vanilla_sgd(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,
 */

static void __pyx_f_5thinc_6neural_6_funcs_vanilla_sgd(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_weights, __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_gradient, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_weight, struct __pyx_t_5thinc_7structs_ConstantsC const *__pyx_v_hp) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("vanilla_sgd", __pyx_f[0], 167, 1, __PYX_ERR(0, 167, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":169
 * cdef void vanilla_sgd(weight_t* weights, weight_t* gradient,
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,             # <<<<<<<<<<<<<<
 *         100.0, nr_weight)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_clip_gradient(__pyx_v_gradient, 100.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":172
 *         100.0, nr_weight)
 * 
 *     VecVec.add_i(weights,             # <<<<<<<<<<<<<<
 *         gradient, -hp.e, nr_weight)
 *     memset(gradient,
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_weights, __pyx_v_gradient, (-__pyx_v_hp->e), __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":174
 *     VecVec.add_i(weights,
 *         gradient, -hp.e, nr_weight)
 *     memset(gradient,             # <<<<<<<<<<<<<<
 *         0, sizeof(gradient[0]) * nr_weight)
 * 
 */
  memset(__pyx_v_gradient, 0, ((sizeof((__pyx_v_gradient[0]))) * __pyx_v_nr_weight));

  /* "thinc/neural/_funcs.pyx":177
 *         0, sizeof(gradient[0]) * nr_weight)
 * 
 *     update_averages(weights+nr_weight,             # <<<<<<<<<<<<<<
 *         weights, nr_weight, hp.t)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_update_averages((__pyx_v_weights + __pyx_v_nr_weight), __pyx_v_weights, __pyx_v_nr_weight, __pyx_v_hp->t);

  /* "thinc/neural/_funcs.pyx":167
 * 
 * @cython.cdivision(True)
 * cdef void vanilla_sgd(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.vanilla_sgd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":182
 * 
 * @cython.cdivision(True)
 * cdef void sgd_cm(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     '''
 */

static void __pyx_f_5thinc_6neural_6_funcs_sgd_cm(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_weights, __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_gradient, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_weight, struct __pyx_t_5thinc_7structs_ConstantsC const *__pyx_v_hp) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_momentum;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("sgd_cm", __pyx_f[0], 182, 1, __PYX_ERR(0, 182, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":187
 *     Update weights with SGD and classical momentum
 *     '''
 *     clip_gradient(gradient,             # <<<<<<<<<<<<<<
 *         100.0, nr_weight)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_clip_gradient(__pyx_v_gradient, 100.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":190
 *         100.0, nr_weight)
 * 
 *     momentum = weights + nr_weight * 2             # <<<<<<<<<<<<<<
 *     Vec.mul_i(momentum, hp.m, nr_weight)
 *     VecVec.add_i(momentum,
 */
  __pyx_v_momentum = (__pyx_v_weights + (__pyx_v_nr_weight * 2));

  /* "thinc/neural/_funcs.pyx":191
 * 
 *     momentum = weights + nr_weight * 2
 *     Vec.mul_i(momentum, hp.m, nr_weight)             # <<<<<<<<<<<<<<
 *     VecVec.add_i(momentum,
 *         gradient, hp.e, nr_weight)
 */
  __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_momentum, __pyx_v_hp->m, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":192
 *     momentum = weights + nr_weight * 2
 *     Vec.mul_i(momentum, hp.m, nr_weight)
 *     VecVec.add_i(momentum,             # <<<<<<<<<<<<<<
 *         gradient, hp.e, nr_weight)
 *     VecVec.add_i(weights,
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_momentum, __pyx_v_gradient, __pyx_v_hp->e, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":194
 *     VecVec.add_i(momentum,
 *         gradient, hp.e, nr_weight)
 *     VecVec.add_i(weights,             # <<<<<<<<<<<<<<
 *         momentum, -1.0, nr_weight)
 * 
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_weights, __pyx_v_momentum, -1.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":197
 *         momentum, -1.0, nr_weight)
 * 
 *     memset(gradient,             # <<<<<<<<<<<<<<
 *         0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,
 */
  memset(__pyx_v_gradient, 0, ((sizeof((__pyx_v_gradient[0]))) * __pyx_v_nr_weight));

  /* "thinc/neural/_funcs.pyx":199
 *     memset(gradient,
 *         0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,             # <<<<<<<<<<<<<<
 *         weights, nr_weight, hp.t)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_update_averages((__pyx_v_weights + __pyx_v_nr_weight), __pyx_v_weights, __pyx_v_nr_weight, __pyx_v_hp->t);

  /* "thinc/neural/_funcs.pyx":182
 * 
 * @cython.cdivision(True)
 * cdef void sgd_cm(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     '''
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.sgd_cm", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":205
 * 
 * @cython.cdivision(True)
 * cdef void nag(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     '''
 */

static void __pyx_f_5thinc_6neural_6_funcs_nag(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_weights, __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_gradient, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_weight, struct __pyx_t_5thinc_7structs_ConstantsC const *__pyx_v_hp) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_momentum;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("nag", __pyx_f[0], 205, 1, __PYX_ERR(0, 205, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":210
 *     Update weights with SGD and Nesterov momentum
 *     '''
 *     clip_gradient(gradient,             # <<<<<<<<<<<<<<
 *         100.0, nr_weight)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_clip_gradient(__pyx_v_gradient, 100.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":213
 *         100.0, nr_weight)
 * 
 *     momentum = weights + nr_weight * 2             # <<<<<<<<<<<<<<
 *     # http://cs231n.github.io/neural-networks-3/
 *     # v_prev = v # back this up
 */
  __pyx_v_momentum = (__pyx_v_weights + (__pyx_v_nr_weight * 2));

  /* "thinc/neural/_funcs.pyx":223
 *     # v -= lr * gradient
 *     # x += (1+mu) * v
 *     VecVec.add_i(weights,             # <<<<<<<<<<<<<<
 *         momentum, -hp.m, nr_weight)
 *     Vec.mul_i(momentum,
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_weights, __pyx_v_momentum, (-__pyx_v_hp->m), __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":225
 *     VecVec.add_i(weights,
 *         momentum, -hp.m, nr_weight)
 *     Vec.mul_i(momentum,             # <<<<<<<<<<<<<<
 *         hp.m, nr_weight)
 *     VecVec.add_i(momentum,
 */
  __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_momentum, __pyx_v_hp->m, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":227
 *     Vec.mul_i(momentum,
 *         hp.m, nr_weight)
 *     VecVec.add_i(momentum,             # <<<<<<<<<<<<<<
 *         gradient, -hp.e, nr_weight)
 *     VecVec.add_i(weights,
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_momentum, __pyx_v_gradient, (-__pyx_v_hp->e), __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":229
 *     VecVec.add_i(momentum,
 *         gradient, -hp.e, nr_weight)
 *     VecVec.add_i(weights,             # <<<<<<<<<<<<<<
 *         momentum, 1+hp.m, nr_weight)
 * 
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_weights, __pyx_v_momentum, (1.0 + __pyx_v_hp->m), __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":232
 *         momentum, 1+hp.m, nr_weight)
 * 
 *     memset(gradient,             # <<<<<<<<<<<<<<
 *         0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,
 */
  memset(__pyx_v_gradient, 0, ((sizeof((__pyx_v_gradient[0]))) * __pyx_v_nr_weight));

  /* "thinc/neural/_funcs.pyx":234
 *     memset(gradient,
 *         0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,             # <<<<<<<<<<<<<<
 *         weights, nr_weight, hp.t)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_update_averages((__pyx_v_weights + __pyx_v_nr_weight), __pyx_v_weights, __pyx_v_nr_weight, __pyx_v_hp->t);

  /* "thinc/neural/_funcs.pyx":205
 * 
 * @cython.cdivision(True)
 * cdef void nag(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     '''
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.nag", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":239
 * 
 * @cython.cdivision(True)
 * cdef void adam(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,
 */

static void __pyx_f_5thinc_6neural_6_funcs_adam(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_weights, __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_gradient, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_weight, struct __pyx_t_5thinc_7structs_ConstantsC const *__pyx_v_hp) {
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_beta1;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_beta2;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_eps;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_learn_rate;
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_mom1;
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_mom2;
  __pyx_t_5thinc_8typedefs_len_t __pyx_v_i;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_a_t;
  __Pyx_TraceDeclarations
  __pyx_t_5thinc_8typedefs_weight_t __pyx_t_1;
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_2;
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_3;
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_4;
  __Pyx_TraceCall("adam", __pyx_f[0], 239, 1, __PYX_ERR(0, 239, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":241
 * cdef void adam(weight_t* weights, weight_t* gradient,
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,             # <<<<<<<<<<<<<<
 *         100.0, nr_weight)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_clip_gradient(__pyx_v_gradient, 100.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":244
 *         100.0, nr_weight)
 * 
 *     cdef weight_t beta1 = 0.90             # <<<<<<<<<<<<<<
 *     cdef weight_t beta2 = 0.999
 *     cdef weight_t eps = 1e-08
 */
  __pyx_v_beta1 = 0.90;

  /* "thinc/neural/_funcs.pyx":245
 * 
 *     cdef weight_t beta1 = 0.90
 *     cdef weight_t beta2 = 0.999             # <<<<<<<<<<<<<<
 *     cdef weight_t eps = 1e-08
 *     cdef weight_t learn_rate = hp.e
 */
  __pyx_v_beta2 = 0.999;

  /* "thinc/neural/_funcs.pyx":246
 *     cdef weight_t beta1 = 0.90
 *     cdef weight_t beta2 = 0.999
 *     cdef weight_t eps = 1e-08             # <<<<<<<<<<<<<<
 *     cdef weight_t learn_rate = hp.e
 *     mom1 = weights + nr_weight * 2
 */
  __pyx_v_eps = 1e-08;

  /* "thinc/neural/_funcs.pyx":247
 *     cdef weight_t beta2 = 0.999
 *     cdef weight_t eps = 1e-08
 *     cdef weight_t learn_rate = hp.e             # <<<<<<<<<<<<<<
 *     mom1 = weights + nr_weight * 2
 *     Vec.mul_i(mom1,
 */
  __pyx_t_1 = __pyx_v_hp->e;
  __pyx_v_learn_rate = __pyx_t_1;

  /* "thinc/neural/_funcs.pyx":248
 *     cdef weight_t eps = 1e-08
 *     cdef weight_t learn_rate = hp.e
 *     mom1 = weights + nr_weight * 2             # <<<<<<<<<<<<<<
 *     Vec.mul_i(mom1,
 *         beta1, nr_weight)
 */
  __pyx_v_mom1 = (__pyx_v_weights + (__pyx_v_nr_weight * 2));

  /* "thinc/neural/_funcs.pyx":249
 *     cdef weight_t learn_rate = hp.e
 *     mom1 = weights + nr_weight * 2
 *     Vec.mul_i(mom1,             # <<<<<<<<<<<<<<
 *         beta1, nr_weight)
 *     VecVec.add_i(mom1,
 */
  __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_mom1, __pyx_v_beta1, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":251
 *     Vec.mul_i(mom1,
 *         beta1, nr_weight)
 *     VecVec.add_i(mom1,             # <<<<<<<<<<<<<<
 *         gradient, 1-beta1, nr_weight)
 *     mom2 = weights + nr_weight * 3
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_mom1, __pyx_v_gradient, (1.0 - __pyx_v_beta1), __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":253
 *     VecVec.add_i(mom1,
 *         gradient, 1-beta1, nr_weight)
 *     mom2 = weights + nr_weight * 3             # <<<<<<<<<<<<<<
 *     for i in range(nr_weight):
 *         mom2[i] = (beta2 * mom2[i]) + ((1-beta2) * gradient[i] * gradient[i])
 */
  __pyx_v_mom2 = (__pyx_v_weights + (__pyx_v_nr_weight * 3));

  /* "thinc/neural/_funcs.pyx":254
 *         gradient, 1-beta1, nr_weight)
 *     mom2 = weights + nr_weight * 3
 *     for i in range(nr_weight):             # <<<<<<<<<<<<<<
 *         mom2[i] = (beta2 * mom2[i]) + ((1-beta2) * gradient[i] * gradient[i])
 *     # More efficient version, from the paper
 */
  __pyx_t_2 = __pyx_v_nr_weight;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "thinc/neural/_funcs.pyx":255
 *     mom2 = weights + nr_weight * 3
 *     for i in range(nr_weight):
 *         mom2[i] = (beta2 * mom2[i]) + ((1-beta2) * gradient[i] * gradient[i])             # <<<<<<<<<<<<<<
 *     # More efficient version, from the paper
 *     cdef weight_t a_t = learn_rate * sqrt(1-beta2**hp.t) / (1-beta1**hp.t)
 */
    (__pyx_v_mom2[__pyx_v_i]) = ((__pyx_v_beta2 * (__pyx_v_mom2[__pyx_v_i])) + (((1.0 - __pyx_v_beta2) * (__pyx_v_gradient[__pyx_v_i])) * (__pyx_v_gradient[__pyx_v_i])));
  }

  /* "thinc/neural/_funcs.pyx":257
 *         mom2[i] = (beta2 * mom2[i]) + ((1-beta2) * gradient[i] * gradient[i])
 *     # More efficient version, from the paper
 *     cdef weight_t a_t = learn_rate * sqrt(1-beta2**hp.t) / (1-beta1**hp.t)             # <<<<<<<<<<<<<<
 *     for i in range(nr_weight):
 *         weights[i] -= a_t * (mom1[i] / (sqrt(mom2[i]) + eps))
 */
  __pyx_v_a_t = ((__pyx_v_learn_rate * sqrt((1.0 - powf(__pyx_v_beta2, ((__pyx_t_5thinc_8typedefs_weight_t)__pyx_v_hp->t))))) / (1.0 - powf(__pyx_v_beta1, ((__pyx_t_5thinc_8typedefs_weight_t)__pyx_v_hp->t))));

  /* "thinc/neural/_funcs.pyx":258
 *     # More efficient version, from the paper
 *     cdef weight_t a_t = learn_rate * sqrt(1-beta2**hp.t) / (1-beta1**hp.t)
 *     for i in range(nr_weight):             # <<<<<<<<<<<<<<
 *         weights[i] -= a_t * (mom1[i] / (sqrt(mom2[i]) + eps))
 * 
 */
  __pyx_t_2 = __pyx_v_nr_weight;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "thinc/neural/_funcs.pyx":259
 *     cdef weight_t a_t = learn_rate * sqrt(1-beta2**hp.t) / (1-beta1**hp.t)
 *     for i in range(nr_weight):
 *         weights[i] -= a_t * (mom1[i] / (sqrt(mom2[i]) + eps))             # <<<<<<<<<<<<<<
 * 
 *     memset(gradient, 0, sizeof(gradient[0]) * nr_weight)
 */
    __pyx_t_4 = __pyx_v_i;
    (__pyx_v_weights[__pyx_t_4]) = ((__pyx_v_weights[__pyx_t_4]) - (__pyx_v_a_t * ((__pyx_v_mom1[__pyx_v_i]) / (sqrt((__pyx_v_mom2[__pyx_v_i])) + __pyx_v_eps))));
  }

  /* "thinc/neural/_funcs.pyx":261
 *         weights[i] -= a_t * (mom1[i] / (sqrt(mom2[i]) + eps))
 * 
 *     memset(gradient, 0, sizeof(gradient[0]) * nr_weight)             # <<<<<<<<<<<<<<
 *     update_averages(weights+nr_weight,
 *         weights, nr_weight, hp.t)
 */
  memset(__pyx_v_gradient, 0, ((sizeof((__pyx_v_gradient[0]))) * __pyx_v_nr_weight));

  /* "thinc/neural/_funcs.pyx":262
 * 
 *     memset(gradient, 0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,             # <<<<<<<<<<<<<<
 *         weights, nr_weight, hp.t)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_update_averages((__pyx_v_weights + __pyx_v_nr_weight), __pyx_v_weights, __pyx_v_nr_weight, __pyx_v_hp->t);

  /* "thinc/neural/_funcs.pyx":239
 * 
 * @cython.cdivision(True)
 * cdef void adam(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.adam", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":267
 * 
 * @cython.cdivision(True)
 * cdef void adagrad(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,
 */

static void __pyx_f_5thinc_6neural_6_funcs_adagrad(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_weights, __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_gradient, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_weight, struct __pyx_t_5thinc_7structs_ConstantsC const *__pyx_v_hp) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_momentum;
  __pyx_t_5thinc_8typedefs_len_t __pyx_v_i;
  __Pyx_TraceDeclarations
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_1;
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_2;
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_3;
  __Pyx_TraceCall("adagrad", __pyx_f[0], 267, 1, __PYX_ERR(0, 267, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":269
 * cdef void adagrad(weight_t* weights, weight_t* gradient,
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,             # <<<<<<<<<<<<<<
 *         100.0, nr_weight)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_clip_gradient(__pyx_v_gradient, 100.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":272
 *         100.0, nr_weight)
 * 
 *     momentum = weights + nr_weight * 2             # <<<<<<<<<<<<<<
 *     VecVec.add_pow_i(momentum,
 *         gradient, 2.0, nr_weight)
 */
  __pyx_v_momentum = (__pyx_v_weights + (__pyx_v_nr_weight * 2));

  /* "thinc/neural/_funcs.pyx":273
 * 
 *     momentum = weights + nr_weight * 2
 *     VecVec.add_pow_i(momentum,             # <<<<<<<<<<<<<<
 *         gradient, 2.0, nr_weight)
 *     for i in range(nr_weight):
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_pow_i(__pyx_v_momentum, __pyx_v_gradient, 2.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":275
 *     VecVec.add_pow_i(momentum,
 *         gradient, 2.0, nr_weight)
 *     for i in range(nr_weight):             # <<<<<<<<<<<<<<
 *         gradient[i] *= hp.e / (sqrt(momentum[i]) + EPS)
 *     # Make the (already scaled) update
 */
  __pyx_t_1 = __pyx_v_nr_weight;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":276
 *         gradient, 2.0, nr_weight)
 *     for i in range(nr_weight):
 *         gradient[i] *= hp.e / (sqrt(momentum[i]) + EPS)             # <<<<<<<<<<<<<<
 *     # Make the (already scaled) update
 *     VecVec.add_i(weights,
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_gradient[__pyx_t_3]) = ((__pyx_v_gradient[__pyx_t_3]) * (__pyx_v_hp->e / (sqrt((__pyx_v_momentum[__pyx_v_i])) + 1e-08)));
  }

  /* "thinc/neural/_funcs.pyx":278
 *         gradient[i] *= hp.e / (sqrt(momentum[i]) + EPS)
 *     # Make the (already scaled) update
 *     VecVec.add_i(weights,             # <<<<<<<<<<<<<<
 *         gradient, -1.0, nr_weight)
 * 
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_weights, __pyx_v_gradient, -1.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":281
 *         gradient, -1.0, nr_weight)
 * 
 *     memset(gradient,             # <<<<<<<<<<<<<<
 *         0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,
 */
  memset(__pyx_v_gradient, 0, ((sizeof((__pyx_v_gradient[0]))) * __pyx_v_nr_weight));

  /* "thinc/neural/_funcs.pyx":283
 *     memset(gradient,
 *         0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,             # <<<<<<<<<<<<<<
 *         weights, nr_weight, hp.t)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_update_averages((__pyx_v_weights + __pyx_v_nr_weight), __pyx_v_weights, __pyx_v_nr_weight, __pyx_v_hp->t);

  /* "thinc/neural/_funcs.pyx":267
 * 
 * @cython.cdivision(True)
 * cdef void adagrad(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.adagrad", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "thinc/neural/_funcs.pyx":288
 * 
 * @cython.cdivision(True)
 * cdef void adadelta(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,
 */

static void __pyx_f_5thinc_6neural_6_funcs_adadelta(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_weights, __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_gradient, __pyx_t_5thinc_8typedefs_len_t __pyx_v_nr_weight, struct __pyx_t_5thinc_7structs_ConstantsC const *__pyx_v_hp) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_avg;
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_v_step;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_alpha;
  __pyx_t_5thinc_8typedefs_len_t __pyx_v_i;
  __Pyx_TraceDeclarations
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_1;
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_2;
  __pyx_t_5thinc_8typedefs_len_t __pyx_t_3;
  __Pyx_TraceCall("adadelta", __pyx_f[0], 288, 1, __PYX_ERR(0, 288, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":290
 * cdef void adadelta(weight_t* weights, weight_t* gradient,
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,             # <<<<<<<<<<<<<<
 *         100.0, nr_weight)
 * 
 */
  __pyx_f_5thinc_6neural_6_funcs_clip_gradient(__pyx_v_gradient, 100.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":293
 *         100.0, nr_weight)
 * 
 *     avg = weights + nr_weight * 2             # <<<<<<<<<<<<<<
 *     step = weights + nr_weight * 3
 *     cdef weight_t alpha = 0.90
 */
  __pyx_v_avg = (__pyx_v_weights + (__pyx_v_nr_weight * 2));

  /* "thinc/neural/_funcs.pyx":294
 * 
 *     avg = weights + nr_weight * 2
 *     step = weights + nr_weight * 3             # <<<<<<<<<<<<<<
 *     cdef weight_t alpha = 0.90
 *     Vec.mul_i(avg,
 */
  __pyx_v_step = (__pyx_v_weights + (__pyx_v_nr_weight * 3));

  /* "thinc/neural/_funcs.pyx":295
 *     avg = weights + nr_weight * 2
 *     step = weights + nr_weight * 3
 *     cdef weight_t alpha = 0.90             # <<<<<<<<<<<<<<
 *     Vec.mul_i(avg,
 *         alpha, nr_weight)
 */
  __pyx_v_alpha = 0.90;

  /* "thinc/neural/_funcs.pyx":296
 *     step = weights + nr_weight * 3
 *     cdef weight_t alpha = 0.90
 *     Vec.mul_i(avg,             # <<<<<<<<<<<<<<
 *         alpha, nr_weight)
 *     for i in range(nr_weight):
 */
  __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_avg, __pyx_v_alpha, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":298
 *     Vec.mul_i(avg,
 *         alpha, nr_weight)
 *     for i in range(nr_weight):             # <<<<<<<<<<<<<<
 *         avg[i] += (1-alpha) * gradient[i] * gradient[i]
 *     for i in range(nr_weight):
 */
  __pyx_t_1 = __pyx_v_nr_weight;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":299
 *         alpha, nr_weight)
 *     for i in range(nr_weight):
 *         avg[i] += (1-alpha) * gradient[i] * gradient[i]             # <<<<<<<<<<<<<<
 *     for i in range(nr_weight):
 *         gradient[i] *= sqrt(step[i] + EPS) / sqrt(avg[i] + EPS)
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_avg[__pyx_t_3]) = ((__pyx_v_avg[__pyx_t_3]) + (((1.0 - __pyx_v_alpha) * (__pyx_v_gradient[__pyx_v_i])) * (__pyx_v_gradient[__pyx_v_i])));
  }

  /* "thinc/neural/_funcs.pyx":300
 *     for i in range(nr_weight):
 *         avg[i] += (1-alpha) * gradient[i] * gradient[i]
 *     for i in range(nr_weight):             # <<<<<<<<<<<<<<
 *         gradient[i] *= sqrt(step[i] + EPS) / sqrt(avg[i] + EPS)
 *     VecVec.add_i(weights,
 */
  __pyx_t_1 = __pyx_v_nr_weight;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "thinc/neural/_funcs.pyx":301
 *         avg[i] += (1-alpha) * gradient[i] * gradient[i]
 *     for i in range(nr_weight):
 *         gradient[i] *= sqrt(step[i] + EPS) / sqrt(avg[i] + EPS)             # <<<<<<<<<<<<<<
 *     VecVec.add_i(weights,
 *         gradient, -1.0, nr_weight)
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_gradient[__pyx_t_3]) = ((__pyx_v_gradient[__pyx_t_3]) * (sqrt(((__pyx_v_step[__pyx_v_i]) + 1e-08)) / sqrt(((__pyx_v_avg[__pyx_v_i]) + 1e-08))));
  }

  /* "thinc/neural/_funcs.pyx":302
 *     for i in range(nr_weight):
 *         gradient[i] *= sqrt(step[i] + EPS) / sqrt(avg[i] + EPS)
 *     VecVec.add_i(weights,             # <<<<<<<<<<<<<<
 *         gradient, -1.0, nr_weight)
 *     Vec.mul_i(step,
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_weights, __pyx_v_gradient, -1.0, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":304
 *     VecVec.add_i(weights,
 *         gradient, -1.0, nr_weight)
 *     Vec.mul_i(step,             # <<<<<<<<<<<<<<
 *         alpha, nr_weight)
 * 
 */
  __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_step, __pyx_v_alpha, __pyx_v_nr_weight);

  /* "thinc/neural/_funcs.pyx":307
 *         alpha, nr_weight)
 * 
 *     memset(gradient,             # <<<<<<<<<<<<<<
 *         0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,
 */
  memset(__pyx_v_gradient, 0, ((sizeof((__pyx_v_gradient[0]))) * __pyx_v_nr_weight));

  /* "thinc/neural/_funcs.pyx":309
 *     memset(gradient,
 *         0, sizeof(gradient[0]) * nr_weight)
 *     update_averages(weights+nr_weight,             # <<<<<<<<<<<<<<
 *         weights, nr_weight, hp.t)
 */
  __pyx_f_5thinc_6neural_6_funcs_update_averages((__pyx_v_weights + __pyx_v_nr_weight), __pyx_v_weights, __pyx_v_nr_weight, __pyx_v_hp->t);

  /* "thinc/neural/_funcs.pyx":288
 * 
 * @cython.cdivision(True)
 * cdef void adadelta(weight_t* weights, weight_t* gradient,             # <<<<<<<<<<<<<<
 *         len_t nr_weight, const ConstantsC* hp) nogil:
 *     clip_gradient(gradient,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.neural._funcs.adadelta", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":31
 * cdef class Vec:
 *     @staticmethod
 *     cdef inline int arg_max(const weight_t* scores, const int n_classes) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef int best = 0
 */

static CYTHON_INLINE int __pyx_f_5thinc_6linalg_3Vec_arg_max(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_scores, int const __pyx_v_n_classes) {
  int __pyx_v_i;
  int __pyx_v_best;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_mode;
  int __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("arg_max", __pyx_f[1], 31, 1, __PYX_ERR(1, 31, __pyx_L1_error));

  /* "linalg.pxd":33
 *     cdef inline int arg_max(const weight_t* scores, const int n_classes) nogil:
 *         cdef int i
 *         cdef int best = 0             # <<<<<<<<<<<<<<
 *         cdef weight_t mode = scores[0]
 *         for i in range(1, n_classes):
 */
  __pyx_v_best = 0;

  /* "linalg.pxd":34
 *         cdef int i
 *         cdef int best = 0
 *         cdef weight_t mode = scores[0]             # <<<<<<<<<<<<<<
 *         for i in range(1, n_classes):
 *             if scores[i] > mode:
 */
  __pyx_v_mode = (__pyx_v_scores[0]);

  /* "linalg.pxd":35
 *         cdef int best = 0
 *         cdef weight_t mode = scores[0]
 *         for i in range(1, n_classes):             # <<<<<<<<<<<<<<
 *             if scores[i] > mode:
 *                 mode = scores[i]
 */
  __pyx_t_1 = __pyx_v_n_classes;
  for (__pyx_t_2 = 1; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":36
 *         cdef weight_t mode = scores[0]
 *         for i in range(1, n_classes):
 *             if scores[i] > mode:             # <<<<<<<<<<<<<<
 *                 mode = scores[i]
 *                 best = i
 */
    __pyx_t_3 = (((__pyx_v_scores[__pyx_v_i]) > __pyx_v_mode) != 0);
    if (__pyx_t_3) {

      /* "linalg.pxd":37
 *         for i in range(1, n_classes):
 *             if scores[i] > mode:
 *                 mode = scores[i]             # <<<<<<<<<<<<<<
 *                 best = i
 *         return best
 */
      __pyx_v_mode = (__pyx_v_scores[__pyx_v_i]);

      /* "linalg.pxd":38
 *             if scores[i] > mode:
 *                 mode = scores[i]
 *                 best = i             # <<<<<<<<<<<<<<
 *         return best
 * 
 */
      __pyx_v_best = __pyx_v_i;

      /* "linalg.pxd":36
 *         cdef weight_t mode = scores[0]
 *         for i in range(1, n_classes):
 *             if scores[i] > mode:             # <<<<<<<<<<<<<<
 *                 mode = scores[i]
 *                 best = i
 */
    }
  }

  /* "linalg.pxd":39
 *                 mode = scores[i]
 *                 best = i
 *         return best             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = __pyx_v_best;
  goto __pyx_L0;

  /* "linalg.pxd":31
 * cdef class Vec:
 *     @staticmethod
 *     cdef inline int arg_max(const weight_t* scores, const int n_classes) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef int best = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.arg_max", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "linalg.pxd":42
 * 
 *     @staticmethod
 *     cdef inline weight_t max(const weight_t* x, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         if nr == 0:
 *             return 0
 */

static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_max(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_mode;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int32_t __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("max", __pyx_f[1], 42, 1, __PYX_ERR(1, 42, __pyx_L1_error));

  /* "linalg.pxd":43
 *     @staticmethod
 *     cdef inline weight_t max(const weight_t* x, int32_t nr) nogil:
 *         if nr == 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         cdef int i
 */
  __pyx_t_1 = ((__pyx_v_nr == 0) != 0);
  if (__pyx_t_1) {

    /* "linalg.pxd":44
 *     cdef inline weight_t max(const weight_t* x, int32_t nr) nogil:
 *         if nr == 0:
 *             return 0             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef weight_t mode = x[0]
 */
    __pyx_r = 0.0;
    goto __pyx_L0;

    /* "linalg.pxd":43
 *     @staticmethod
 *     cdef inline weight_t max(const weight_t* x, int32_t nr) nogil:
 *         if nr == 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         cdef int i
 */
  }

  /* "linalg.pxd":46
 *             return 0
 *         cdef int i
 *         cdef weight_t mode = x[0]             # <<<<<<<<<<<<<<
 *         for i in range(1, nr):
 *             if x[i] > mode:
 */
  __pyx_v_mode = (__pyx_v_x[0]);

  /* "linalg.pxd":47
 *         cdef int i
 *         cdef weight_t mode = x[0]
 *         for i in range(1, nr):             # <<<<<<<<<<<<<<
 *             if x[i] > mode:
 *                 mode = x[i]
 */
  __pyx_t_2 = __pyx_v_nr;
  for (__pyx_t_3 = 1; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "linalg.pxd":48
 *         cdef weight_t mode = x[0]
 *         for i in range(1, nr):
 *             if x[i] > mode:             # <<<<<<<<<<<<<<
 *                 mode = x[i]
 *         return mode
 */
    __pyx_t_1 = (((__pyx_v_x[__pyx_v_i]) > __pyx_v_mode) != 0);
    if (__pyx_t_1) {

      /* "linalg.pxd":49
 *         for i in range(1, nr):
 *             if x[i] > mode:
 *                 mode = x[i]             # <<<<<<<<<<<<<<
 *         return mode
 * 
 */
      __pyx_v_mode = (__pyx_v_x[__pyx_v_i]);

      /* "linalg.pxd":48
 *         cdef weight_t mode = x[0]
 *         for i in range(1, nr):
 *             if x[i] > mode:             # <<<<<<<<<<<<<<
 *                 mode = x[i]
 *         return mode
 */
    }
  }

  /* "linalg.pxd":50
 *             if x[i] > mode:
 *                 mode = x[i]
 *         return mode             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = __pyx_v_mode;
  goto __pyx_L0;

  /* "linalg.pxd":42
 * 
 *     @staticmethod
 *     cdef inline weight_t max(const weight_t* x, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         if nr == 0:
 *             return 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.max", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "linalg.pxd":53
 * 
 *     @staticmethod
 *     cdef inline weight_t sum(const weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef weight_t total = 0
 */

static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_sum(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_total;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_r;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("sum", __pyx_f[1], 53, 1, __PYX_ERR(1, 53, __pyx_L1_error));

  /* "linalg.pxd":55
 *     cdef inline weight_t sum(const weight_t* vec, int32_t nr) nogil:
 *         cdef int i
 *         cdef weight_t total = 0             # <<<<<<<<<<<<<<
 *         for i in range(nr):
 *             total += vec[i]
 */
  __pyx_v_total = 0.0;

  /* "linalg.pxd":56
 *         cdef int i
 *         cdef weight_t total = 0
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             total += vec[i]
 *         return total
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":57
 *         cdef weight_t total = 0
 *         for i in range(nr):
 *             total += vec[i]             # <<<<<<<<<<<<<<
 *         return total
 * 
 */
    __pyx_v_total = (__pyx_v_total + (__pyx_v_vec[__pyx_v_i]));
  }

  /* "linalg.pxd":58
 *         for i in range(nr):
 *             total += vec[i]
 *         return total             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = __pyx_v_total;
  goto __pyx_L0;

  /* "linalg.pxd":53
 * 
 *     @staticmethod
 *     cdef inline weight_t sum(const weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef weight_t total = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.sum", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "linalg.pxd":61
 * 
 *     @staticmethod
 *     cdef inline weight_t norm(const weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef weight_t total = 0
 *         for i in range(nr):
 */

static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_3Vec_norm(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr) {
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_total;
  int32_t __pyx_v_i;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_r;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int32_t __pyx_t_2;
  __Pyx_TraceCall("norm", __pyx_f[1], 61, 1, __PYX_ERR(1, 61, __pyx_L1_error));

  /* "linalg.pxd":62
 *     @staticmethod
 *     cdef inline weight_t norm(const weight_t* vec, int32_t nr) nogil:
 *         cdef weight_t total = 0             # <<<<<<<<<<<<<<
 *         for i in range(nr):
 *             total += vec[i] ** 2
 */
  __pyx_v_total = 0.0;

  /* "linalg.pxd":63
 *     cdef inline weight_t norm(const weight_t* vec, int32_t nr) nogil:
 *         cdef weight_t total = 0
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             total += vec[i] ** 2
 *         return sqrt(total)
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":64
 *         cdef weight_t total = 0
 *         for i in range(nr):
 *             total += vec[i] ** 2             # <<<<<<<<<<<<<<
 *         return sqrt(total)
 * 
 */
    __pyx_v_total = (__pyx_v_total + powf(((__pyx_t_5thinc_8typedefs_weight_t)(__pyx_v_vec[__pyx_v_i])), 2.0));
  }

  /* "linalg.pxd":65
 *         for i in range(nr):
 *             total += vec[i] ** 2
 *         return sqrt(total)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = sqrt(__pyx_v_total);
  goto __pyx_L0;

  /* "linalg.pxd":61
 * 
 *     @staticmethod
 *     cdef inline weight_t norm(const weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef weight_t total = 0
 *         for i in range(nr):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.norm", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "linalg.pxd":68
 * 
 *     @staticmethod
 *     cdef inline void add(weight_t* output, const weight_t* x,             # <<<<<<<<<<<<<<
 *             weight_t inc, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_add(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_inc, int32_t __pyx_v_nr) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("add", __pyx_f[1], 68, 1, __PYX_ERR(1, 68, __pyx_L1_error));

  /* "linalg.pxd":70
 *     cdef inline void add(weight_t* output, const weight_t* x,
 *             weight_t inc, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)             # <<<<<<<<<<<<<<
 *         Vec.add_i(output, inc, nr)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_x, ((sizeof((__pyx_v_output[0]))) * __pyx_v_nr));

  /* "linalg.pxd":71
 *             weight_t inc, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 *         Vec.add_i(output, inc, nr)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_3Vec_add_i(__pyx_v_output, __pyx_v_inc, __pyx_v_nr);

  /* "linalg.pxd":68
 * 
 *     @staticmethod
 *     cdef inline void add(weight_t* output, const weight_t* x,             # <<<<<<<<<<<<<<
 *             weight_t inc, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.add", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":74
 * 
 *     @staticmethod
 *     cdef inline void add_i(weight_t* vec, weight_t inc, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_inc, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("add_i", __pyx_f[1], 74, 1, __PYX_ERR(1, 74, __pyx_L1_error));

  /* "linalg.pxd":76
 *     cdef inline void add_i(weight_t* vec, weight_t inc, int32_t nr) nogil:
 *         cdef int i
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             vec[i] += inc
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":77
 *         cdef int i
 *         for i in range(nr):
 *             vec[i] += inc             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_vec[__pyx_t_3]) = ((__pyx_v_vec[__pyx_t_3]) + __pyx_v_inc);
  }

  /* "linalg.pxd":74
 * 
 *     @staticmethod
 *     cdef inline void add_i(weight_t* vec, weight_t inc, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.add_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":80
 * 
 *     @staticmethod
 *     cdef inline void mul(weight_t* output, const weight_t* vec, weight_t scal,             # <<<<<<<<<<<<<<
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_mul(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scal, int32_t __pyx_v_nr) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("mul", __pyx_f[1], 80, 1, __PYX_ERR(1, 80, __pyx_L1_error));

  /* "linalg.pxd":82
 *     cdef inline void mul(weight_t* output, const weight_t* vec, weight_t scal,
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)             # <<<<<<<<<<<<<<
 *         Vec.mul_i(output, scal, nr)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_vec, ((sizeof((__pyx_v_output[0]))) * __pyx_v_nr));

  /* "linalg.pxd":83
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 *         Vec.mul_i(output, scal, nr)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_output, __pyx_v_scal, __pyx_v_nr);

  /* "linalg.pxd":80
 * 
 *     @staticmethod
 *     cdef inline void mul(weight_t* output, const weight_t* vec, weight_t scal,             # <<<<<<<<<<<<<<
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.mul", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":86
 * 
 *     @staticmethod
 *     cdef inline void mul_i(weight_t* vec, weight_t scal, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         IF USE_BLAS:
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scal, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("mul_i", __pyx_f[1], 86, 1, __PYX_ERR(1, 86, __pyx_L1_error));

  /* "linalg.pxd":91
 *             blis.scalv(blis.NO_CONJUGATE, nr, scal, vec, 1)
 *         ELSE:
 *             for i in range(nr):             # <<<<<<<<<<<<<<
 *                 vec[i] *= scal
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":92
 *         ELSE:
 *             for i in range(nr):
 *                 vec[i] *= scal             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_vec[__pyx_t_3]) = ((__pyx_v_vec[__pyx_t_3]) * __pyx_v_scal);
  }

  /* "linalg.pxd":86
 * 
 *     @staticmethod
 *     cdef inline void mul_i(weight_t* vec, weight_t scal, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         IF USE_BLAS:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.mul_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":95
 * 
 *     @staticmethod
 *     cdef inline void pow(weight_t* output, const weight_t* vec, weight_t scal,             # <<<<<<<<<<<<<<
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_pow(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scal, int32_t __pyx_v_nr) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("pow", __pyx_f[1], 95, 1, __PYX_ERR(1, 95, __pyx_L1_error));

  /* "linalg.pxd":97
 *     cdef inline void pow(weight_t* output, const weight_t* vec, weight_t scal,
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)             # <<<<<<<<<<<<<<
 *         Vec.pow_i(output, scal, nr)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_vec, ((sizeof((__pyx_v_output[0]))) * __pyx_v_nr));

  /* "linalg.pxd":98
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 *         Vec.pow_i(output, scal, nr)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_3Vec_pow_i(__pyx_v_output, __pyx_v_scal, __pyx_v_nr);

  /* "linalg.pxd":95
 * 
 *     @staticmethod
 *     cdef inline void pow(weight_t* output, const weight_t* vec, weight_t scal,             # <<<<<<<<<<<<<<
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.pow", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":101
 * 
 *     @staticmethod
 *     cdef inline void pow_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_pow_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t const __pyx_v_scal, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("pow_i", __pyx_f[1], 101, 1, __PYX_ERR(1, 101, __pyx_L1_error));

  /* "linalg.pxd":103
 *     cdef inline void pow_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:
 *         cdef int i
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             vec[i] **= scal
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":104
 *         cdef int i
 *         for i in range(nr):
 *             vec[i] **= scal             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_vec[__pyx_t_3]) = powf((__pyx_v_vec[__pyx_t_3]), ((__pyx_t_5thinc_8typedefs_weight_t)__pyx_v_scal));
  }

  /* "linalg.pxd":101
 * 
 *     @staticmethod
 *     cdef inline void pow_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.pow_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":108
 *     @staticmethod
 *     @cython.cdivision(True)
 *     cdef inline void div(weight_t* output, const weight_t* vec, weight_t scal,             # <<<<<<<<<<<<<<
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_div(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scal, int32_t __pyx_v_nr) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("div", __pyx_f[1], 108, 1, __PYX_ERR(1, 108, __pyx_L1_error));

  /* "linalg.pxd":110
 *     cdef inline void div(weight_t* output, const weight_t* vec, weight_t scal,
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)             # <<<<<<<<<<<<<<
 *         Vec.div_i(output, scal, nr)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_vec, ((sizeof((__pyx_v_output[0]))) * __pyx_v_nr));

  /* "linalg.pxd":111
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 *         Vec.div_i(output, scal, nr)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_3Vec_div_i(__pyx_v_output, __pyx_v_scal, __pyx_v_nr);

  /* "linalg.pxd":108
 *     @staticmethod
 *     @cython.cdivision(True)
 *     cdef inline void div(weight_t* output, const weight_t* vec, weight_t scal,             # <<<<<<<<<<<<<<
 *             int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.div", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":115
 *     @staticmethod
 *     @cython.cdivision(True)
 *     cdef inline void div_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_div_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t const __pyx_v_scal, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("div_i", __pyx_f[1], 115, 1, __PYX_ERR(1, 115, __pyx_L1_error));

  /* "linalg.pxd":117
 *     cdef inline void div_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:
 *         cdef int i
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             vec[i] /= scal
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":118
 *         cdef int i
 *         for i in range(nr):
 *             vec[i] /= scal             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_vec[__pyx_t_3]) = ((__pyx_v_vec[__pyx_t_3]) / __pyx_v_scal);
  }

  /* "linalg.pxd":115
 *     @staticmethod
 *     @cython.cdivision(True)
 *     cdef inline void div_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.div_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":121
 * 
 *     @staticmethod
 *     cdef inline void exp(weight_t* output, const weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 *         Vec.exp_i(output, nr)
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_exp(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("exp", __pyx_f[1], 121, 1, __PYX_ERR(1, 121, __pyx_L1_error));

  /* "linalg.pxd":122
 *     @staticmethod
 *     cdef inline void exp(weight_t* output, const weight_t* vec, int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)             # <<<<<<<<<<<<<<
 *         Vec.exp_i(output, nr)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_vec, ((sizeof((__pyx_v_output[0]))) * __pyx_v_nr));

  /* "linalg.pxd":123
 *     cdef inline void exp(weight_t* output, const weight_t* vec, int32_t nr) nogil:
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 *         Vec.exp_i(output, nr)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_3Vec_exp_i(__pyx_v_output, __pyx_v_nr);

  /* "linalg.pxd":121
 * 
 *     @staticmethod
 *     cdef inline void exp(weight_t* output, const weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         memcpy(output, vec, sizeof(output[0]) * nr)
 *         Vec.exp_i(output, nr)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.exp", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":126
 * 
 *     @staticmethod
 *     cdef inline void exp_i(weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_exp_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("exp_i", __pyx_f[1], 126, 1, __PYX_ERR(1, 126, __pyx_L1_error));

  /* "linalg.pxd":128
 *     cdef inline void exp_i(weight_t* vec, int32_t nr) nogil:
 *         cdef int i
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             vec[i] = exp(vec[i])
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":129
 *         cdef int i
 *         for i in range(nr):
 *             vec[i] = exp(vec[i])             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    (__pyx_v_vec[__pyx_v_i]) = exp((__pyx_v_vec[__pyx_v_i]));
  }

  /* "linalg.pxd":126
 * 
 *     @staticmethod
 *     cdef inline void exp_i(weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.exp_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":132
 * 
 *     @staticmethod
 *     cdef inline void reciprocal_i(weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Vec_reciprocal_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_vec, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("reciprocal_i", __pyx_f[1], 132, 1, __PYX_ERR(1, 132, __pyx_L1_error));

  /* "linalg.pxd":134
 *     cdef inline void reciprocal_i(weight_t* vec, int32_t nr) nogil:
 *         cdef int i
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             vec[i] = 1.0 / vec[i]
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":135
 *         cdef int i
 *         for i in range(nr):
 *             vec[i] = 1.0 / vec[i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_vec[__pyx_v_i]) = (1.0 / (__pyx_v_vec[__pyx_v_i]));
  }

  /* "linalg.pxd":132
 * 
 *     @staticmethod
 *     cdef inline void reciprocal_i(weight_t* vec, int32_t nr) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(nr):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Vec.reciprocal_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":140
 * cdef class VecVec:
 *     @staticmethod
 *     cdef inline void add(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* x,
 *                          const weight_t* y,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scale, int32_t __pyx_v_nr) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("add", __pyx_f[1], 140, 1, __PYX_ERR(1, 140, __pyx_L1_error));

  /* "linalg.pxd":145
 *                          weight_t scale,
 *                          int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)             # <<<<<<<<<<<<<<
 *         VecVec.add_i(output, y, scale, nr)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_x, ((sizeof((__pyx_v_output[0]))) * __pyx_v_nr));

  /* "linalg.pxd":146
 *                          int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 *         VecVec.add_i(output, y, scale, nr)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_output, __pyx_v_y, __pyx_v_scale, __pyx_v_nr);

  /* "linalg.pxd":140
 * cdef class VecVec:
 *     @staticmethod
 *     cdef inline void add(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* x,
 *                          const weight_t* y,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.add", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":149
 * 
 *     @staticmethod
 *     cdef inline void add_i(weight_t* x,             # <<<<<<<<<<<<<<
 *                            const weight_t* y,
 *                            weight_t scale,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scale, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("add_i", __pyx_f[1], 149, 1, __PYX_ERR(1, 149, __pyx_L1_error));

  /* "linalg.pxd":157
 *             blis.axpyv(blis.NO_CONJUGATE, nr, scale, <weight_t*>y, 1, x, 1)
 *         ELSE:
 *             for i in range(nr):             # <<<<<<<<<<<<<<
 *                 x[i] += y[i] * scale
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":158
 *         ELSE:
 *             for i in range(nr):
 *                 x[i] += y[i] * scale             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_x[__pyx_t_3]) = ((__pyx_v_x[__pyx_t_3]) + ((__pyx_v_y[__pyx_v_i]) * __pyx_v_scale));
  }

  /* "linalg.pxd":149
 * 
 *     @staticmethod
 *     cdef inline void add_i(weight_t* x,             # <<<<<<<<<<<<<<
 *                            const weight_t* y,
 *                            weight_t scale,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.add_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":161
 * 
 *     @staticmethod
 *     cdef inline void batch_add_i(weight_t* x,             # <<<<<<<<<<<<<<
 *                            const weight_t* y,
 *                            weight_t scale,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_batch_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scale, int32_t __pyx_v_nr, int32_t __pyx_v_nr_batch) {
  CYTHON_UNUSED int __pyx_v__;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("batch_add_i", __pyx_f[1], 161, 1, __PYX_ERR(1, 161, __pyx_L1_error));

  /* "linalg.pxd":167
 *         # For fixed x, matrix of y
 *         cdef int i, _
 *         for _ in range(nr_batch):             # <<<<<<<<<<<<<<
 *             VecVec.add_i(x,
 *                 y, scale, nr)
 */
  __pyx_t_1 = __pyx_v_nr_batch;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v__ = __pyx_t_2;

    /* "linalg.pxd":168
 *         cdef int i, _
 *         for _ in range(nr_batch):
 *             VecVec.add_i(x,             # <<<<<<<<<<<<<<
 *                 y, scale, nr)
 *             y += nr
 */
    __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_x, __pyx_v_y, __pyx_v_scale, __pyx_v_nr);

    /* "linalg.pxd":170
 *             VecVec.add_i(x,
 *                 y, scale, nr)
 *             y += nr             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_v_y = (__pyx_v_y + __pyx_v_nr);
  }

  /* "linalg.pxd":161
 * 
 *     @staticmethod
 *     cdef inline void batch_add_i(weight_t* x,             # <<<<<<<<<<<<<<
 *                            const weight_t* y,
 *                            weight_t scale,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.batch_add_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":173
 * 
 *     @staticmethod
 *     cdef inline void add_pow(weight_t* output,             # <<<<<<<<<<<<<<
 *             const weight_t* x, const weight_t* y, weight_t power, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_pow(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_power, int32_t __pyx_v_nr) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("add_pow", __pyx_f[1], 173, 1, __PYX_ERR(1, 173, __pyx_L1_error));

  /* "linalg.pxd":175
 *     cdef inline void add_pow(weight_t* output,
 *             const weight_t* x, const weight_t* y, weight_t power, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)             # <<<<<<<<<<<<<<
 *         VecVec.add_pow_i(output, y, power, nr)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_x, ((sizeof((__pyx_v_output[0]))) * __pyx_v_nr));

  /* "linalg.pxd":176
 *             const weight_t* x, const weight_t* y, weight_t power, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 *         VecVec.add_pow_i(output, y, power, nr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_5thinc_6linalg_6VecVec_add_pow_i(__pyx_v_output, __pyx_v_y, __pyx_v_power, __pyx_v_nr);

  /* "linalg.pxd":173
 * 
 *     @staticmethod
 *     cdef inline void add_pow(weight_t* output,             # <<<<<<<<<<<<<<
 *             const weight_t* x, const weight_t* y, weight_t power, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.add_pow", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":180
 * 
 *     @staticmethod
 *     cdef inline void add_pow_i(weight_t* x,             # <<<<<<<<<<<<<<
 *             const weight_t* y, weight_t power, int32_t nr) nogil:
 *         cdef int i
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_add_pow_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_power, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("add_pow_i", __pyx_f[1], 180, 1, __PYX_ERR(1, 180, __pyx_L1_error));

  /* "linalg.pxd":183
 *             const weight_t* y, weight_t power, int32_t nr) nogil:
 *         cdef int i
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             x[i] += y[i] ** power
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":184
 *         cdef int i
 *         for i in range(nr):
 *             x[i] += y[i] ** power             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_x[__pyx_t_3]) = ((__pyx_v_x[__pyx_t_3]) + powf(((__pyx_t_5thinc_8typedefs_weight_t)(__pyx_v_y[__pyx_v_i])), __pyx_v_power));
  }

  /* "linalg.pxd":180
 * 
 *     @staticmethod
 *     cdef inline void add_pow_i(weight_t* x,             # <<<<<<<<<<<<<<
 *             const weight_t* y, weight_t power, int32_t nr) nogil:
 *         cdef int i
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.add_pow_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":187
 * 
 *     @staticmethod
 *     cdef inline void mul(weight_t* output,             # <<<<<<<<<<<<<<
 *             const weight_t* x, const weight_t* y, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_mul(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("mul", __pyx_f[1], 187, 1, __PYX_ERR(1, 187, __pyx_L1_error));

  /* "linalg.pxd":189
 *     cdef inline void mul(weight_t* output,
 *             const weight_t* x, const weight_t* y, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)             # <<<<<<<<<<<<<<
 *         VecVec.mul_i(output, y, nr)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_x, ((sizeof((__pyx_v_output[0]))) * __pyx_v_nr));

  /* "linalg.pxd":190
 *             const weight_t* x, const weight_t* y, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 *         VecVec.mul_i(output, y, nr)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_6VecVec_mul_i(__pyx_v_output, __pyx_v_y, __pyx_v_nr);

  /* "linalg.pxd":187
 * 
 *     @staticmethod
 *     cdef inline void mul(weight_t* output,             # <<<<<<<<<<<<<<
 *             const weight_t* x, const weight_t* y, int32_t nr) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.mul", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":193
 * 
 *     @staticmethod
 *     cdef inline void mul_i(weight_t* x,             # <<<<<<<<<<<<<<
 *             const weight_t* y, int32_t nr) nogil:
 *         cdef int i
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6VecVec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("mul_i", __pyx_f[1], 193, 1, __PYX_ERR(1, 193, __pyx_L1_error));

  /* "linalg.pxd":196
 *             const weight_t* y, int32_t nr) nogil:
 *         cdef int i
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             x[i] *= y[i]
 * 
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":197
 *         cdef int i
 *         for i in range(nr):
 *             x[i] *= y[i]             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_x[__pyx_t_3]) = ((__pyx_v_x[__pyx_t_3]) * (__pyx_v_y[__pyx_v_i]));
  }

  /* "linalg.pxd":193
 * 
 *     @staticmethod
 *     cdef inline void mul_i(weight_t* x,             # <<<<<<<<<<<<<<
 *             const weight_t* y, int32_t nr) nogil:
 *         cdef int i
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.mul_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":200
 * 
 *     @staticmethod
 *     cdef inline weight_t dot(             # <<<<<<<<<<<<<<
 *             const weight_t* x, const weight_t* y, int32_t nr) nogil:
 *         cdef int i
 */

static CYTHON_INLINE __pyx_t_5thinc_8typedefs_weight_t __pyx_f_5thinc_6linalg_6VecVec_dot(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr) {
  int __pyx_v_i;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_total;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_r;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("dot", __pyx_f[1], 200, 1, __PYX_ERR(1, 200, __pyx_L1_error));

  /* "linalg.pxd":203
 *             const weight_t* x, const weight_t* y, int32_t nr) nogil:
 *         cdef int i
 *         cdef weight_t total = 0             # <<<<<<<<<<<<<<
 *         for i in range(nr):
 *             total += x[i] * y[i]
 */
  __pyx_v_total = 0.0;

  /* "linalg.pxd":204
 *         cdef int i
 *         cdef weight_t total = 0
 *         for i in range(nr):             # <<<<<<<<<<<<<<
 *             total += x[i] * y[i]
 *         return total
 */
  __pyx_t_1 = __pyx_v_nr;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":205
 *         cdef weight_t total = 0
 *         for i in range(nr):
 *             total += x[i] * y[i]             # <<<<<<<<<<<<<<
 *         return total
 * 
 */
    __pyx_v_total = (__pyx_v_total + ((__pyx_v_x[__pyx_v_i]) * (__pyx_v_y[__pyx_v_i])));
  }

  /* "linalg.pxd":206
 *         for i in range(nr):
 *             total += x[i] * y[i]
 *         return total             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = __pyx_v_total;
  goto __pyx_L0;

  /* "linalg.pxd":200
 * 
 *     @staticmethod
 *     cdef inline weight_t dot(             # <<<<<<<<<<<<<<
 *             const weight_t* x, const weight_t* y, int32_t nr) nogil:
 *         cdef int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.dot", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "linalg.pxd":209
 * 
 *     @staticmethod
 *     cdef inline int arg_max_if_true(             # <<<<<<<<<<<<<<
 *             const weight_t* scores, const int* is_valid, const int n_classes) nogil:
 *         cdef int i
 */

static CYTHON_INLINE int __pyx_f_5thinc_6linalg_6VecVec_arg_max_if_true(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_scores, int const *__pyx_v_is_valid, int const __pyx_v_n_classes) {
  int __pyx_v_i;
  int __pyx_v_best;
  int __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_TraceCall("arg_max_if_true", __pyx_f[1], 209, 1, __PYX_ERR(1, 209, __pyx_L1_error));

  /* "linalg.pxd":212
 *             const weight_t* scores, const int* is_valid, const int n_classes) nogil:
 *         cdef int i
 *         cdef int best = -1             # <<<<<<<<<<<<<<
 *         for i in range(n_classes):
 *             if is_valid[i] and (best == -1 or scores[i] > scores[best]):
 */
  __pyx_v_best = -1;

  /* "linalg.pxd":213
 *         cdef int i
 *         cdef int best = -1
 *         for i in range(n_classes):             # <<<<<<<<<<<<<<
 *             if is_valid[i] and (best == -1 or scores[i] > scores[best]):
 *                 best = i
 */
  __pyx_t_1 = __pyx_v_n_classes;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":214
 *         cdef int best = -1
 *         for i in range(n_classes):
 *             if is_valid[i] and (best == -1 or scores[i] > scores[best]):             # <<<<<<<<<<<<<<
 *                 best = i
 *         return best
 */
    __pyx_t_4 = ((__pyx_v_is_valid[__pyx_v_i]) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_best == -1L) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_scores[__pyx_v_i]) > (__pyx_v_scores[__pyx_v_best])) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_3) {

      /* "linalg.pxd":215
 *         for i in range(n_classes):
 *             if is_valid[i] and (best == -1 or scores[i] > scores[best]):
 *                 best = i             # <<<<<<<<<<<<<<
 *         return best
 * 
 */
      __pyx_v_best = __pyx_v_i;

      /* "linalg.pxd":214
 *         cdef int best = -1
 *         for i in range(n_classes):
 *             if is_valid[i] and (best == -1 or scores[i] > scores[best]):             # <<<<<<<<<<<<<<
 *                 best = i
 *         return best
 */
    }
  }

  /* "linalg.pxd":216
 *             if is_valid[i] and (best == -1 or scores[i] > scores[best]):
 *                 best = i
 *         return best             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = __pyx_v_best;
  goto __pyx_L0;

  /* "linalg.pxd":209
 * 
 *     @staticmethod
 *     cdef inline int arg_max_if_true(             # <<<<<<<<<<<<<<
 *             const weight_t* scores, const int* is_valid, const int n_classes) nogil:
 *         cdef int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.arg_max_if_true", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "linalg.pxd":219
 * 
 *     @staticmethod
 *     cdef inline int arg_max_if_zero(             # <<<<<<<<<<<<<<
 *             const weight_t* scores, const weight_t* costs, const int n_classes) nogil:
 *         cdef int i
 */

static CYTHON_INLINE int __pyx_f_5thinc_6linalg_6VecVec_arg_max_if_zero(__pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_scores, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_costs, int const __pyx_v_n_classes) {
  int __pyx_v_i;
  int __pyx_v_best;
  int __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_TraceCall("arg_max_if_zero", __pyx_f[1], 219, 1, __PYX_ERR(1, 219, __pyx_L1_error));

  /* "linalg.pxd":222
 *             const weight_t* scores, const weight_t* costs, const int n_classes) nogil:
 *         cdef int i
 *         cdef int best = -1             # <<<<<<<<<<<<<<
 *         for i in range(n_classes):
 *             if costs[i] == 0 and (best == -1 or scores[i] > scores[best]):
 */
  __pyx_v_best = -1;

  /* "linalg.pxd":223
 *         cdef int i
 *         cdef int best = -1
 *         for i in range(n_classes):             # <<<<<<<<<<<<<<
 *             if costs[i] == 0 and (best == -1 or scores[i] > scores[best]):
 *                 best = i
 */
  __pyx_t_1 = __pyx_v_n_classes;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":224
 *         cdef int best = -1
 *         for i in range(n_classes):
 *             if costs[i] == 0 and (best == -1 or scores[i] > scores[best]):             # <<<<<<<<<<<<<<
 *                 best = i
 *         return best
 */
    __pyx_t_4 = (((__pyx_v_costs[__pyx_v_i]) == 0.0) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_best == -1L) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_scores[__pyx_v_i]) > (__pyx_v_scores[__pyx_v_best])) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_3) {

      /* "linalg.pxd":225
 *         for i in range(n_classes):
 *             if costs[i] == 0 and (best == -1 or scores[i] > scores[best]):
 *                 best = i             # <<<<<<<<<<<<<<
 *         return best
 * 
 */
      __pyx_v_best = __pyx_v_i;

      /* "linalg.pxd":224
 *         cdef int best = -1
 *         for i in range(n_classes):
 *             if costs[i] == 0 and (best == -1 or scores[i] > scores[best]):             # <<<<<<<<<<<<<<
 *                 best = i
 *         return best
 */
    }
  }

  /* "linalg.pxd":226
 *             if costs[i] == 0 and (best == -1 or scores[i] > scores[best]):
 *                 best = i
 *         return best             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_best;
  goto __pyx_L0;

  /* "linalg.pxd":219
 * 
 *     @staticmethod
 *     cdef inline int arg_max_if_zero(             # <<<<<<<<<<<<<<
 *             const weight_t* scores, const weight_t* costs, const int n_classes) nogil:
 *         cdef int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.VecVec.arg_max_if_zero", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "linalg.pxd":231
 * cdef class Mat:
 *     @staticmethod
 *     cdef inline void mean_row(weight_t* Ex,             # <<<<<<<<<<<<<<
 *             const weight_t* mat, int32_t nr_row, int32_t nr_col) nogil:
 *         memset(Ex, 0, sizeof(Ex[0]) * nr_col)
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Mat_mean_row(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_Ex, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  int32_t __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int32_t __pyx_t_2;
  __Pyx_TraceCall("mean_row", __pyx_f[1], 231, 1, __PYX_ERR(1, 231, __pyx_L1_error));

  /* "linalg.pxd":233
 *     cdef inline void mean_row(weight_t* Ex,
 *             const weight_t* mat, int32_t nr_row, int32_t nr_col) nogil:
 *         memset(Ex, 0, sizeof(Ex[0]) * nr_col)             # <<<<<<<<<<<<<<
 *         for i in range(nr_row):
 *             VecVec.add_i(Ex, &mat[i * nr_col], 1.0, nr_col)
 */
  memset(__pyx_v_Ex, 0, ((sizeof((__pyx_v_Ex[0]))) * __pyx_v_nr_col));

  /* "linalg.pxd":234
 *             const weight_t* mat, int32_t nr_row, int32_t nr_col) nogil:
 *         memset(Ex, 0, sizeof(Ex[0]) * nr_col)
 *         for i in range(nr_row):             # <<<<<<<<<<<<<<
 *             VecVec.add_i(Ex, &mat[i * nr_col], 1.0, nr_col)
 *         Vec.mul_i(Ex, 1.0 / nr_row, nr_col)
 */
  __pyx_t_1 = __pyx_v_nr_row;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":235
 *         memset(Ex, 0, sizeof(Ex[0]) * nr_col)
 *         for i in range(nr_row):
 *             VecVec.add_i(Ex, &mat[i * nr_col], 1.0, nr_col)             # <<<<<<<<<<<<<<
 *         Vec.mul_i(Ex, 1.0 / nr_row, nr_col)
 * 
 */
    __pyx_f_5thinc_6linalg_6VecVec_add_i(__pyx_v_Ex, (&(__pyx_v_mat[(__pyx_v_i * __pyx_v_nr_col)])), 1.0, __pyx_v_nr_col);
  }

  /* "linalg.pxd":236
 *         for i in range(nr_row):
 *             VecVec.add_i(Ex, &mat[i * nr_col], 1.0, nr_col)
 *         Vec.mul_i(Ex, 1.0 / nr_row, nr_col)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_3Vec_mul_i(__pyx_v_Ex, (1.0 / __pyx_v_nr_row), __pyx_v_nr_col);

  /* "linalg.pxd":231
 * cdef class Mat:
 *     @staticmethod
 *     cdef inline void mean_row(weight_t* Ex,             # <<<<<<<<<<<<<<
 *             const weight_t* mat, int32_t nr_row, int32_t nr_col) nogil:
 *         memset(Ex, 0, sizeof(Ex[0]) * nr_col)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Mat.mean_row", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":239
 * 
 *     @staticmethod
 *     cdef inline void var_row(weight_t* Vx,             # <<<<<<<<<<<<<<
 *             const weight_t* mat, const weight_t* Ex,
 *             int32_t nr_row, int32_t nr_col, weight_t eps) nogil:
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_3Mat_var_row(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_Vx, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_Ex, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_eps) {
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_sum_;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_sum2;
  int32_t __pyx_v_i;
  int32_t __pyx_v_j;
  __pyx_t_5thinc_8typedefs_weight_t __pyx_v_x;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int32_t __pyx_t_3;
  int32_t __pyx_t_4;
  int32_t __pyx_t_5;
  int32_t __pyx_t_6;
  __Pyx_TraceCall("var_row", __pyx_f[1], 239, 1, __PYX_ERR(1, 239, __pyx_L1_error));

  /* "linalg.pxd":243
 *             int32_t nr_row, int32_t nr_col, weight_t eps) nogil:
 *         # From https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
 *         if nr_row == 0 or nr_col == 0:             # <<<<<<<<<<<<<<
 *             return
 *         cdef weight_t sum_, sum2
 */
  __pyx_t_2 = ((__pyx_v_nr_row == 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_nr_col == 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "linalg.pxd":244
 *         # From https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
 *         if nr_row == 0 or nr_col == 0:
 *             return             # <<<<<<<<<<<<<<
 *         cdef weight_t sum_, sum2
 *         for i in range(nr_col):
 */
    goto __pyx_L0;

    /* "linalg.pxd":243
 *             int32_t nr_row, int32_t nr_col, weight_t eps) nogil:
 *         # From https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
 *         if nr_row == 0 or nr_col == 0:             # <<<<<<<<<<<<<<
 *             return
 *         cdef weight_t sum_, sum2
 */
  }

  /* "linalg.pxd":246
 *             return
 *         cdef weight_t sum_, sum2
 *         for i in range(nr_col):             # <<<<<<<<<<<<<<
 *             sum_ = 0.0
 *             sum2 = 0.0
 */
  __pyx_t_3 = __pyx_v_nr_col;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "linalg.pxd":247
 *         cdef weight_t sum_, sum2
 *         for i in range(nr_col):
 *             sum_ = 0.0             # <<<<<<<<<<<<<<
 *             sum2 = 0.0
 *             for j in range(nr_row):
 */
    __pyx_v_sum_ = 0.0;

    /* "linalg.pxd":248
 *         for i in range(nr_col):
 *             sum_ = 0.0
 *             sum2 = 0.0             # <<<<<<<<<<<<<<
 *             for j in range(nr_row):
 *                 x = mat[j * nr_col + i]
 */
    __pyx_v_sum2 = 0.0;

    /* "linalg.pxd":249
 *             sum_ = 0.0
 *             sum2 = 0.0
 *             for j in range(nr_row):             # <<<<<<<<<<<<<<
 *                 x = mat[j * nr_col + i]
 *                 sum2 += (x - Ex[i]) ** 2
 */
    __pyx_t_5 = __pyx_v_nr_row;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "linalg.pxd":250
 *             sum2 = 0.0
 *             for j in range(nr_row):
 *                 x = mat[j * nr_col + i]             # <<<<<<<<<<<<<<
 *                 sum2 += (x - Ex[i]) ** 2
 *                 sum_ += x - Ex[i]
 */
      __pyx_v_x = (__pyx_v_mat[((__pyx_v_j * __pyx_v_nr_col) + __pyx_v_i)]);

      /* "linalg.pxd":251
 *             for j in range(nr_row):
 *                 x = mat[j * nr_col + i]
 *                 sum2 += (x - Ex[i]) ** 2             # <<<<<<<<<<<<<<
 *                 sum_ += x - Ex[i]
 *             Vx[i] = (sum2 - sum_**2 / nr_row) / nr_row
 */
      __pyx_v_sum2 = (__pyx_v_sum2 + powf((__pyx_v_x - (__pyx_v_Ex[__pyx_v_i])), 2.0));

      /* "linalg.pxd":252
 *                 x = mat[j * nr_col + i]
 *                 sum2 += (x - Ex[i]) ** 2
 *                 sum_ += x - Ex[i]             # <<<<<<<<<<<<<<
 *             Vx[i] = (sum2 - sum_**2 / nr_row) / nr_row
 *             Vx[i] += eps
 */
      __pyx_v_sum_ = (__pyx_v_sum_ + (__pyx_v_x - (__pyx_v_Ex[__pyx_v_i])));
    }

    /* "linalg.pxd":253
 *                 sum2 += (x - Ex[i]) ** 2
 *                 sum_ += x - Ex[i]
 *             Vx[i] = (sum2 - sum_**2 / nr_row) / nr_row             # <<<<<<<<<<<<<<
 *             Vx[i] += eps
 * 
 */
    (__pyx_v_Vx[__pyx_v_i]) = ((__pyx_v_sum2 - (powf(__pyx_v_sum_, 2.0) / __pyx_v_nr_row)) / __pyx_v_nr_row);

    /* "linalg.pxd":254
 *                 sum_ += x - Ex[i]
 *             Vx[i] = (sum2 - sum_**2 / nr_row) / nr_row
 *             Vx[i] += eps             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_5 = __pyx_v_i;
    (__pyx_v_Vx[__pyx_t_5]) = ((__pyx_v_Vx[__pyx_t_5]) + __pyx_v_eps);
  }

  /* "linalg.pxd":239
 * 
 *     @staticmethod
 *     cdef inline void var_row(weight_t* Vx,             # <<<<<<<<<<<<<<
 *             const weight_t* mat, const weight_t* Ex,
 *             int32_t nr_row, int32_t nr_col, weight_t eps) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.Mat.var_row", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":259
 * cdef class MatVec:
 *     @staticmethod
 *     cdef inline void add_i(weight_t* mat,             # <<<<<<<<<<<<<<
 *             const weight_t* vec, weight_t scale, int32_t nr_row, int32_t nr_col) nogil:
 *         cdef int i
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, __pyx_t_5thinc_8typedefs_weight_t __pyx_v_scale, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("add_i", __pyx_f[1], 259, 1, __PYX_ERR(1, 259, __pyx_L1_error));

  /* "linalg.pxd":262
 *             const weight_t* vec, weight_t scale, int32_t nr_row, int32_t nr_col) nogil:
 *         cdef int i
 *         for i in range(nr_row):             # <<<<<<<<<<<<<<
 *             VecVec.add_i(mat + (i * nr_col),
 *                 vec, scale, nr_col)
 */
  __pyx_t_1 = __pyx_v_nr_row;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":263
 *         cdef int i
 *         for i in range(nr_row):
 *             VecVec.add_i(mat + (i * nr_col),             # <<<<<<<<<<<<<<
 *                 vec, scale, nr_col)
 * 
 */
    __pyx_f_5thinc_6linalg_6VecVec_add_i((__pyx_v_mat + (__pyx_v_i * __pyx_v_nr_col)), __pyx_v_vec, __pyx_v_scale, __pyx_v_nr_col);
  }

  /* "linalg.pxd":259
 * cdef class MatVec:
 *     @staticmethod
 *     cdef inline void add_i(weight_t* mat,             # <<<<<<<<<<<<<<
 *             const weight_t* vec, weight_t scale, int32_t nr_row, int32_t nr_col) nogil:
 *         cdef int i
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatVec.add_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":267
 * 
 *     @staticmethod
 *     cdef inline void mul(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* mat,
 *                          const weight_t* vec,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_mul(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("mul", __pyx_f[1], 267, 1, __PYX_ERR(1, 267, __pyx_L1_error));

  /* "linalg.pxd":271
 *                          const weight_t* vec,
 *                          int32_t nr_row, int32_t nr_col) nogil:
 *         memcpy(output, mat, sizeof(output[0]) * nr_row * nr_col)             # <<<<<<<<<<<<<<
 *         MatVec.mul_i(output, vec, nr_row, nr_col)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_mat, (((sizeof((__pyx_v_output[0]))) * __pyx_v_nr_row) * __pyx_v_nr_col));

  /* "linalg.pxd":272
 *                          int32_t nr_row, int32_t nr_col) nogil:
 *         memcpy(output, mat, sizeof(output[0]) * nr_row * nr_col)
 *         MatVec.mul_i(output, vec, nr_row, nr_col)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_6MatVec_mul_i(__pyx_v_output, __pyx_v_vec, __pyx_v_nr_row, __pyx_v_nr_col);

  /* "linalg.pxd":267
 * 
 *     @staticmethod
 *     cdef inline void mul(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* mat,
 *                          const weight_t* vec,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatVec.mul", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":275
 * 
 *     @staticmethod
 *     cdef inline void mul_i(weight_t* mat,             # <<<<<<<<<<<<<<
 *                            const weight_t* vec,
 *                            int32_t nr_row, int32_t nr_col) nogil:
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_mul_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  int __pyx_v_i;
  int __pyx_v_row;
  int __pyx_v_col;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int32_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_TraceCall("mul_i", __pyx_f[1], 275, 1, __PYX_ERR(1, 275, __pyx_L1_error));

  /* "linalg.pxd":279
 *                            int32_t nr_row, int32_t nr_col) nogil:
 *         cdef int i, row, col
 *         for i in range(nr_row):             # <<<<<<<<<<<<<<
 *             row = i * nr_col
 *             for col in range(nr_col):
 */
  __pyx_t_1 = __pyx_v_nr_row;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":280
 *         cdef int i, row, col
 *         for i in range(nr_row):
 *             row = i * nr_col             # <<<<<<<<<<<<<<
 *             for col in range(nr_col):
 *                 mat[row + col] *= vec[col]
 */
    __pyx_v_row = (__pyx_v_i * __pyx_v_nr_col);

    /* "linalg.pxd":281
 *         for i in range(nr_row):
 *             row = i * nr_col
 *             for col in range(nr_col):             # <<<<<<<<<<<<<<
 *                 mat[row + col] *= vec[col]
 * 
 */
    __pyx_t_3 = __pyx_v_nr_col;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_col = __pyx_t_4;

      /* "linalg.pxd":282
 *             row = i * nr_col
 *             for col in range(nr_col):
 *                 mat[row + col] *= vec[col]             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
      __pyx_t_5 = (__pyx_v_row + __pyx_v_col);
      (__pyx_v_mat[__pyx_t_5]) = ((__pyx_v_mat[__pyx_t_5]) * (__pyx_v_vec[__pyx_v_col]));
    }
  }

  /* "linalg.pxd":275
 * 
 *     @staticmethod
 *     cdef inline void mul_i(weight_t* mat,             # <<<<<<<<<<<<<<
 *                            const weight_t* vec,
 *                            int32_t nr_row, int32_t nr_col) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatVec.mul_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":285
 * 
 *     @staticmethod
 *     cdef inline void dot(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* mat,
 *                          const weight_t* vec,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_dot(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  int __pyx_v_i;
  int __pyx_v_row;
  int __pyx_v_col;
  CYTHON_UNUSED double __pyx_v_zero;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int32_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_TraceCall("dot", __pyx_f[1], 285, 1, __PYX_ERR(1, 285, __pyx_L1_error));

  /* "linalg.pxd":290
 *                          int32_t nr_row, int32_t nr_col) nogil:
 *         cdef int i, row, col
 *         cdef double zero = 0.0             # <<<<<<<<<<<<<<
 *         IF USE_BLAS:
 *             blis.gemv(
 */
  __pyx_v_zero = 0.0;

  /* "linalg.pxd":304
 *             )
 *         ELSE:
 *             for i in range(nr_row):             # <<<<<<<<<<<<<<
 *                 row = i * nr_col
 *                 for col in range(nr_col):
 */
  __pyx_t_1 = __pyx_v_nr_row;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":305
 *         ELSE:
 *             for i in range(nr_row):
 *                 row = i * nr_col             # <<<<<<<<<<<<<<
 *                 for col in range(nr_col):
 *                     output[i] += mat[row + col] * vec[col]
 */
    __pyx_v_row = (__pyx_v_i * __pyx_v_nr_col);

    /* "linalg.pxd":306
 *             for i in range(nr_row):
 *                 row = i * nr_col
 *                 for col in range(nr_col):             # <<<<<<<<<<<<<<
 *                     output[i] += mat[row + col] * vec[col]
 * 
 */
    __pyx_t_3 = __pyx_v_nr_col;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_col = __pyx_t_4;

      /* "linalg.pxd":307
 *                 row = i * nr_col
 *                 for col in range(nr_col):
 *                     output[i] += mat[row + col] * vec[col]             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
      __pyx_t_5 = __pyx_v_i;
      (__pyx_v_output[__pyx_t_5]) = ((__pyx_v_output[__pyx_t_5]) + ((__pyx_v_mat[(__pyx_v_row + __pyx_v_col)]) * (__pyx_v_vec[__pyx_v_col])));
    }
  }

  /* "linalg.pxd":285
 * 
 *     @staticmethod
 *     cdef inline void dot(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* mat,
 *                          const weight_t* vec,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatVec.dot", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":310
 * 
 *     @staticmethod
 *     cdef inline void batch_dot(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* mat,
 *                          const weight_t* vec,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_batch_dot(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col, int32_t __pyx_v_nr_batch) {
  CYTHON_UNUSED double __pyx_v_one;
  CYTHON_UNUSED int32_t __pyx_v_b;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int32_t __pyx_t_2;
  __Pyx_TraceCall("batch_dot", __pyx_f[1], 310, 1, __PYX_ERR(1, 310, __pyx_L1_error));

  /* "linalg.pxd":325
 *         # out:   M * N
 *         cdef int i, row, col
 *         cdef double one = 1.0             # <<<<<<<<<<<<<<
 *         IF USE_BLAS:
 *             blis.gemm(
 */
  __pyx_v_one = 1.0;

  /* "linalg.pxd":345
 *                 1)
 *         ELSE:
 *             for b in range(nr_batch):             # <<<<<<<<<<<<<<
 *                 MatVec.dot(output,
 *                     mat, vec, nr_row, nr_col)
 */
  __pyx_t_1 = __pyx_v_nr_batch;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_b = __pyx_t_2;

    /* "linalg.pxd":346
 *         ELSE:
 *             for b in range(nr_batch):
 *                 MatVec.dot(output,             # <<<<<<<<<<<<<<
 *                     mat, vec, nr_row, nr_col)
 *                 output += nr_row
 */
    __pyx_f_5thinc_6linalg_6MatVec_dot(__pyx_v_output, __pyx_v_mat, __pyx_v_vec, __pyx_v_nr_row, __pyx_v_nr_col);

    /* "linalg.pxd":348
 *                 MatVec.dot(output,
 *                     mat, vec, nr_row, nr_col)
 *                 output += nr_row             # <<<<<<<<<<<<<<
 *                 vec += nr_col
 * 
 */
    __pyx_v_output = (__pyx_v_output + __pyx_v_nr_row);

    /* "linalg.pxd":349
 *                     mat, vec, nr_row, nr_col)
 *                 output += nr_row
 *                 vec += nr_col             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_v_vec = (__pyx_v_vec + __pyx_v_nr_col);
  }

  /* "linalg.pxd":310
 * 
 *     @staticmethod
 *     cdef inline void batch_dot(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* mat,
 *                          const weight_t* vec,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatVec.batch_dot", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":352
 * 
 *     @staticmethod
 *     cdef inline void T_dot(weight_t* output,             # <<<<<<<<<<<<<<
 *                              const weight_t* mat,
 *                              const weight_t* vec,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_T_dot(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  int __pyx_v_row;
  int __pyx_v_col;
  CYTHON_UNUSED double __pyx_v_zero;
  CYTHON_UNUSED double __pyx_v_one;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int32_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_TraceCall("T_dot", __pyx_f[1], 352, 1, __PYX_ERR(1, 352, __pyx_L1_error));

  /* "linalg.pxd":358
 *                              int32_t nr_col) nogil:
 *         cdef int i, row, col
 *         cdef double zero = 0.0             # <<<<<<<<<<<<<<
 *         cdef double one = 1.0
 *         IF USE_BLAS:
 */
  __pyx_v_zero = 0.0;

  /* "linalg.pxd":359
 *         cdef int i, row, col
 *         cdef double zero = 0.0
 *         cdef double one = 1.0             # <<<<<<<<<<<<<<
 *         IF USE_BLAS:
 *             blis.gemv(
 */
  __pyx_v_one = 1.0;

  /* "linalg.pxd":372
 *             )
 *         ELSE:
 *             for row in range(nr_row):             # <<<<<<<<<<<<<<
 *                 for col in range(nr_col):
 *                     output[col] += vec[row] * mat[(row * nr_col) + col]
 */
  __pyx_t_1 = __pyx_v_nr_row;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_row = __pyx_t_2;

    /* "linalg.pxd":373
 *         ELSE:
 *             for row in range(nr_row):
 *                 for col in range(nr_col):             # <<<<<<<<<<<<<<
 *                     output[col] += vec[row] * mat[(row * nr_col) + col]
 * 
 */
    __pyx_t_3 = __pyx_v_nr_col;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_col = __pyx_t_4;

      /* "linalg.pxd":374
 *             for row in range(nr_row):
 *                 for col in range(nr_col):
 *                     output[col] += vec[row] * mat[(row * nr_col) + col]             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
      __pyx_t_5 = __pyx_v_col;
      (__pyx_v_output[__pyx_t_5]) = ((__pyx_v_output[__pyx_t_5]) + ((__pyx_v_vec[__pyx_v_row]) * (__pyx_v_mat[((__pyx_v_row * __pyx_v_nr_col) + __pyx_v_col)])));
    }
  }

  /* "linalg.pxd":352
 * 
 *     @staticmethod
 *     cdef inline void T_dot(weight_t* output,             # <<<<<<<<<<<<<<
 *                              const weight_t* mat,
 *                              const weight_t* vec,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatVec.T_dot", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":377
 * 
 *     @staticmethod
 *     cdef inline void batch_T_dot(weight_t* output,             # <<<<<<<<<<<<<<
 *                              const weight_t* mat,
 *                              const weight_t* vec,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatVec_batch_T_dot(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_vec, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col, int32_t __pyx_v_nr_batch) {
  CYTHON_UNUSED int __pyx_v__;
  CYTHON_UNUSED double __pyx_v_one;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("batch_T_dot", __pyx_f[1], 377, 1, __PYX_ERR(1, 377, __pyx_L1_error));

  /* "linalg.pxd":384
 *                              int32_t nr_batch) nogil:
 *         cdef int _
 *         cdef double one = 1.0             # <<<<<<<<<<<<<<
 *         IF USE_BLAS:
 *             # output is (nr_batch, nr_col)
 */
  __pyx_v_one = 1.0;

  /* "linalg.pxd":416
 *                 1)
 *         ELSE:
 *             for _ in range(nr_batch):             # <<<<<<<<<<<<<<
 *                 MatVec.T_dot(output,
 *                     mat, vec, nr_row, nr_col)
 */
  __pyx_t_1 = __pyx_v_nr_batch;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v__ = __pyx_t_2;

    /* "linalg.pxd":417
 *         ELSE:
 *             for _ in range(nr_batch):
 *                 MatVec.T_dot(output,             # <<<<<<<<<<<<<<
 *                     mat, vec, nr_row, nr_col)
 *                 output += nr_col
 */
    __pyx_f_5thinc_6linalg_6MatVec_T_dot(__pyx_v_output, __pyx_v_mat, __pyx_v_vec, __pyx_v_nr_row, __pyx_v_nr_col);

    /* "linalg.pxd":419
 *                 MatVec.T_dot(output,
 *                     mat, vec, nr_row, nr_col)
 *                 output += nr_col             # <<<<<<<<<<<<<<
 *                 vec += nr_row
 * 
 */
    __pyx_v_output = (__pyx_v_output + __pyx_v_nr_col);

    /* "linalg.pxd":420
 *                     mat, vec, nr_row, nr_col)
 *                 output += nr_col
 *                 vec += nr_row             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_vec = (__pyx_v_vec + __pyx_v_nr_row);
  }

  /* "linalg.pxd":377
 * 
 *     @staticmethod
 *     cdef inline void batch_T_dot(weight_t* output,             # <<<<<<<<<<<<<<
 *                              const weight_t* mat,
 *                              const weight_t* vec,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatVec.batch_T_dot", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":425
 * cdef class MatMat:
 *     @staticmethod
 *     cdef inline void add(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* x,
 *                          const weight_t* y,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("add", __pyx_f[1], 425, 1, __PYX_ERR(1, 425, __pyx_L1_error));

  /* "linalg.pxd":429
 *                          const weight_t* y,
 *                          int32_t nr_row, int32_t nr_col) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr_row * nr_col)             # <<<<<<<<<<<<<<
 *         MatMat.add_i(output, y, nr_row, nr_col)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_x, (((sizeof((__pyx_v_output[0]))) * __pyx_v_nr_row) * __pyx_v_nr_col));

  /* "linalg.pxd":430
 *                          int32_t nr_row, int32_t nr_col) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr_row * nr_col)
 *         MatMat.add_i(output, y, nr_row, nr_col)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_6MatMat_add_i(__pyx_v_output, __pyx_v_y, __pyx_v_nr_row, __pyx_v_nr_col);

  /* "linalg.pxd":425
 * cdef class MatMat:
 *     @staticmethod
 *     cdef inline void add(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* x,
 *                          const weight_t* y,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatMat.add", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":433
 * 
 *     @staticmethod
 *     cdef inline void add_i(weight_t* x,             # <<<<<<<<<<<<<<
 *                            const weight_t* y,
 *                            int32_t nr_row, int32_t nr_col) nogil:
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  int __pyx_v_i;
  int __pyx_v_row;
  int __pyx_v_col;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int32_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_TraceCall("add_i", __pyx_f[1], 433, 1, __PYX_ERR(1, 433, __pyx_L1_error));

  /* "linalg.pxd":437
 *                            int32_t nr_row, int32_t nr_col) nogil:
 *         cdef int i, row, col
 *         for i in range(nr_row):             # <<<<<<<<<<<<<<
 *             row = i * nr_col
 *             for col in range(nr_col):
 */
  __pyx_t_1 = __pyx_v_nr_row;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":438
 *         cdef int i, row, col
 *         for i in range(nr_row):
 *             row = i * nr_col             # <<<<<<<<<<<<<<
 *             for col in range(nr_col):
 *                 x[row + col] += y[row + col]
 */
    __pyx_v_row = (__pyx_v_i * __pyx_v_nr_col);

    /* "linalg.pxd":439
 *         for i in range(nr_row):
 *             row = i * nr_col
 *             for col in range(nr_col):             # <<<<<<<<<<<<<<
 *                 x[row + col] += y[row + col]
 * 
 */
    __pyx_t_3 = __pyx_v_nr_col;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_col = __pyx_t_4;

      /* "linalg.pxd":440
 *             row = i * nr_col
 *             for col in range(nr_col):
 *                 x[row + col] += y[row + col]             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
      __pyx_t_5 = (__pyx_v_row + __pyx_v_col);
      (__pyx_v_x[__pyx_t_5]) = ((__pyx_v_x[__pyx_t_5]) + (__pyx_v_y[(__pyx_v_row + __pyx_v_col)]));
    }
  }

  /* "linalg.pxd":433
 * 
 *     @staticmethod
 *     cdef inline void add_i(weight_t* x,             # <<<<<<<<<<<<<<
 *                            const weight_t* y,
 *                            int32_t nr_row, int32_t nr_col) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatMat.add_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":443
 * 
 *     @staticmethod
 *     cdef inline void mul(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* x,
 *                          const weight_t* y,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_mul(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("mul", __pyx_f[1], 443, 1, __PYX_ERR(1, 443, __pyx_L1_error));

  /* "linalg.pxd":447
 *                          const weight_t* y,
 *                          int32_t nr_row, int32_t nr_col) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr_row * nr_col)             # <<<<<<<<<<<<<<
 *         MatMat.mul_i(output, y, nr_row, nr_col)
 * 
 */
  memcpy(__pyx_v_output, __pyx_v_x, (((sizeof((__pyx_v_output[0]))) * __pyx_v_nr_row) * __pyx_v_nr_col));

  /* "linalg.pxd":448
 *                          int32_t nr_row, int32_t nr_col) nogil:
 *         memcpy(output, x, sizeof(output[0]) * nr_row * nr_col)
 *         MatMat.mul_i(output, y, nr_row, nr_col)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_f_5thinc_6linalg_6MatMat_mul_i(__pyx_v_output, __pyx_v_y, __pyx_v_nr_row, __pyx_v_nr_col);

  /* "linalg.pxd":443
 * 
 *     @staticmethod
 *     cdef inline void mul(weight_t* output,             # <<<<<<<<<<<<<<
 *                          const weight_t* x,
 *                          const weight_t* y,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatMat.mul", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":451
 * 
 *     @staticmethod
 *     cdef inline void mul_i(weight_t* x,             # <<<<<<<<<<<<<<
 *                            const weight_t* y,
 *                            int32_t nr_row, int32_t nr_col) nogil:
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_mul_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  int __pyx_v_i;
  int __pyx_v_row;
  int __pyx_v_col;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int32_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_TraceCall("mul_i", __pyx_f[1], 451, 1, __PYX_ERR(1, 451, __pyx_L1_error));

  /* "linalg.pxd":455
 *                            int32_t nr_row, int32_t nr_col) nogil:
 *         cdef int i, row, col
 *         for i in range(nr_row):             # <<<<<<<<<<<<<<
 *             row = i * nr_col
 *             for col in range(nr_col):
 */
  __pyx_t_1 = __pyx_v_nr_row;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":456
 *         cdef int i, row, col
 *         for i in range(nr_row):
 *             row = i * nr_col             # <<<<<<<<<<<<<<
 *             for col in range(nr_col):
 *                 x[row + col] *= y[row + col]
 */
    __pyx_v_row = (__pyx_v_i * __pyx_v_nr_col);

    /* "linalg.pxd":457
 *         for i in range(nr_row):
 *             row = i * nr_col
 *             for col in range(nr_col):             # <<<<<<<<<<<<<<
 *                 x[row + col] *= y[row + col]
 * 
 */
    __pyx_t_3 = __pyx_v_nr_col;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_col = __pyx_t_4;

      /* "linalg.pxd":458
 *             row = i * nr_col
 *             for col in range(nr_col):
 *                 x[row + col] *= y[row + col]             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
      __pyx_t_5 = (__pyx_v_row + __pyx_v_col);
      (__pyx_v_x[__pyx_t_5]) = ((__pyx_v_x[__pyx_t_5]) * (__pyx_v_y[(__pyx_v_row + __pyx_v_col)]));
    }
  }

  /* "linalg.pxd":451
 * 
 *     @staticmethod
 *     cdef inline void mul_i(weight_t* x,             # <<<<<<<<<<<<<<
 *                            const weight_t* y,
 *                            int32_t nr_row, int32_t nr_col) nogil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatMat.mul_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":461
 * 
 *     @staticmethod
 *     cdef inline void add_outer_i(weight_t* mat,             # <<<<<<<<<<<<<<
 *                                  const weight_t* x,
 *                                  const weight_t* y,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_add_outer_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_mat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_row;
  CYTHON_UNUSED double __pyx_v_one;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  int32_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_TraceCall("add_outer_i", __pyx_f[1], 461, 1, __PYX_ERR(1, 461, __pyx_L1_error));

  /* "linalg.pxd":467
 *                                  int32_t nr_col) nogil:
 *         cdef int i, j, row
 *         cdef double one = 1.0             # <<<<<<<<<<<<<<
 *         IF USE_BLAS:
 *             blis.ger(
 */
  __pyx_v_one = 1.0;

  /* "linalg.pxd":478
 *             )
 *         ELSE:
 *             for i in range(nr_row):             # <<<<<<<<<<<<<<
 *                 row = i * nr_col
 *                 for j in range(nr_col):
 */
  __pyx_t_1 = __pyx_v_nr_row;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "linalg.pxd":479
 *         ELSE:
 *             for i in range(nr_row):
 *                 row = i * nr_col             # <<<<<<<<<<<<<<
 *                 for j in range(nr_col):
 *                     mat[row + j] += x[i] * y[j]
 */
    __pyx_v_row = (__pyx_v_i * __pyx_v_nr_col);

    /* "linalg.pxd":480
 *             for i in range(nr_row):
 *                 row = i * nr_col
 *                 for j in range(nr_col):             # <<<<<<<<<<<<<<
 *                     mat[row + j] += x[i] * y[j]
 * 
 */
    __pyx_t_3 = __pyx_v_nr_col;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_j = __pyx_t_4;

      /* "linalg.pxd":481
 *                 row = i * nr_col
 *                 for j in range(nr_col):
 *                     mat[row + j] += x[i] * y[j]             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
      __pyx_t_5 = (__pyx_v_row + __pyx_v_j);
      (__pyx_v_mat[__pyx_t_5]) = ((__pyx_v_mat[__pyx_t_5]) + ((__pyx_v_x[__pyx_v_i]) * (__pyx_v_y[__pyx_v_j])));
    }
  }

  /* "linalg.pxd":461
 * 
 *     @staticmethod
 *     cdef inline void add_outer_i(weight_t* mat,             # <<<<<<<<<<<<<<
 *                                  const weight_t* x,
 *                                  const weight_t* y,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatMat.add_outer_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "linalg.pxd":484
 * 
 *     @staticmethod
 *     cdef inline void batch_add_outer_i(weight_t* output,             # <<<<<<<<<<<<<<
 *                                  const weight_t* x,
 *                                  const weight_t* y,
 */

static CYTHON_INLINE void __pyx_f_5thinc_6linalg_6MatMat_batch_add_outer_i(__pyx_t_5thinc_8typedefs_weight_t *__pyx_v_output, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_x, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_y, int32_t __pyx_v_nr_row, int32_t __pyx_v_nr_col, int32_t __pyx_v_nr_batch) {
  CYTHON_UNUSED double __pyx_v_one;
  CYTHON_UNUSED int32_t __pyx_v__;
  int32_t __pyx_v_i;
  int32_t __pyx_v_row;
  int32_t __pyx_v_j;
  __Pyx_TraceDeclarations
  int32_t __pyx_t_1;
  int32_t __pyx_t_2;
  int32_t __pyx_t_3;
  int32_t __pyx_t_4;
  int32_t __pyx_t_5;
  int32_t __pyx_t_6;
  int32_t __pyx_t_7;
  __Pyx_TraceCall("batch_add_outer_i", __pyx_f[1], 484, 1, __PYX_ERR(1, 484, __pyx_L1_error));

  /* "linalg.pxd":502
 *         # y:    K * N
 *         # out:  M * N
 *         cdef double one = 1.0             # <<<<<<<<<<<<<<
 *         IF USE_BLAS:
 *             blis.gemm(
 */
  __pyx_v_one = 1.0;

  /* "linalg.pxd":522
 *                 1)
 *         ELSE:
 *             for _ in range(nr_batch):             # <<<<<<<<<<<<<<
 *                 for i in range(nr_row):
 *                     row = i * nr_col
 */
  __pyx_t_1 = __pyx_v_nr_batch;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v__ = __pyx_t_2;

    /* "linalg.pxd":523
 *         ELSE:
 *             for _ in range(nr_batch):
 *                 for i in range(nr_row):             # <<<<<<<<<<<<<<
 *                     row = i * nr_col
 *                     for j in range(nr_col):
 */
    __pyx_t_3 = __pyx_v_nr_row;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "linalg.pxd":524
 *             for _ in range(nr_batch):
 *                 for i in range(nr_row):
 *                     row = i * nr_col             # <<<<<<<<<<<<<<
 *                     for j in range(nr_col):
 *                         output[row + j] += x[i] * y[j]
 */
      __pyx_v_row = (__pyx_v_i * __pyx_v_nr_col);

      /* "linalg.pxd":525
 *                 for i in range(nr_row):
 *                     row = i * nr_col
 *                     for j in range(nr_col):             # <<<<<<<<<<<<<<
 *                         output[row + j] += x[i] * y[j]
 *                 x += nr_row
 */
      __pyx_t_5 = __pyx_v_nr_col;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_j = __pyx_t_6;

        /* "linalg.pxd":526
 *                     row = i * nr_col
 *                     for j in range(nr_col):
 *                         output[row + j] += x[i] * y[j]             # <<<<<<<<<<<<<<
 *                 x += nr_row
 *                 y += nr_col
 */
        __pyx_t_7 = (__pyx_v_row + __pyx_v_j);
        (__pyx_v_output[__pyx_t_7]) = ((__pyx_v_output[__pyx_t_7]) + ((__pyx_v_x[__pyx_v_i]) * (__pyx_v_y[__pyx_v_j])));
      }
    }

    /* "linalg.pxd":527
 *                     for j in range(nr_col):
 *                         output[row + j] += x[i] * y[j]
 *                 x += nr_row             # <<<<<<<<<<<<<<
 *                 y += nr_col
 */
    __pyx_v_x = (__pyx_v_x + __pyx_v_nr_row);

    /* "linalg.pxd":528
 *                         output[row + j] += x[i] * y[j]
 *                 x += nr_row
 *                 y += nr_col             # <<<<<<<<<<<<<<
 */
    __pyx_v_y = (__pyx_v_y + __pyx_v_nr_col);
  }

  /* "linalg.pxd":484
 * 
 *     @staticmethod
 *     cdef inline void batch_add_outer_i(weight_t* output,             # <<<<<<<<<<<<<<
 *                                  const weight_t* x,
 *                                  const weight_t* y,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("thinc.linalg.MatMat.batch_add_outer_i", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "structs.pxd":140
 *     int batch_size
 * 
 *     __init__(len_t* widths, int nr_layer, int batch_size) nogil:             # <<<<<<<<<<<<<<
 *         this.i = 0
 *         this.nr_layer = nr_layer
 */

void __pyx_t_5thinc_7structs_MinibatchC::__pyx_f___init__MinibatchC(__pyx_t_5thinc_8typedefs_len_t *__pyx_v_widths, int __pyx_v_nr_layer, int __pyx_v_batch_size) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("<init>", __pyx_f[2], 140, 1, __PYX_ERR(2, 140, __pyx_L1_error));

  /* "structs.pxd":141
 * 
 *     __init__(len_t* widths, int nr_layer, int batch_size) nogil:
 *         this.i = 0             # <<<<<<<<<<<<<<
 *         this.nr_layer = nr_layer
 *         this.batch_size = batch_size
 */
  this->i = 0;

  /* "structs.pxd":142
 *     __init__(len_t* widths, int nr_layer, int batch_size) nogil:
 *         this.i = 0
 *         this.nr_layer = nr_layer             # <<<<<<<<<<<<<<
 *         this.batch_size = batch_size
 *         this.widths = <len_t*>calloc(nr_layer, sizeof(len_t))
 */
  this->nr_layer = __pyx_v_nr_layer;

  /* "structs.pxd":143
 *         this.i = 0
 *         this.nr_layer = nr_layer
 *         this.batch_size = batch_size             # <<<<<<<<<<<<<<
 *         this.widths = <len_t*>calloc(nr_layer, sizeof(len_t))
 *         this._fwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
 */
  this->batch_size = __pyx_v_batch_size;

  /* "structs.pxd":144
 *         this.nr_layer = nr_layer
 *         this.batch_size = batch_size
 *         this.widths = <len_t*>calloc(nr_layer, sizeof(len_t))             # <<<<<<<<<<<<<<
 *         this._fwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
 *         this._bwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
 */
  this->widths = ((__pyx_t_5thinc_8typedefs_len_t *)calloc(__pyx_v_nr_layer, (sizeof(__pyx_t_5thinc_8typedefs_len_t))));

  /* "structs.pxd":145
 *         this.batch_size = batch_size
 *         this.widths = <len_t*>calloc(nr_layer, sizeof(len_t))
 *         this._fwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))             # <<<<<<<<<<<<<<
 *         this._bwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
 *         for i in range(nr_layer):
 */
  this->_fwd = ((__pyx_t_5thinc_8typedefs_weight_t **)calloc(__pyx_v_nr_layer, (sizeof(__pyx_t_5thinc_8typedefs_weight_t *))));

  /* "structs.pxd":146
 *         this.widths = <len_t*>calloc(nr_layer, sizeof(len_t))
 *         this._fwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
 *         this._bwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))             # <<<<<<<<<<<<<<
 *         for i in range(nr_layer):
 *             this.widths[i] = widths[i]
 */
  this->_bwd = ((__pyx_t_5thinc_8typedefs_weight_t **)calloc(__pyx_v_nr_layer, (sizeof(__pyx_t_5thinc_8typedefs_weight_t *))));

  /* "structs.pxd":147
 *         this._fwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
 *         this._bwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
 *         for i in range(nr_layer):             # <<<<<<<<<<<<<<
 *             this.widths[i] = widths[i]
 *             this._fwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
 */
  __pyx_t_1 = __pyx_v_nr_layer;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "structs.pxd":148
 *         this._bwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
 *         for i in range(nr_layer):
 *             this.widths[i] = widths[i]             # <<<<<<<<<<<<<<
 *             this._fwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
 *             this._bwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
 */
    (this->widths[__pyx_v_i]) = (__pyx_v_widths[__pyx_v_i]);

    /* "structs.pxd":149
 *         for i in range(nr_layer):
 *             this.widths[i] = widths[i]
 *             this._fwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))             # <<<<<<<<<<<<<<
 *             this._bwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
 *         this._feats = <FeatureC**>calloc(batch_size, sizeof(void*))
 */
    (this->_fwd[__pyx_v_i]) = ((__pyx_t_5thinc_8typedefs_weight_t *)calloc(((this->widths[__pyx_v_i]) * __pyx_v_batch_size), (sizeof(__pyx_t_5thinc_8typedefs_weight_t))));

    /* "structs.pxd":150
 *             this.widths[i] = widths[i]
 *             this._fwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
 *             this._bwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))             # <<<<<<<<<<<<<<
 *         this._feats = <FeatureC**>calloc(batch_size, sizeof(void*))
 *         this._nr_feat = <len_t*>calloc(batch_size, sizeof(len_t))
 */
    (this->_bwd[__pyx_v_i]) = ((__pyx_t_5thinc_8typedefs_weight_t *)calloc(((this->widths[__pyx_v_i]) * __pyx_v_batch_size), (sizeof(__pyx_t_5thinc_8typedefs_weight_t))));
  }

  /* "structs.pxd":151
 *             this._fwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
 *             this._bwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
 *         this._feats = <FeatureC**>calloc(batch_size, sizeof(void*))             # <<<<<<<<<<<<<<
 *         this._nr_feat = <len_t*>calloc(batch_size, sizeof(len_t))
 *         this._is_valid = <int*>calloc(batch_size * widths[nr_layer-1], sizeof(int))
 */
  this->_feats = ((struct __pyx_t_5thinc_7structs_FeatureC **)calloc(__pyx_v_batch_size, (sizeof(void *))));

  /* "structs.pxd":152
 *             this._bwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
 *         this._feats = <FeatureC**>calloc(batch_size, sizeof(void*))
 *         this._nr_feat = <len_t*>calloc(batch_size, sizeof(len_t))             # <<<<<<<<<<<<<<
 *         this._is_valid = <int*>calloc(batch_size * widths[nr_layer-1], sizeof(int))
 *         this._costs = <weight_t*>calloc(batch_size * widths[nr_layer-1], sizeof(weight_t))
 */
  this->_nr_feat = ((__pyx_t_5thinc_8typedefs_len_t *)calloc(__pyx_v_batch_size, (sizeof(__pyx_t_5thinc_8typedefs_len_t))));

  /* "structs.pxd":153
 *         this._feats = <FeatureC**>calloc(batch_size, sizeof(void*))
 *         this._nr_feat = <len_t*>calloc(batch_size, sizeof(len_t))
 *         this._is_valid = <int*>calloc(batch_size * widths[nr_layer-1], sizeof(int))             # <<<<<<<<<<<<<<
 *         this._costs = <weight_t*>calloc(batch_size * widths[nr_layer-1], sizeof(weight_t))
 *         this.signatures = <uint64_t*>calloc(batch_size, sizeof(uint64_t))
 */
  this->_is_valid = ((int *)calloc((__pyx_v_batch_size * (__pyx_v_widths[(__pyx_v_nr_layer - 1)])), (sizeof(int))));

  /* "structs.pxd":154
 *         this._nr_feat = <len_t*>calloc(batch_size, sizeof(len_t))
 *         this._is_valid = <int*>calloc(batch_size * widths[nr_layer-1], sizeof(int))
 *         this._costs = <weight_t*>calloc(batch_size * widths[nr_layer-1], sizeof(weight_t))             # <<<<<<<<<<<<<<
 *         this.signatures = <uint64_t*>calloc(batch_size, sizeof(uint64_t))
 * 
 */
  this->_costs = ((__pyx_t_5thinc_8typedefs_weight_t *)calloc((__pyx_v_batch_size * (__pyx_v_widths[(__pyx_v_nr_layer - 1)])), (sizeof(__pyx_t_5thinc_8typedefs_weight_t))));

  /* "structs.pxd":155
 *         this._is_valid = <int*>calloc(batch_size * widths[nr_layer-1], sizeof(int))
 *         this._costs = <weight_t*>calloc(batch_size * widths[nr_layer-1], sizeof(weight_t))
 *         this.signatures = <uint64_t*>calloc(batch_size, sizeof(uint64_t))             # <<<<<<<<<<<<<<
 * 
 *     __dealloc__() nogil:
 */
  this->signatures = ((uint64_t *)calloc(__pyx_v_batch_size, (sizeof(uint64_t))));

  /* "structs.pxd":140
 *     int batch_size
 * 
 *     __init__(len_t* widths, int nr_layer, int batch_size) nogil:             # <<<<<<<<<<<<<<
 *         this.i = 0
 *         this.nr_layer = nr_layer
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.<init>", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "structs.pxd":157
 *         this.signatures = <uint64_t*>calloc(batch_size, sizeof(uint64_t))
 * 
 *     __dealloc__() nogil:             # <<<<<<<<<<<<<<
 *         free(this.widths)
 *         for i in range(this.nr_layer):
 */

void __pyx_t_5thinc_7structs_MinibatchC::__pyx_f___dealloc__MinibatchC(void) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("<del>", __pyx_f[2], 157, 1, __PYX_ERR(2, 157, __pyx_L1_error));

  /* "structs.pxd":158
 * 
 *     __dealloc__() nogil:
 *         free(this.widths)             # <<<<<<<<<<<<<<
 *         for i in range(this.nr_layer):
 *             free(this._fwd[i])
 */
  free(this->widths);

  /* "structs.pxd":159
 *     __dealloc__() nogil:
 *         free(this.widths)
 *         for i in range(this.nr_layer):             # <<<<<<<<<<<<<<
 *             free(this._fwd[i])
 *             free(this._bwd[i])
 */
  __pyx_t_1 = this->nr_layer;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "structs.pxd":160
 *         free(this.widths)
 *         for i in range(this.nr_layer):
 *             free(this._fwd[i])             # <<<<<<<<<<<<<<
 *             free(this._bwd[i])
 *         for i in range(this.i):
 */
    free((this->_fwd[__pyx_v_i]));

    /* "structs.pxd":161
 *         for i in range(this.nr_layer):
 *             free(this._fwd[i])
 *             free(this._bwd[i])             # <<<<<<<<<<<<<<
 *         for i in range(this.i):
 *             free(this._feats[i])
 */
    free((this->_bwd[__pyx_v_i]));
  }

  /* "structs.pxd":162
 *             free(this._fwd[i])
 *             free(this._bwd[i])
 *         for i in range(this.i):             # <<<<<<<<<<<<<<
 *             free(this._feats[i])
 *         free(this._fwd)
 */
  __pyx_t_1 = this->i;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "structs.pxd":163
 *             free(this._bwd[i])
 *         for i in range(this.i):
 *             free(this._feats[i])             # <<<<<<<<<<<<<<
 *         free(this._fwd)
 *         free(this._bwd)
 */
    free((this->_feats[__pyx_v_i]));
  }

  /* "structs.pxd":164
 *         for i in range(this.i):
 *             free(this._feats[i])
 *         free(this._fwd)             # <<<<<<<<<<<<<<
 *         free(this._bwd)
 *         free(this._feats)
 */
  free(this->_fwd);

  /* "structs.pxd":165
 *             free(this._feats[i])
 *         free(this._fwd)
 *         free(this._bwd)             # <<<<<<<<<<<<<<
 *         free(this._feats)
 *         free(this._nr_feat)
 */
  free(this->_bwd);

  /* "structs.pxd":166
 *         free(this._fwd)
 *         free(this._bwd)
 *         free(this._feats)             # <<<<<<<<<<<<<<
 *         free(this._nr_feat)
 *         free(this._is_valid)
 */
  free(this->_feats);

  /* "structs.pxd":167
 *         free(this._bwd)
 *         free(this._feats)
 *         free(this._nr_feat)             # <<<<<<<<<<<<<<
 *         free(this._is_valid)
 *         free(this._costs)
 */
  free(this->_nr_feat);

  /* "structs.pxd":168
 *         free(this._feats)
 *         free(this._nr_feat)
 *         free(this._is_valid)             # <<<<<<<<<<<<<<
 *         free(this._costs)
 *         free(this.signatures)
 */
  free(this->_is_valid);

  /* "structs.pxd":169
 *         free(this._nr_feat)
 *         free(this._is_valid)
 *         free(this._costs)             # <<<<<<<<<<<<<<
 *         free(this.signatures)
 * 
 */
  free(this->_costs);

  /* "structs.pxd":170
 *         free(this._is_valid)
 *         free(this._costs)
 *         free(this.signatures)             # <<<<<<<<<<<<<<
 * 
 *     void reset() nogil:
 */
  free(this->signatures);

  /* "structs.pxd":157
 *         this.signatures = <uint64_t*>calloc(batch_size, sizeof(uint64_t))
 * 
 *     __dealloc__() nogil:             # <<<<<<<<<<<<<<
 *         free(this.widths)
 *         for i in range(this.nr_layer):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.<del>", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "structs.pxd":172
 *         free(this.signatures)
 * 
 *     void reset() nogil:             # <<<<<<<<<<<<<<
 *         for i in range(this.nr_layer):
 *             memset(this._fwd[i],
 */

void __pyx_t_5thinc_7structs_MinibatchC::reset(void) {
  int __pyx_v_i;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_TraceCall("reset", __pyx_f[2], 172, 1, __PYX_ERR(2, 172, __pyx_L1_error));

  /* "structs.pxd":173
 * 
 *     void reset() nogil:
 *         for i in range(this.nr_layer):             # <<<<<<<<<<<<<<
 *             memset(this._fwd[i],
 *                 0, sizeof(this._fwd[i][0]) * this.batch_size * this.widths[i])
 */
  __pyx_t_1 = this->nr_layer;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "structs.pxd":174
 *     void reset() nogil:
 *         for i in range(this.nr_layer):
 *             memset(this._fwd[i],             # <<<<<<<<<<<<<<
 *                 0, sizeof(this._fwd[i][0]) * this.batch_size * this.widths[i])
 *             memset(this._bwd[i],
 */
    memset((this->_fwd[__pyx_v_i]), 0, (((sizeof(((this->_fwd[__pyx_v_i])[0]))) * this->batch_size) * (this->widths[__pyx_v_i])));

    /* "structs.pxd":176
 *             memset(this._fwd[i],
 *                 0, sizeof(this._fwd[i][0]) * this.batch_size * this.widths[i])
 *             memset(this._bwd[i],             # <<<<<<<<<<<<<<
 *                 0, sizeof(this._bwd[i][0]) * this.batch_size * this.widths[i])
 *         memset(this._nr_feat, 0, sizeof(this._nr_feat[0]) * this.batch_size)
 */
    memset((this->_bwd[__pyx_v_i]), 0, (((sizeof(((this->_bwd[__pyx_v_i])[0]))) * this->batch_size) * (this->widths[__pyx_v_i])));
  }

  /* "structs.pxd":178
 *             memset(this._bwd[i],
 *                 0, sizeof(this._bwd[i][0]) * this.batch_size * this.widths[i])
 *         memset(this._nr_feat, 0, sizeof(this._nr_feat[0]) * this.batch_size)             # <<<<<<<<<<<<<<
 *         memset(this.signatures, 0, sizeof(this.signatures[0]) * this.batch_size)
 *         memset(this._costs,
 */
  memset(this->_nr_feat, 0, ((sizeof((this->_nr_feat[0]))) * this->batch_size));

  /* "structs.pxd":179
 *                 0, sizeof(this._bwd[i][0]) * this.batch_size * this.widths[i])
 *         memset(this._nr_feat, 0, sizeof(this._nr_feat[0]) * this.batch_size)
 *         memset(this.signatures, 0, sizeof(this.signatures[0]) * this.batch_size)             # <<<<<<<<<<<<<<
 *         memset(this._costs,
 *             0, sizeof(this._costs[0]) * this.nr_out() * this.batch_size)
 */
  memset(this->signatures, 0, ((sizeof((this->signatures[0]))) * this->batch_size));

  /* "structs.pxd":180
 *         memset(this._nr_feat, 0, sizeof(this._nr_feat[0]) * this.batch_size)
 *         memset(this.signatures, 0, sizeof(this.signatures[0]) * this.batch_size)
 *         memset(this._costs,             # <<<<<<<<<<<<<<
 *             0, sizeof(this._costs[0]) * this.nr_out() * this.batch_size)
 *         memset(this._is_valid,
 */
  memset(this->_costs, 0, (((sizeof((this->_costs[0]))) * this->nr_out()) * this->batch_size));

  /* "structs.pxd":182
 *         memset(this._costs,
 *             0, sizeof(this._costs[0]) * this.nr_out() * this.batch_size)
 *         memset(this._is_valid,             # <<<<<<<<<<<<<<
 *             0, sizeof(this._is_valid[0]) * this.nr_out() * this.batch_size)
 *         for i in range(this.i):
 */
  memset(this->_is_valid, 0, (((sizeof((this->_is_valid[0]))) * this->nr_out()) * this->batch_size));

  /* "structs.pxd":184
 *         memset(this._is_valid,
 *             0, sizeof(this._is_valid[0]) * this.nr_out() * this.batch_size)
 *         for i in range(this.i):             # <<<<<<<<<<<<<<
 *             free(this._feats[i])
 *             this._feats[i] = NULL
 */
  __pyx_t_1 = this->i;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "structs.pxd":185
 *             0, sizeof(this._is_valid[0]) * this.nr_out() * this.batch_size)
 *         for i in range(this.i):
 *             free(this._feats[i])             # <<<<<<<<<<<<<<
 *             this._feats[i] = NULL
 *         this.i = 0
 */
    free((this->_feats[__pyx_v_i]));

    /* "structs.pxd":186
 *         for i in range(this.i):
 *             free(this._feats[i])
 *             this._feats[i] = NULL             # <<<<<<<<<<<<<<
 *         this.i = 0
 * 
 */
    (this->_feats[__pyx_v_i]) = NULL;
  }

  /* "structs.pxd":187
 *             free(this._feats[i])
 *             this._feats[i] = NULL
 *         this.i = 0             # <<<<<<<<<<<<<<
 * 
 *     int nr_in() nogil:
 */
  this->i = 0;

  /* "structs.pxd":172
 *         free(this.signatures)
 * 
 *     void reset() nogil:             # <<<<<<<<<<<<<<
 *         for i in range(this.nr_layer):
 *             memset(this._fwd[i],
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.reset", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
}

/* "structs.pxd":189
 *         this.i = 0
 * 
 *     int nr_in() nogil:             # <<<<<<<<<<<<<<
 *         return this.widths[0]
 * 
 */

int __pyx_t_5thinc_7structs_MinibatchC::nr_in(void) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("nr_in", __pyx_f[2], 189, 1, __PYX_ERR(2, 189, __pyx_L1_error));

  /* "structs.pxd":190
 * 
 *     int nr_in() nogil:
 *         return this.widths[0]             # <<<<<<<<<<<<<<
 * 
 *     int nr_out() nogil:
 */
  __pyx_r = (this->widths[0]);
  goto __pyx_L0;

  /* "structs.pxd":189
 *         this.i = 0
 * 
 *     int nr_in() nogil:             # <<<<<<<<<<<<<<
 *         return this.widths[0]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.nr_in", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":192
 *         return this.widths[0]
 * 
 *     int nr_out() nogil:             # <<<<<<<<<<<<<<
 *         return this.widths[this.nr_layer - 1]
 * 
 */

int __pyx_t_5thinc_7structs_MinibatchC::nr_out(void) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("nr_out", __pyx_f[2], 192, 1, __PYX_ERR(2, 192, __pyx_L1_error));

  /* "structs.pxd":193
 * 
 *     int nr_out() nogil:
 *         return this.widths[this.nr_layer - 1]             # <<<<<<<<<<<<<<
 * 
 *     int push_back(const FeatureC* feats, int nr_feat,
 */
  __pyx_r = (this->widths[(this->nr_layer - 1)]);
  goto __pyx_L0;

  /* "structs.pxd":192
 *         return this.widths[0]
 * 
 *     int nr_out() nogil:             # <<<<<<<<<<<<<<
 *         return this.widths[this.nr_layer - 1]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.nr_out", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":195
 *         return this.widths[this.nr_layer - 1]
 * 
 *     int push_back(const FeatureC* feats, int nr_feat,             # <<<<<<<<<<<<<<
 *             const weight_t* costs, const int* is_valid, uint64_t key) nogil:
 *         # Hash the features, to see if the batch has a matching input.
 */

int __pyx_t_5thinc_7structs_MinibatchC::push_back(struct __pyx_t_5thinc_7structs_FeatureC const *__pyx_v_feats, int __pyx_v_nr_feat, __pyx_t_5thinc_8typedefs_weight_t const *__pyx_v_costs, int const *__pyx_v_is_valid, uint64_t __pyx_v_key) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_TraceDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_TraceCall("push_back", __pyx_f[2], 195, 1, __PYX_ERR(2, 195, __pyx_L1_error));

  /* "structs.pxd":199
 *         # Hash the features, to see if the batch has a matching input.
 *         # If it does, just update the gradient for it.
 *         if key != 0:             # <<<<<<<<<<<<<<
 *             for i in range(this.i):
 *                 if this.signatures[i] == key:
 */
  __pyx_t_1 = ((__pyx_v_key != 0) != 0);
  if (__pyx_t_1) {

    /* "structs.pxd":200
 *         # If it does, just update the gradient for it.
 *         if key != 0:
 *             for i in range(this.i):             # <<<<<<<<<<<<<<
 *                 if this.signatures[i] == key:
 *                     VecVec.add_i(this.costs(i),
 */
    __pyx_t_2 = this->i;
    for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "structs.pxd":201
 *         if key != 0:
 *             for i in range(this.i):
 *                 if this.signatures[i] == key:             # <<<<<<<<<<<<<<
 *                     VecVec.add_i(this.costs(i),
 *                         costs, 1.0, this.nr_out())
 */
      __pyx_t_1 = (((this->signatures[__pyx_v_i]) == __pyx_v_key) != 0);
      if (__pyx_t_1) {

        /* "structs.pxd":202
 *             for i in range(this.i):
 *                 if this.signatures[i] == key:
 *                     VecVec.add_i(this.costs(i),             # <<<<<<<<<<<<<<
 *                         costs, 1.0, this.nr_out())
 *                     return 0
 */
        __pyx_f_5thinc_6linalg_6VecVec_add_i(this->costs(__pyx_v_i), __pyx_v_costs, 1.0, this->nr_out());

        /* "structs.pxd":204
 *                     VecVec.add_i(this.costs(i),
 *                         costs, 1.0, this.nr_out())
 *                     return 0             # <<<<<<<<<<<<<<
 *         if this.i >= this.batch_size:
 *             this.reset()
 */
        __pyx_r = 0;
        goto __pyx_L0;

        /* "structs.pxd":201
 *         if key != 0:
 *             for i in range(this.i):
 *                 if this.signatures[i] == key:             # <<<<<<<<<<<<<<
 *                     VecVec.add_i(this.costs(i),
 *                         costs, 1.0, this.nr_out())
 */
      }
    }

    /* "structs.pxd":199
 *         # Hash the features, to see if the batch has a matching input.
 *         # If it does, just update the gradient for it.
 *         if key != 0:             # <<<<<<<<<<<<<<
 *             for i in range(this.i):
 *                 if this.signatures[i] == key:
 */
  }

  /* "structs.pxd":205
 *                         costs, 1.0, this.nr_out())
 *                     return 0
 *         if this.i >= this.batch_size:             # <<<<<<<<<<<<<<
 *             this.reset()
 *             this.i = 0 # This is done in reset() --- but make it obvious
 */
  __pyx_t_1 = ((this->i >= this->batch_size) != 0);
  if (__pyx_t_1) {

    /* "structs.pxd":206
 *                     return 0
 *         if this.i >= this.batch_size:
 *             this.reset()             # <<<<<<<<<<<<<<
 *             this.i = 0 # This is done in reset() --- but make it obvious
 * 
 */
    this->reset();

    /* "structs.pxd":207
 *         if this.i >= this.batch_size:
 *             this.reset()
 *             this.i = 0 # This is done in reset() --- but make it obvious             # <<<<<<<<<<<<<<
 * 
 *         this.signatures[this.i] = key
 */
    this->i = 0;

    /* "structs.pxd":205
 *                         costs, 1.0, this.nr_out())
 *                     return 0
 *         if this.i >= this.batch_size:             # <<<<<<<<<<<<<<
 *             this.reset()
 *             this.i = 0 # This is done in reset() --- but make it obvious
 */
  }

  /* "structs.pxd":209
 *             this.i = 0 # This is done in reset() --- but make it obvious
 * 
 *         this.signatures[this.i] = key             # <<<<<<<<<<<<<<
 *         this._nr_feat[this.i] = nr_feat
 *         this._feats[this.i] = <FeatureC*>calloc(nr_feat, sizeof(FeatureC))
 */
  (this->signatures[this->i]) = __pyx_v_key;

  /* "structs.pxd":210
 * 
 *         this.signatures[this.i] = key
 *         this._nr_feat[this.i] = nr_feat             # <<<<<<<<<<<<<<
 *         this._feats[this.i] = <FeatureC*>calloc(nr_feat, sizeof(FeatureC))
 *         memcpy(this._feats[this.i],
 */
  (this->_nr_feat[this->i]) = __pyx_v_nr_feat;

  /* "structs.pxd":211
 *         this.signatures[this.i] = key
 *         this._nr_feat[this.i] = nr_feat
 *         this._feats[this.i] = <FeatureC*>calloc(nr_feat, sizeof(FeatureC))             # <<<<<<<<<<<<<<
 *         memcpy(this._feats[this.i],
 *             feats, nr_feat * sizeof(this._feats[this.i][0]))
 */
  (this->_feats[this->i]) = ((struct __pyx_t_5thinc_7structs_FeatureC *)calloc(__pyx_v_nr_feat, (sizeof(struct __pyx_t_5thinc_7structs_FeatureC))));

  /* "structs.pxd":212
 *         this._nr_feat[this.i] = nr_feat
 *         this._feats[this.i] = <FeatureC*>calloc(nr_feat, sizeof(FeatureC))
 *         memcpy(this._feats[this.i],             # <<<<<<<<<<<<<<
 *             feats, nr_feat * sizeof(this._feats[this.i][0]))
 * 
 */
  memcpy((this->_feats[this->i]), __pyx_v_feats, (__pyx_v_nr_feat * (sizeof(((this->_feats[this->i])[0])))));

  /* "structs.pxd":215
 *             feats, nr_feat * sizeof(this._feats[this.i][0]))
 * 
 *         memcpy(this.costs(this.i),             # <<<<<<<<<<<<<<
 *             costs, this.nr_out() * sizeof(costs[0]))
 *         if is_valid is not NULL:
 */
  memcpy(this->costs(this->i), __pyx_v_costs, (this->nr_out() * (sizeof((__pyx_v_costs[0])))));

  /* "structs.pxd":217
 *         memcpy(this.costs(this.i),
 *             costs, this.nr_out() * sizeof(costs[0]))
 *         if is_valid is not NULL:             # <<<<<<<<<<<<<<
 *             memcpy(this.is_valid(this.i),
 *                 is_valid, this.nr_out() * sizeof(is_valid[0]))
 */
  __pyx_t_1 = ((__pyx_v_is_valid != NULL) != 0);
  if (__pyx_t_1) {

    /* "structs.pxd":218
 *             costs, this.nr_out() * sizeof(costs[0]))
 *         if is_valid is not NULL:
 *             memcpy(this.is_valid(this.i),             # <<<<<<<<<<<<<<
 *                 is_valid, this.nr_out() * sizeof(is_valid[0]))
 *         else:
 */
    memcpy(this->is_valid(this->i), __pyx_v_is_valid, (this->nr_out() * (sizeof((__pyx_v_is_valid[0])))));

    /* "structs.pxd":217
 *         memcpy(this.costs(this.i),
 *             costs, this.nr_out() * sizeof(costs[0]))
 *         if is_valid is not NULL:             # <<<<<<<<<<<<<<
 *             memcpy(this.is_valid(this.i),
 *                 is_valid, this.nr_out() * sizeof(is_valid[0]))
 */
    goto __pyx_L8;
  }

  /* "structs.pxd":221
 *                 is_valid, this.nr_out() * sizeof(is_valid[0]))
 *         else:
 *             for i in range(this.nr_out()):             # <<<<<<<<<<<<<<
 *                 this.is_valid(this.i)[i] = 1
 *         this.i += 1
 */
  /*else*/ {
    __pyx_t_2 = this->nr_out();
    for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "structs.pxd":222
 *         else:
 *             for i in range(this.nr_out()):
 *                 this.is_valid(this.i)[i] = 1             # <<<<<<<<<<<<<<
 *         this.i += 1
 *         return this.i >= this.batch_size
 */
      (this->is_valid(this->i)[__pyx_v_i]) = 1;
    }
  }
  __pyx_L8:;

  /* "structs.pxd":223
 *             for i in range(this.nr_out()):
 *                 this.is_valid(this.i)[i] = 1
 *         this.i += 1             # <<<<<<<<<<<<<<
 *         return this.i >= this.batch_size
 * 
 */
  this->i = (this->i + 1);

  /* "structs.pxd":224
 *                 this.is_valid(this.i)[i] = 1
 *         this.i += 1
 *         return this.i >= this.batch_size             # <<<<<<<<<<<<<<
 * 
 *     FeatureC* features(int i) nogil:
 */
  __pyx_r = (this->i >= this->batch_size);
  goto __pyx_L0;

  /* "structs.pxd":195
 *         return this.widths[this.nr_layer - 1]
 * 
 *     int push_back(const FeatureC* feats, int nr_feat,             # <<<<<<<<<<<<<<
 *             const weight_t* costs, const int* is_valid, uint64_t key) nogil:
 *         # Hash the features, to see if the batch has a matching input.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.push_back", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":226
 *         return this.i >= this.batch_size
 * 
 *     FeatureC* features(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this._feats[i]
 * 
 */

struct __pyx_t_5thinc_7structs_FeatureC *__pyx_t_5thinc_7structs_MinibatchC::features(int __pyx_v_i) {
  struct __pyx_t_5thinc_7structs_FeatureC *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("features", __pyx_f[2], 226, 1, __PYX_ERR(2, 226, __pyx_L1_error));

  /* "structs.pxd":227
 * 
 *     FeatureC* features(int i) nogil:
 *         return this._feats[i]             # <<<<<<<<<<<<<<
 * 
 *     int nr_feat(int i) nogil:
 */
  __pyx_r = (this->_feats[__pyx_v_i]);
  goto __pyx_L0;

  /* "structs.pxd":226
 *         return this.i >= this.batch_size
 * 
 *     FeatureC* features(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this._feats[i]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.features", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":229
 *         return this._feats[i]
 * 
 *     int nr_feat(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this._nr_feat[i]
 * 
 */

int __pyx_t_5thinc_7structs_MinibatchC::nr_feat(int __pyx_v_i) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("nr_feat", __pyx_f[2], 229, 1, __PYX_ERR(2, 229, __pyx_L1_error));

  /* "structs.pxd":230
 * 
 *     int nr_feat(int i) nogil:
 *         return this._nr_feat[i]             # <<<<<<<<<<<<<<
 * 
 *     weight_t* fwd(int i, int j) nogil:
 */
  __pyx_r = (this->_nr_feat[__pyx_v_i]);
  goto __pyx_L0;

  /* "structs.pxd":229
 *         return this._feats[i]
 * 
 *     int nr_feat(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this._nr_feat[i]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.nr_feat", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":232
 *         return this._nr_feat[i]
 * 
 *     weight_t* fwd(int i, int j) nogil:             # <<<<<<<<<<<<<<
 *         return this._fwd[i] + (j * this.widths[i])
 * 
 */

__pyx_t_5thinc_8typedefs_weight_t *__pyx_t_5thinc_7structs_MinibatchC::fwd(int __pyx_v_i, int __pyx_v_j) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("fwd", __pyx_f[2], 232, 1, __PYX_ERR(2, 232, __pyx_L1_error));

  /* "structs.pxd":233
 * 
 *     weight_t* fwd(int i, int j) nogil:
 *         return this._fwd[i] + (j * this.widths[i])             # <<<<<<<<<<<<<<
 * 
 *     weight_t* bwd(int i, int j) nogil:
 */
  __pyx_r = ((this->_fwd[__pyx_v_i]) + (__pyx_v_j * (this->widths[__pyx_v_i])));
  goto __pyx_L0;

  /* "structs.pxd":232
 *         return this._nr_feat[i]
 * 
 *     weight_t* fwd(int i, int j) nogil:             # <<<<<<<<<<<<<<
 *         return this._fwd[i] + (j * this.widths[i])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.fwd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":235
 *         return this._fwd[i] + (j * this.widths[i])
 * 
 *     weight_t* bwd(int i, int j) nogil:             # <<<<<<<<<<<<<<
 *         return this._bwd[i] + (j * this.widths[i])
 * 
 */

__pyx_t_5thinc_8typedefs_weight_t *__pyx_t_5thinc_7structs_MinibatchC::bwd(int __pyx_v_i, int __pyx_v_j) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("bwd", __pyx_f[2], 235, 1, __PYX_ERR(2, 235, __pyx_L1_error));

  /* "structs.pxd":236
 * 
 *     weight_t* bwd(int i, int j) nogil:
 *         return this._bwd[i] + (j * this.widths[i])             # <<<<<<<<<<<<<<
 * 
 *     weight_t* scores(int i) nogil:
 */
  __pyx_r = ((this->_bwd[__pyx_v_i]) + (__pyx_v_j * (this->widths[__pyx_v_i])));
  goto __pyx_L0;

  /* "structs.pxd":235
 *         return this._fwd[i] + (j * this.widths[i])
 * 
 *     weight_t* bwd(int i, int j) nogil:             # <<<<<<<<<<<<<<
 *         return this._bwd[i] + (j * this.widths[i])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.bwd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":238
 *         return this._bwd[i] + (j * this.widths[i])
 * 
 *     weight_t* scores(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this.fwd(this.nr_layer-1, i)
 * 
 */

__pyx_t_5thinc_8typedefs_weight_t *__pyx_t_5thinc_7structs_MinibatchC::scores(int __pyx_v_i) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("scores", __pyx_f[2], 238, 1, __PYX_ERR(2, 238, __pyx_L1_error));

  /* "structs.pxd":239
 * 
 *     weight_t* scores(int i) nogil:
 *         return this.fwd(this.nr_layer-1, i)             # <<<<<<<<<<<<<<
 * 
 *     weight_t* losses(int i) nogil:
 */
  __pyx_r = this->fwd((this->nr_layer - 1), __pyx_v_i);
  goto __pyx_L0;

  /* "structs.pxd":238
 *         return this._bwd[i] + (j * this.widths[i])
 * 
 *     weight_t* scores(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this.fwd(this.nr_layer-1, i)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.scores", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":241
 *         return this.fwd(this.nr_layer-1, i)
 * 
 *     weight_t* losses(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this.bwd(this.nr_layer-1, i)
 * 
 */

__pyx_t_5thinc_8typedefs_weight_t *__pyx_t_5thinc_7structs_MinibatchC::losses(int __pyx_v_i) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("losses", __pyx_f[2], 241, 1, __PYX_ERR(2, 241, __pyx_L1_error));

  /* "structs.pxd":242
 * 
 *     weight_t* losses(int i) nogil:
 *         return this.bwd(this.nr_layer-1, i)             # <<<<<<<<<<<<<<
 * 
 *     weight_t* costs(int i) nogil:
 */
  __pyx_r = this->bwd((this->nr_layer - 1), __pyx_v_i);
  goto __pyx_L0;

  /* "structs.pxd":241
 *         return this.fwd(this.nr_layer-1, i)
 * 
 *     weight_t* losses(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this.bwd(this.nr_layer-1, i)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.losses", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":244
 *         return this.bwd(this.nr_layer-1, i)
 * 
 *     weight_t* costs(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this._costs + (i * this.nr_out())
 * 
 */

__pyx_t_5thinc_8typedefs_weight_t *__pyx_t_5thinc_7structs_MinibatchC::costs(int __pyx_v_i) {
  __pyx_t_5thinc_8typedefs_weight_t *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("costs", __pyx_f[2], 244, 1, __PYX_ERR(2, 244, __pyx_L1_error));

  /* "structs.pxd":245
 * 
 *     weight_t* costs(int i) nogil:
 *         return this._costs + (i * this.nr_out())             # <<<<<<<<<<<<<<
 * 
 *     int* is_valid(int i) nogil:
 */
  __pyx_r = (this->_costs + (__pyx_v_i * this->nr_out()));
  goto __pyx_L0;

  /* "structs.pxd":244
 *         return this.bwd(this.nr_layer-1, i)
 * 
 *     weight_t* costs(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this._costs + (i * this.nr_out())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.costs", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":247
 *         return this._costs + (i * this.nr_out())
 * 
 *     int* is_valid(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this._is_valid + (i * this.nr_out())
 * 
 */

int *__pyx_t_5thinc_7structs_MinibatchC::is_valid(int __pyx_v_i) {
  int *__pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("is_valid", __pyx_f[2], 247, 1, __PYX_ERR(2, 247, __pyx_L1_error));

  /* "structs.pxd":248
 * 
 *     int* is_valid(int i) nogil:
 *         return this._is_valid + (i * this.nr_out())             # <<<<<<<<<<<<<<
 * 
 *     int guess(int i) nogil:
 */
  __pyx_r = (this->_is_valid + (__pyx_v_i * this->nr_out()));
  goto __pyx_L0;

  /* "structs.pxd":247
 *         return this._costs + (i * this.nr_out())
 * 
 *     int* is_valid(int i) nogil:             # <<<<<<<<<<<<<<
 *         return this._is_valid + (i * this.nr_out())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.is_valid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":250
 *         return this._is_valid + (i * this.nr_out())
 * 
 *     int guess(int i) nogil:             # <<<<<<<<<<<<<<
 *         return VecVec.arg_max_if_true(this.scores(i), this.is_valid(i), this.nr_out())
 * 
 */

int __pyx_t_5thinc_7structs_MinibatchC::guess(int __pyx_v_i) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("guess", __pyx_f[2], 250, 1, __PYX_ERR(2, 250, __pyx_L1_error));

  /* "structs.pxd":251
 * 
 *     int guess(int i) nogil:
 *         return VecVec.arg_max_if_true(this.scores(i), this.is_valid(i), this.nr_out())             # <<<<<<<<<<<<<<
 * 
 *     int best(int i) nogil:
 */
  __pyx_r = __pyx_f_5thinc_6linalg_6VecVec_arg_max_if_true(this->scores(__pyx_v_i), this->is_valid(__pyx_v_i), this->nr_out());
  goto __pyx_L0;

  /* "structs.pxd":250
 *         return this._is_valid + (i * this.nr_out())
 * 
 *     int guess(int i) nogil:             # <<<<<<<<<<<<<<
 *         return VecVec.arg_max_if_true(this.scores(i), this.is_valid(i), this.nr_out())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.guess", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

/* "structs.pxd":253
 *         return VecVec.arg_max_if_true(this.scores(i), this.is_valid(i), this.nr_out())
 * 
 *     int best(int i) nogil:             # <<<<<<<<<<<<<<
 *         return VecVec.arg_max_if_zero(this.scores(i), this.costs(i), this.nr_out())
 * 
 */

int __pyx_t_5thinc_7structs_MinibatchC::best(int __pyx_v_i) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_TraceCall("best", __pyx_f[2], 253, 1, __PYX_ERR(2, 253, __pyx_L1_error));

  /* "structs.pxd":254
 * 
 *     int best(int i) nogil:
 *         return VecVec.arg_max_if_zero(this.scores(i), this.costs(i), this.nr_out())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_5thinc_6linalg_6VecVec_arg_max_if_zero(this->scores(__pyx_v_i), this->costs(__pyx_v_i), this->nr_out());
  goto __pyx_L0;

  /* "structs.pxd":253
 *         return VecVec.arg_max_if_true(this.scores(i), this.is_valid(i), this.nr_out())
 * 
 *     int best(int i) nogil:             # <<<<<<<<<<<<<<
 *         return VecVec.arg_max_if_zero(this.scores(i), this.costs(i), this.nr_out())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("MinibatchC.best", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 1);
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__funcs(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__funcs},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_funcs",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 23, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_funcs(void); /*proto*/
PyMODINIT_FUNC init_funcs(void)
#else
PyMODINIT_FUNC PyInit__funcs(void); /*proto*/
PyMODINIT_FUNC PyInit__funcs(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec__funcs(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  __Pyx_TraceDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #endif
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__funcs(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_funcs", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_thinc__neural___funcs) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "thinc.neural._funcs")) {
      if (unlikely(PyDict_SetItemString(modules, "thinc.neural._funcs", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("ELU", (void (*)(void))__pyx_f_5thinc_6neural_6_funcs_ELU, "void (__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_len_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ReLu", (void (*)(void))__pyx_f_5thinc_6neural_6_funcs_ReLu, "void (__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_len_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("softmax", (void (*)(void))__pyx_f_5thinc_6neural_6_funcs_softmax, "void (__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_len_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("d_ReLu", (void (*)(void))__pyx_f_5thinc_6neural_6_funcs_d_ReLu, "void (__pyx_t_5thinc_8typedefs_weight_t *, __pyx_t_5thinc_8typedefs_weight_t const *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  __pyx_ptype_5cymem_5cymem_Pool = __Pyx_ImportType("cymem.cymem", "Pool", sizeof(struct __pyx_obj_5cymem_5cymem_Pool), 1); if (unlikely(!__pyx_ptype_5cymem_5cymem_Pool)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_vtabptr_5cymem_5cymem_Pool = (struct __pyx_vtabstruct_5cymem_5cymem_Pool*)__Pyx_GetVtable(__pyx_ptype_5cymem_5cymem_Pool->tp_dict); if (unlikely(!__pyx_vtabptr_5cymem_5cymem_Pool)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_ptype_5cymem_5cymem_Address = __Pyx_ImportType("cymem.cymem", "Address", sizeof(struct __pyx_obj_5cymem_5cymem_Address), 1); if (unlikely(!__pyx_ptype_5cymem_5cymem_Address)) __PYX_ERR(3, 11, __pyx_L1_error)
  __pyx_ptype_7preshed_4maps_PreshMap = __Pyx_ImportType("preshed.maps", "PreshMap", sizeof(struct __pyx_obj_7preshed_4maps_PreshMap), 1); if (unlikely(!__pyx_ptype_7preshed_4maps_PreshMap)) __PYX_ERR(4, 36, __pyx_L1_error)
  __pyx_vtabptr_7preshed_4maps_PreshMap = (struct __pyx_vtabstruct_7preshed_4maps_PreshMap*)__Pyx_GetVtable(__pyx_ptype_7preshed_4maps_PreshMap->tp_dict); if (unlikely(!__pyx_vtabptr_7preshed_4maps_PreshMap)) __PYX_ERR(4, 36, __pyx_L1_error)
  __pyx_ptype_7preshed_4maps_PreshMapArray = __Pyx_ImportType("preshed.maps", "PreshMapArray", sizeof(struct __pyx_obj_7preshed_4maps_PreshMapArray), 1); if (unlikely(!__pyx_ptype_7preshed_4maps_PreshMapArray)) __PYX_ERR(4, 44, __pyx_L1_error)
  __pyx_vtabptr_7preshed_4maps_PreshMapArray = (struct __pyx_vtabstruct_7preshed_4maps_PreshMapArray*)__Pyx_GetVtable(__pyx_ptype_7preshed_4maps_PreshMapArray->tp_dict); if (unlikely(!__pyx_vtabptr_7preshed_4maps_PreshMapArray)) __PYX_ERR(4, 44, __pyx_L1_error)
  __pyx_ptype_5thinc_6linalg_Matrix = __Pyx_ImportType("thinc.linalg", "Matrix", sizeof(struct __pyx_obj_5thinc_6linalg_Matrix), 1); if (unlikely(!__pyx_ptype_5thinc_6linalg_Matrix)) __PYX_ERR(1, 22, __pyx_L1_error)
  __pyx_ptype_5thinc_6linalg_Vec = __Pyx_ImportType("thinc.linalg", "Vec", sizeof(struct __pyx_obj_5thinc_6linalg_Vec), 1); if (unlikely(!__pyx_ptype_5thinc_6linalg_Vec)) __PYX_ERR(1, 29, __pyx_L1_error)
  __pyx_vtabptr_5thinc_6linalg_Vec = (struct __pyx_vtabstruct_5thinc_6linalg_Vec*)__Pyx_GetVtable(__pyx_ptype_5thinc_6linalg_Vec->tp_dict); if (unlikely(!__pyx_vtabptr_5thinc_6linalg_Vec)) __PYX_ERR(1, 29, __pyx_L1_error)
  __pyx_ptype_5thinc_6linalg_VecVec = __Pyx_ImportType("thinc.linalg", "VecVec", sizeof(struct __pyx_obj_5thinc_6linalg_VecVec), 1); if (unlikely(!__pyx_ptype_5thinc_6linalg_VecVec)) __PYX_ERR(1, 138, __pyx_L1_error)
  __pyx_vtabptr_5thinc_6linalg_VecVec = (struct __pyx_vtabstruct_5thinc_6linalg_VecVec*)__Pyx_GetVtable(__pyx_ptype_5thinc_6linalg_VecVec->tp_dict); if (unlikely(!__pyx_vtabptr_5thinc_6linalg_VecVec)) __PYX_ERR(1, 138, __pyx_L1_error)
  __pyx_ptype_5thinc_6linalg_Mat = __Pyx_ImportType("thinc.linalg", "Mat", sizeof(struct __pyx_obj_5thinc_6linalg_Mat), 1); if (unlikely(!__pyx_ptype_5thinc_6linalg_Mat)) __PYX_ERR(1, 229, __pyx_L1_error)
  __pyx_vtabptr_5thinc_6linalg_Mat = (struct __pyx_vtabstruct_5thinc_6linalg_Mat*)__Pyx_GetVtable(__pyx_ptype_5thinc_6linalg_Mat->tp_dict); if (unlikely(!__pyx_vtabptr_5thinc_6linalg_Mat)) __PYX_ERR(1, 229, __pyx_L1_error)
  __pyx_ptype_5thinc_6linalg_MatVec = __Pyx_ImportType("thinc.linalg", "MatVec", sizeof(struct __pyx_obj_5thinc_6linalg_MatVec), 1); if (unlikely(!__pyx_ptype_5thinc_6linalg_MatVec)) __PYX_ERR(1, 257, __pyx_L1_error)
  __pyx_vtabptr_5thinc_6linalg_MatVec = (struct __pyx_vtabstruct_5thinc_6linalg_MatVec*)__Pyx_GetVtable(__pyx_ptype_5thinc_6linalg_MatVec->tp_dict); if (unlikely(!__pyx_vtabptr_5thinc_6linalg_MatVec)) __PYX_ERR(1, 257, __pyx_L1_error)
  __pyx_ptype_5thinc_6linalg_MatMat = __Pyx_ImportType("thinc.linalg", "MatMat", sizeof(struct __pyx_obj_5thinc_6linalg_MatMat), 1); if (unlikely(!__pyx_ptype_5thinc_6linalg_MatMat)) __PYX_ERR(1, 423, __pyx_L1_error)
  __pyx_vtabptr_5thinc_6linalg_MatMat = (struct __pyx_vtabstruct_5thinc_6linalg_MatMat*)__Pyx_GetVtable(__pyx_ptype_5thinc_6linalg_MatMat->tp_dict); if (unlikely(!__pyx_vtabptr_5thinc_6linalg_MatMat)) __PYX_ERR(1, 423, __pyx_L1_error)
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("murmurhash.mrmr"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "real_hash64", (void (**)(void))&__pyx_f_10murmurhash_4mrmr_real_hash64, "uint64_t (void *, int, uint64_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __Pyx_TraceCall("PyMODINIT_FUNC PyInit__funcs(void)", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "thinc/neural/_funcs.pyx":1
 * # cython: profile=True             # <<<<<<<<<<<<<<
 * # cython: cdivision=True
 * # cython: infer_types=True
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "structs.pxd":253
 *         return VecVec.arg_max_if_true(this.scores(i), this.is_valid(i), this.nr_out())
 * 
 *     int best(int i) nogil:             # <<<<<<<<<<<<<<
 *         return VecVec.arg_max_if_zero(this.scores(i), this.costs(i), this.nr_out())
 * 
 */
  __Pyx_TraceReturn(Py_None, 0);

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init thinc.neural._funcs", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init thinc.neural._funcs");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* Profile */
#if CYTHON_PROFILE
static int __Pyx_TraceSetupAndCall(PyCodeObject** code,
                                   PyFrameObject** frame,
                                   PyThreadState* tstate,
                                   const char *funcname,
                                   const char *srcfile,
                                   int firstlineno) {
    PyObject *type, *value, *traceback;
    int retval;
    if (*frame == NULL || !CYTHON_PROFILE_REUSE_FRAME) {
        if (*code == NULL) {
            *code = __Pyx_createFrameCodeObject(funcname, srcfile, firstlineno);
            if (*code == NULL) return 0;
        }
        *frame = PyFrame_New(
            tstate,                          /*PyThreadState *tstate*/
            *code,                           /*PyCodeObject *code*/
            __pyx_d,                  /*PyObject *globals*/
            0                                /*PyObject *locals*/
        );
        if (*frame == NULL) return 0;
        if (CYTHON_TRACE && (*frame)->f_trace == NULL) {
            Py_INCREF(Py_None);
            (*frame)->f_trace = Py_None;
        }
#if PY_VERSION_HEX < 0x030400B1
    } else {
        (*frame)->f_tstate = tstate;
#endif
    }
      __Pyx_PyFrame_SetLineNumber(*frame, firstlineno);
    retval = 1;
    tstate->tracing++;
    tstate->use_tracing = 0;
    PyErr_Fetch(&type, &value, &traceback);
    #if CYTHON_TRACE
    if (tstate->c_tracefunc)
        retval = tstate->c_tracefunc(tstate->c_traceobj, *frame, PyTrace_CALL, NULL) == 0;
    if (retval && tstate->c_profilefunc)
    #endif
        retval = tstate->c_profilefunc(tstate->c_profileobj, *frame, PyTrace_CALL, NULL) == 0;
    tstate->use_tracing = (tstate->c_profilefunc ||
                           (CYTHON_TRACE && tstate->c_tracefunc));
    tstate->tracing--;
    if (retval) {
        PyErr_Restore(type, value, traceback);
        return tstate->use_tracing && retval;
    } else {
        Py_XDECREF(type);
        Py_XDECREF(value);
        Py_XDECREF(traceback);
        return -1;
    }
}
static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno) {
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    PyCodeObject *py_code = 0;
    #if PY_MAJOR_VERSION < 3
    py_funcname = PyString_FromString(funcname);
    py_srcfile = PyString_FromString(srcfile);
    #else
    py_funcname = PyUnicode_FromString(funcname);
    py_srcfile = PyUnicode_FromString(srcfile);
    #endif
    if (!py_funcname | !py_srcfile) goto bad;
    py_code = PyCode_New(
        0,
        #if PY_MAJOR_VERSION >= 3
        0,
        #endif
        0,
        0,
        CO_OPTIMIZED | CO_NEWLOCALS,
        __pyx_empty_bytes,     /*PyObject *code,*/
        __pyx_empty_tuple,     /*PyObject *consts,*/
        __pyx_empty_tuple,     /*PyObject *names,*/
        __pyx_empty_tuple,     /*PyObject *varnames,*/
        __pyx_empty_tuple,     /*PyObject *freevars,*/
        __pyx_empty_tuple,     /*PyObject *cellvars,*/
        py_srcfile,       /*PyObject *filename,*/
        py_funcname,      /*PyObject *name,*/
        firstlineno,
        __pyx_empty_bytes      /*PyObject *lnotab*/
    );
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return py_code;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = PyDict_GetItem(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value) {
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *x) {
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 2 * PyLong_SHIFT) {
                            return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 3 * PyLong_SHIFT) {
                            return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 4 * PyLong_SHIFT) {
                            return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int32_t) -1;
        }
    } else {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyInt_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            PyErr_Clear();
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
